/*************************************************************
*  This file is part of the Surface Evolver source code.     *
*  Programmer:  Ken Brakke, brakke@susqu.edu                 *
*************************************************************/


/*************************************************************
*
*     file:        dump.c
*
*     Contents:  Functions for making ASCII dump file of surface.
*/

char *color_names[] = {
     "clear",  
     "black",
     "blue",
     "green",
     "cyan",
     "red",
     "magenta",
     "brown",
     "lightgray",
     "darkgray",
     "lightblue",
     "lightgreen",
     "lightcyan",
     "lightred",
     "lightmagenta",
     "yellow",
     "white",
     };
#define COLORNAME(n) (((n)<-1)||((n)>WHITE)?"black":color_names[(n)+1])

/********************************************************
*
*  Function: dump()
*
*  Purpose: Dumps configuration to ASCII file in format
*           suitable for read-in.  This is just the 
*           interactive part, which calls do_dump().
*
*/

#include "include.h"
#include "ytab.h"

void dump()
{
  char name[200];
  char defaultname[200];

  /* construct default name */
  strncpy(defaultname,datafilename,sizeof(defaultname)-5);
  strcat(defaultname,".dmp");
  sprintf(msg,"Enter name of dump file (%s): ",defaultname);
  prompt(msg,name,sizeof(name));
  if ( name[0] == 0 ) do_dump(defaultname);
  else do_dump(name);
}

/************************************************************************
*
* function: top_dump()
*
* purpose: Write out top part of datafile.
*/

void top_dump(fd)
FILE *fd;  /* destination file */
{
  int i,j,k;
  struct gen_quant *q;
  struct method_instance *mi;
  FILE *old_fd = outfd;
  int e_type;

  outfd = fd;

  sprintf(msg,"// datafilename: %s\n",datafilename);
  outstring(msg);
  if ( needed_version[0] )
  { sprintf(msg,"\nEvolver_version %s  // minimal version needed\n\n",
        needed_version);
    outstring(msg);
  }

  if ( keep_macros_flag )
  { outstring("keep_macros\n");
    dump_macros();
  }

#ifdef LONGDOUBLE
  sprintf(msg,"// Total energy: %2.*Lg\n",DPREC,web.total_energy);
#else
  sprintf(msg,"// Total energy: %2.15g\n",web.total_energy);
#endif 
  outstring(msg);
  /* dynamic libraries */
  for ( i = 0 ; i < MAX_DLL ; i++ )
     if ( dll_list[i].name )
     { sprintf(msg,"LOAD_LIBRARY \"%s\"\n",dll_list[i].name);
       outstring(msg);
     }
  /* model info dump */
  if ( web.representation == SIMPLEX ) outstring("SIMPLEX_REPRESENTATION\n"); 
  if ( SDIM != 3 ) 
      { sprintf(msg,"SPACE_DIMENSION %d\n",SDIM); outstring(msg); }
  if ( web.representation == STRING ) outstring("STRING\n\n");
  else if ( web.representation == SOAPFILM ) outstring("SOAPFILM\n\n");
  else { sprintf(msg,"SURFACE_DIMENSION %d\n\n",web.dimension); outstring(msg); }
  if ( web.modeltype == LINEAR ) outstring("LINEAR\n\n");
  else if ( web.modeltype == LAGRANGE ) 
  { sprintf(msg, "LAGRANGE\nlagrange_order %d\n\n",web.lagrange_order);
    outstring(msg);
  }
  else outstring("QUADRATIC\n\n");
  if ( web.symmetry_flag && !web.torus_flag )
     { sprintf(msg,"SYMMETRY_GROUP \"%s\"\n\n",symmetry_name); outstring(msg); }
  /* global variables */
  for ( i = 0 ; i < web.global_count ; i++ )
     { if ( square_curvature_flag && (i == square_curvature_param) ) continue;
       if ( sqgauss_flag && (i == sqgauss_param) ) continue;
       if ( globals[i].flags & FILE_VALUES )
        {
          sprintf(msg,"PARAMETER %s ",globals[i].name);
          outstring(msg); 
          if ( reflevel == 0 )
             sprintf(msg,"PARAMETER_FILE \"%s\" \n",
                  globals[i].value.file.value_file); 
          else    sprintf(msg,"PARAMETER_FILE \"%s\" \n","not dumped");
          outstring(msg); 
        }
        else if ( (globals[i].flags & ORDINARY_PARAM) &&
              !(globals[i].flags & QUANTITY_TYPE) )
          { if ( globals[i].flags & OPTIMIZING_PARAMETER )
                sprintf(msg,"OPTIMIZING_PARAMETER %s ",globals[i].name);
            else 
                sprintf(msg,"PARAMETER %s ",globals[i].name); outstring(msg); 
#ifdef LONGDOUBLE
             sprintf(msg,"= %2.*Lg ",DPREC,globals[i].value.real); outstring(msg); 
#else
             sprintf(msg,"= %2.15g ",globals[i].value.real); outstring(msg); 
#endif 
             if ( globals[i].delta != OPTPARAM_DELTA )
             {
#ifdef LONGDOUBLE
                sprintf(msg,"pdelta = %2.*Lg ",DPREC,globals[i].delta); outstring(msg); 
#else
                sprintf(msg,"pdelta = %2.15g ",globals[i].delta); outstring(msg); 
#endif 
             }
             outstring("\n"); 
             if ( globals[i].scale != 1.0 )
             {
#ifdef LONGDOUBLE
                sprintf(msg,"scale = %2.*Lg ",DPREC,globals[i].scale); outstring(msg); 
#else
                sprintf(msg,"scale = %2.15g ",globals[i].scale); outstring(msg); 
#endif 
             }
          }
     }

  if ( web.full_flag ) outstring("TORUS_FILLED\n");
  else if ( web.torus_flag ) outstring("TORUS\n");
  if ( web.torus_flag || web.inverse_periods )
     { outstring("PERIODS\n");
       for ( i = 0 ; i < SDIM ; i++ )
       { for ( j = 0 ; j < SDIM ; j++ )
         { outstring(" "); 
           outstring(print_express(&torus_period_expr[i][j],' '));
         }
         outstring("\n"); 
       }
       outstring("//Numerical values of periods:\n");
       for ( i = 0 ; i < SDIM ; i++ )
       { outstring("// ");
         for ( j = 0 ; j < SDIM ; j++ )
         { 
#ifdef LONGDOUBLE
           sprintf(msg," %*.*Lf ",DWIDTH,DPREC,web.torus_period[i][j]);
#else
           sprintf(msg," %18.15f ",web.torus_period[i][j]);
#endif 
           outstring(msg);
         }
         outstring("\n"); 
       }
       outstring("\n"); 
     }
  /* view matrix */
  outstring("VIEW_MATRIX \n");
  for ( i = 0 ; i <= SDIM ;i++ )
     { for ( j = 0 ; j <= SDIM ; j++ )
          { sprintf(msg,"%7.5f  ", (DOUBLE)view[i][j]); 
            outstring(msg); 
          }
       outstring("\n"); 
     }
  outstring("\n"); 
  if ( transform_gen_count > 0 )
     { int n = web.torus_flag?transform_gen_count-SDIM:transform_gen_count;
       if ( n > 0 )
        { sprintf(msg,"VIEW_TRANSFORM_GENERATORS %d \n",n); 
          outstring(msg); 
          for ( k = 0 ; k < n ; k++ )
          { if ( transform_gen_swap[k] ) 
                outstring("swap_colors\n");
            for ( i = 0 ; i <= SDIM ;i++ )
             { for ( j = 0 ; j <= SDIM ; j++ )
                { sprintf(msg,"%7.5g  ",(DOUBLE)view_transform_gens[k][i][j]); 
                  outstring(msg); 
                }
                outstring("\n"); 
             }
             outstring("\n"); 
          }
        }
     }
  else if ( transform_count > 0 )
     { { sprintf(msg,"VIEW_TRANSFORMS %d \n",transform_count); outstring(msg); }
        for ( k = 0 ; k < transform_count ; k++ )
        { if ( transform_colors[k] == SWAP_COLORS )
             outstring("swap_colors\n");
          else if ( transform_colors[k] != SAME_COLOR )
             { sprintf(msg,"color %s\n",COLORNAME(transform_colors[k]));
               outstring(msg);
             }
          for ( i = 0 ; i <= SDIM ;i++ )
          { for ( j = 0 ; j <= SDIM ; j++ )
                { sprintf(msg,"%7.5g  ",(DOUBLE)view_transforms[k][i][j]); outstring(msg); }
             outstring("\n"); 
          }
         outstring("\n"); 
        }
     }
  if ( web.meritfactor != 0.0 )
     { sprintf(msg,"MERIT_FACTOR: %2.15g\n\n",(DOUBLE)web.meritfactor); outstring(msg); }
  if ( web.gravflag )
#ifdef LONGDOUBLE
     { sprintf(msg,"GRAVITY_CONSTANT: %2.*Lg\n\n",DPREC,web.grav_const); outstring(msg); }
#else
     { sprintf(msg,"GRAVITY_CONSTANT: %2.15g\n\n",web.grav_const); outstring(msg); }
#endif 
  if ( web.diffusion_const != 0.0 )
#ifdef LONGDOUBLE
     { sprintf(msg,"DIFFUSION: %2.*Lg\n\n",DPREC,web.diffusion_const); outstring(msg); }
#else
     { sprintf(msg,"DIFFUSION: %2.15g\n\n",web.diffusion_const); outstring(msg); }
#endif 
  if ( mobility_flag )
     if ( mobility_tensor_flag )
     { outstring("MOBILITY_TENSOR\n"); 
        /* mobility tensor expressions */
          for ( i = 0 ; i < SDIM ; i++ )
             { for ( j = 0 ; j < SDIM ; j++ )
                  { outstring(" "); 
                    outstring(print_express(&mobility_tensor[i][j],'X'));
                  }
               outstring("\n"); 
             }
          outstring("\n"); 
     }
     else
      { 
         outstring("MOBILITY: ");
         outstring(print_express(&mobility_formula,'X'));
         outstring("\n\n"); 
      }
  if ( autochop_flag )
     { sprintf(msg,"AUTOCHOP %2.15g\n\n",(DOUBLE)autochop_size); outstring(msg); }
  if ( web.homothety )
     { sprintf(msg,"HOMOTHETY %2.15g\n\n",(DOUBLE)homothety_target); outstring(msg); }
  if ( autopop_flag )
     outstring("AUTOPOP\n\n");
  if ( effective_area_flag )
     outstring("EFFECTIVE_AREA\n\n"); 
  if ( boundary_curvature_flag )
     outstring("BOUNDARY_CURVATURE\n\n"); 
  if ( normal_curvature_flag )
     outstring("NORMAL_CURVATURE\n\n");
  if ( total_time > 0.0 )
     { sprintf(msg,"TOTAL_TIME %g\n\n",(DOUBLE)total_time); outstring(msg); }
  if ( runge_kutta_flag )
     outstring("RUNGE_KUTTA\n\n"); 
  if ( square_curvature_flag )
     { sprintf(msg,"SQUARE_CURVATURE: %2.15g\n\n",
            (DOUBLE)globals[square_curvature_param].value.real); outstring(msg); }
  if ( sqgauss_flag )
     { sprintf(msg,"SQUARE_GAUSSIAN_CURVATURE: %2.15g\n\n",
            (DOUBLE)globals[sqgauss_param].value.real); outstring(msg); }
  if ( mean_curv_int_flag )
     { sprintf(msg,"MEAN_CURVATURE_INTEGRAL: %2.15g\n\n",
            (DOUBLE)globals[square_curvature_param].value.real); outstring(msg); }
  if ( web.wulff_flag )
     { sprintf(msg,"WULFF:      \"%s\"\n\n",web.wulff_name); outstring(msg); }
  if ( phase_flag )
     { sprintf(msg,"PHASEFILE \"%s\"\n\n",phase_file_name); outstring(msg); }
  if ( web.spring_constant != 1.0 )
     { sprintf(msg,"GAP_CONSTANT: %2.15g\n\n",(DOUBLE)web.spring_constant); outstring(msg); }
  if ( web.motion_flag )
     { sprintf(msg,"SCALE: %2.15g     FIXED\n\n",(DOUBLE)web.scale); outstring(msg); }
  else if ( web.scale != 0.1 )
     { sprintf(msg,"SCALE: %2.15g\n\n",(DOUBLE)web.scale); outstring(msg); }
  if ( web.area_norm_flag )
     { sprintf(msg,"AREA_NORMALIZATION \n\n"); outstring(msg); }
  if ( web.jiggle_flag )
     { sprintf(msg,"JIGGLE\n\n"); outstring(msg); }
  if ( web.temperature != 0.05 )
     { sprintf(msg,"TEMPERATURE: %2.15g\n\n",(DOUBLE)web.temperature); outstring(msg); }
  if ( web.pressure != 0.0 )
     { sprintf(msg,"PRESSURE: %2.15g\n\n",(DOUBLE)web.pressure); outstring(msg); }
  if ( web.maxscale != 1.0 )
     { sprintf(msg,"SCALE_LIMIT: %2.15g\n\n",(DOUBLE)web.maxscale); outstring(msg); }
  if ( web.gauss1D_order != 3 )
     { sprintf(msg,"INTEGRAL_ORDER_1D: %d\n\n",web.gauss1D_order); outstring(msg); }
  if ( (web.dimension > 1) && (web.gauss2D_order != 6) )
     { sprintf(msg,"INTEGRAL_ORDER_2D: %d\n\n",web.gauss2D_order); outstring(msg); }
  if ( web.concount )
     { sprintf(msg,"CONSTRAINT_TOLERANCE: %2.15g\n\n",(DOUBLE)web.tolerance); outstring(msg); }
  if ( web.symmetric_content )
     { sprintf(msg,"SYMMETRIC_CONTENT\n\n"); outstring(msg); }
  if ( ordinal(web.zoom_v) != 0 ) 
     { sprintf(msg,"ZOOM_VERTEX  %d\n\n",ordinal(web.zoom_v)+1); outstring(msg); }
  if ( web.zoom_radius < 9000.0 )
     { sprintf(msg,"ZOOM_RADIUS  %2.15g\n\n",(DOUBLE)web.zoom_radius); outstring(msg); }

  if ( klein_metric_flag ) { sprintf(msg,"KLEIN_METRIC\n\n"); outstring(msg); }
  if ( web.conformal_flag )
    {   sprintf(msg,"CONFORMAL_METRIC\n"); outstring(msg); 
        /* metric expression */
        outstring(print_express(&web.metric[0][0],'X'));
        outstring("\n\n"); 
     }
  else if ( web.metric_flag && web.metric[0][0].root )
     {  sprintf(msg,"METRIC\n"); outstring(msg); 
        /* metric expressions */
        for ( i = 0 ; i < SDIM ; i++ )
             { for ( j = 0 ; j < SDIM ; j++ )
                  { outstring(" "); 
                    outstring(print_express(&web.metric[i][j],'X'));
                  }
                outstring("\n"); 
             }
        outstring("\n"); 
     }

  /* extra attributes for elements */
  for ( e_type = VERTEX ; e_type <= FACETEDGE ; e_type++ )
  { struct extra *ex;
    for ( i = 0, ex = web.skel[e_type].extras ; 
                i < web.skel[e_type].extra_count ; i++, ex++ )
     { if ( !(ex->flags & DUMP_ATTR) ) continue;
        sprintf(msg,"\ndefine %s attribute %s %s\n",typenames[e_type], ex->name,
          attr_type_name[ex->type]);
        if ( ex->dim > 1 )
          sprintf(msg+strlen(msg)-1,"[%d]\n",ex->dim);
        if ( ex->code.start ) 
          sprintf(msg+strlen(msg)-1,
            " // function definition in READ section\n");
        outstring(msg);
     }
  }

  /* quantity integrand info */
  for ( i = 0 ; i < QUANTMAX ; i++ )
     { struct quantity *quan = web.quants + i;

        if ( quan->quanvect[0] == NULL ) continue;
        { sprintf(msg,"\nQUANTITY %d  ",i); outstring(msg); }
        if ( quan->attr & GLOBAL        ) outstring("  GLOBAL");
        if ( quan->attr & QFIXED        )
#ifdef LONGDOUBLE
            { sprintf(msg,"  FIXED = %2.*Lg ",DPREC,quan->target); 
              outstring(msg); }
#else
            { sprintf(msg,"  FIXED = %2.15g ",quan->target); 
              outstring(msg); }
#endif 
        outstring("\n"); 
        for ( j = 0 ; j < SDIM ; j++ )
                { sprintf(msg,"Q%1d: ",j+1); outstring(msg); 
                  outstring(print_express(quan->quanvect[j],'X'));
                  outstring("\n"); 
                }
      } 
  
  /* unattached instances, probably from compound quantities */
  for ( mi = METH_INST, k = 0  ; k < meth_inst_count  ; mi++,k++ )
    { if ( mi->quant >= 0 ) continue; /* has a quantity */ 
      outstring("METHOD_INSTANCE "); 
      outstring(mi->name);
      outstring(" METHOD ");
      outstring(basic_gen_methods[mi->gen_method].name);
      dump_method_specs(k);
    }

  /* named quantities */
  for ( q = GEN_QUANTS, k = 0 ; k < gen_quant_count ; q++,k++ )
     { 
        if ( q->flags & DEFAULT_QUANTITY ) continue;

        /* first, its methods */
        outstring("\n");
        for ( i = 0 ; i < q->method_count ; i++ )
        { mi = METH_INST + q->meth_inst[i]; 
          if ( mi->flags & (IMPLICIT_INSTANCE|DEFAULT_INSTANCE) ) continue;
          outstring("METHOD_INSTANCE "); 
          outstring(mi->name);
          outstring(" METHOD ");
          outstring(basic_gen_methods[mi->gen_method].name);
          dump_method_specs(q->meth_inst[i]);
        }
        switch ( q->flags & (Q_INFO|Q_FIXED|Q_ENERGY) )
         { case Q_INFO:
              sprintf(msg,"QUANTITY %s INFO_ONLY ",q->name); outstring(msg); 
              break;
            case Q_FIXED:
#ifdef LONGDOUBLE
              sprintf(msg,"QUANTITY %s FIXED = %2.*Lg  lagrange_multiplier %2.*Lg",
                  q->name,DPREC,q->target,DPREC,q->pressure); 
#else
              sprintf(msg,"QUANTITY %s FIXED = %2.15g  lagrange_multiplier %2.15g",
                  q->name,q->target,q->pressure); 
#endif 
              outstring(msg); 
              break;
            case Q_ENERGY:
                sprintf(msg,"QUANTITY %s ENERGY ",q->name); outstring(msg); 
              break;
         }
        if ( q->tolerance >= 0.0 ) 
#ifdef LONGDOUBLE
          { sprintf(msg," tolerance %2.*Lg ",DPREC,q->tolerance); outstring(msg); }
#else
          { sprintf(msg," tolerance %2.15g ",q->tolerance); outstring(msg); }
#endif 
        if ( q->modulus != 1.0 ) 
#ifdef LONGDOUBLE
          { sprintf(msg," modulus %2.*Lg ",DPREC,q->modulus); outstring(msg); }
#else
          { sprintf(msg," modulus %2.15g ",q->modulus); outstring(msg); }
#endif 
        if ( q->volconst != 0.0 ) 
#ifdef LONGDOUBLE
          { sprintf(msg," volconst %2.*Lg ",DPREC,q->volconst); outstring(msg); }
#else
          { sprintf(msg," volconst %2.15g ",q->volconst); outstring(msg); }
#endif 
        if ( q->flags & Q_COMPOUND )
         { outstring(" function ");
           outstring(print_express(&q->expr,'\0'));
         }
        else for ( i = 0 ; i < q->method_count ; i++ )
         { mi = METH_INST + q->meth_inst[i]; 
            outstring(" method ");
            if ( mi->flags & (IMPLICIT_INSTANCE|DEFAULT_INSTANCE) )
              { outstring(basic_gen_methods[mi->gen_method].name);
                dump_method_specs(q->meth_inst[i]);
              }
            else outstring(mi->name);
         }
        outstring("\n"); 
     }


  /* print boundary info */
  for ( i = 0 ; i < BDRYMAX ; i++ )
     { struct boundary * bdry = web.boundaries + i;
       if ( bdry->pcount == 0 ) continue;
         sprintf(msg,"\nBOUNDARY %d  PARAMETERS %d",i,bdry->pcount); 
         outstring(msg); 
       if ( bdry->attr & B_CONVEX ) outstring("  CONVEX");
       outstring("\n"); 
       for ( j = 0 ; j < SDIM ; j++ )
          {
             sprintf(msg,"X%1d: ",j+1); outstring(msg); 
             outstring(print_express(bdry->coordf[j],'P'));
             outstring("\n");
          }

       if ( (bdry->attr & CON_ENERGY)  )
          { outstring("ENERGY \n");
             for ( j = 0 ; j < bdry->compcount ; j++ )
                { sprintf(msg,"E%1d: ",j+1); outstring(msg); 
                  outstring(print_express(bdry->envect[j],'X'));
                  outstring("\n"); 
                }
          }
       if ( bdry->attr & CON_CONTENT )
          {
             outstring("CONTENT \n");
             for ( j = 0 ; j < bdry->compcount ; j++ )
                { sprintf(msg,"C%1d: ",j+1); outstring(msg); 
                  outstring(print_express(bdry->convect[j],'X'));
                  outstring("\n"); 
                }
          }

       outstring("\n");
     }

  /* print constraint info */
  for ( i = 0 ; i < MAXCON ; i++ )
     { struct constraint *con = get_constraint(i);

        if ( !web.constraint_addr[i] ) continue;
        sprintf(msg,"\nCONSTRAINT %d  ",i); outstring(msg); 
        if ( con->attr & B_CONVEX )  outstring("  CONVEX"); 
        if ( con->attr & NONNEGATIVE )  outstring("  NONNEGATIVE"); 
        if ( con->attr & NONPOSITIVE )  outstring("  NONPOSITIVE"); 
        if ( con->attr & GLOBAL        )  outstring("  GLOBAL"); 

        outstring("\nFUNCTION:  ");
        outstring(print_express(con->formula,'X'));
        outstring("\n"); 

        if ( (con->attr & CON_ENERGY)  )
          { outstring("ENERGY \n");
             for ( j = 0 ; j < con->compcount ; j++ )
                { sprintf(msg,"E%1d: ",j+1); outstring(msg); 
                  outstring(print_express(con->envect[j],'X'));
                  outstring("\n"); 
                }
          }
        if ( con->attr & CON_CONTENT )
          {
             outstring("CONTENT \n");
             for ( j = 0 ; j < con->compcount ; j++ )
                { sprintf(msg,"C%1d: ",j+1); outstring(msg); 
                  outstring(print_express(con->convect[j],'X'));
                  outstring("\n"); 
                }
          }

        /* quantity integrand info */
        for ( k = 0 ; k < QUANTMAX ; k++ )
          {
             if ( !(con->quantity_map & (1<<k)) ) continue;
             sprintf(msg,"\nQUANTITY %d  \n",k); outstring(msg); 
             for ( j = 0 ; j < con->compcount ; j++ )
                { sprintf(msg,"Q%1d: ",j+1); outstring(msg); 
                  outstring(print_express(con->quanvect[k][j],'X')); 
                  outstring("\n"); 
                }
            } 
      } 

  /* surface energy integrand info */
  for ( i = 0 ; i < SURFENMAX ; i++ )
     { struct surf_energy *surfen = web.surfen + i;

        if ( surfen->envect[0] == NULL ) continue;
        sprintf(msg,"\nSURFACE_ENERGY %d  ",i); outstring(msg); 
        if ( surfen->attr & GLOBAL        ) outstring("  GLOBAL");
        outstring("\n"); 
        for ( j = 0 ; j < SDIM ; j++ )
                { sprintf(msg,"E%1d: ",j+1); outstring(msg); 
                  outstring(print_express(surfen->envect[j],'X'));
                  outstring("\n"); 
                }
      } 

  /* windup */
  outfd = old_fd;
} /* end top_dump() */

/**********************************************************************
*
* function: dump_macros()
*
* purpose: print macros in dumpfile if keep_macros_flag
*
*/

void dump_macros()
{ int n;

  if ( !keep_macros_flag ) return;
  outstring("\n");
  for ( n = 0 ; n < macro_count ; n++ )
  { sprintf(msg,"#define %s ",macros[n].name);
     outstring(msg);
     outstring(macro_subs+macros[n].offset);
     outstring("\n");
  }
  outstring("\n");

}

/**********************************************************************
*
*  function: dump_method_specs()
*
*  purpose: list attributes of method instance
*/

void dump_method_specs(meth)
int meth;
{
  struct method_instance *mi = METH_INST + meth;
  int j;

  if ( mi->flags & GLOBAL_INST ) outstring(" global ");
  if ( mi->flags & IGNORE_CONSTR ) outstring(" ignore_constraints ");
  if ( mi->flags & IGNORE_FIXED ) outstring(" ignore_fixed ");
  if ( mi->modulus != 1.0 )
     { 
#ifdef LONGDOUBLE
        sprintf(msg," modulus %3.*Lg ",DPREC,mi->modulus);
#else
        sprintf(msg," modulus %3.15g ",mi->modulus);
#endif 
        outstring(msg);
     }
  if ( mi->flags & METH_PARAMETER_1 )
  { 
#ifdef LONGDOUBLE
     sprintf(msg," parameter_1 %3.*Lg ",DPREC,mi->parameter_1);
#else
     sprintf(msg," parameter_1 %3.15g ",mi->parameter_1);
#endif 
     outstring(msg);
  }
  outstring("\n");
  if ( basic_gen_methods[mi->gen_method].spec_flags & SPEC_SCALAR )
    { outstring("scalar_integrand: ");
      outstring(print_express(mi->expr[0],'X'));
      outstring("\n");
    }
  if ( basic_gen_methods[mi->gen_method].spec_flags & SPEC_VECTOR )
    { outstring("vector_integrand: \n");
      for ( j = 0 ; j < SDIM ; j++ )
        { sprintf(msg,"Q%1d: ",j+1); outstring(msg); 
          outstring(print_express(mi->expr[j],'X'));
          outstring("\n");
        }
    }
  if ( basic_gen_methods[mi->gen_method].spec_flags & SPEC_KVECTOR )
    { sprintf(msg,"k_vector_order %d\nvector_integrand: \n",mi->vec_order);
      outstring(msg);
      for ( j = 0 ; j < mi->vec_order*SDIM ; j++ )
        { sprintf(msg,"Q%1d: ",j+1); outstring(msg); 
          outstring(print_express(mi->expr[j],'X'));
          outstring("\n");
        }
    }
  if ( basic_gen_methods[mi->gen_method].spec_flags & SPEC_2FORM )
    { outstring("form_integrand: \n");
      for ( j = 0 ; j < (SDIM*(SDIM-1))/2 ; j++ )
        { sprintf(msg,"Q%1d: ",j+1); outstring(msg); 
          outstring(print_express(mi->expr[j],'X'));
          outstring("\n");
        }
    }
} /* end dump_method_specs() */

/*************************************************************************
*
* function: vertex_dump()
*
* purpose: write one vertex definition to datafile
*
*/

void vertex_dump(v_id,fd)
vertex_id v_id;
FILE *fd;   /* destination file */
{ int i,k;
  REAL *x;
  FILE *old_fd = outfd;
  struct extra *ex;

  outfd = fd;
        
  if ( !valid_id(v_id) ) return;
  x = get_coord(v_id);
  if ( get_vattr(v_id) & BOUNDARY )
  { /* print boundary parameters */
    struct boundary *bdry = get_boundary(v_id);
    REAL *param = get_param(v_id);

    sprintf(msg,"%3d",1+ordinal(v_id)); outstring(msg); 
    for ( i = 0 ; i < bdry->pcount ; i++ )
#ifdef LONGDOUBLE
    { sprintf(msg,"  %*.*Lg",DWIDTH,DPREC,param[i]); outstring(msg); }
#else
    { sprintf(msg,"  %17.15g",param[i]); outstring(msg); }
#endif 
    sprintf(msg,"  boundary %d ",bdry-web.boundaries); outstring(msg); 
    sprintf(msg," /* ("); outstring(msg); 
    for ( i = 0; i < SDIM ; i++)
#ifdef LONGDOUBLE
    sprintf(msg," %*.*Lg",DWIDTH,DPREC,x[i]); outstring(msg); 
#else
    sprintf(msg," %17.15g",x[i]); outstring(msg); 
#endif 
    sprintf(msg,") */"); outstring(msg); 
  } /* end BOUNDARY parameters */
  else
  { /* print regular coordinates */
    sprintf(msg,"%3d ",1+ordinal(v_id)); outstring(msg); 
    for ( i = 0; i < SDIM ; i++)
#ifdef LONGDOUBLE
      { sprintf(msg," %*.*Lg",DWIDTH,DPREC,x[i]); outstring(msg); }
#else
      { sprintf(msg," %17.15g",x[i]); outstring(msg); }
#endif 
  }
  if ( get_vattr(v_id) & CONSTRAINT )
  { conmap_t * conmap = get_v_constraint_map(v_id);
    sprintf(msg,"  constraints "); outstring(msg); 
    for ( i = 1 ; i <= (int)conmap[0] ; i++ )
      if ( !(get_constraint(conmap[i])->attr & GLOBAL) )
         { sprintf(msg,"%d ",conmap[i]&CONMASK); outstring(msg); }
  }
  if ( get_vattr(v_id)&FIXED ) { sprintf(msg, " fixed "); outstring(msg); }
  if ( get_vattr(v_id)&AXIAL_POINT ) 
     { sprintf(msg, " axial_point "); outstring(msg); }
  if ( get_vattr(v_id)&BARE_NAKED )
     { sprintf(msg, " bare "); outstring(msg); }

  i = get_original(v_id);
  if ( (i!=NO_ORIGINAL) && (i != 1+ordinal(v_id)) ) 
  { sprintf(msg," original %d",i); outstring(msg); }

  if ( elptr(v_id)->method_count )
  { struct method_instance *mi;
    int meth_offset = web.skel[VERTEX].extras[web.meth_attr[VERTEX]].offset;
    int *methlist = (int*)((char*)elptr(v_id) + meth_offset);
    for ( i = 0 ; i < (int)elptr(v_id)->method_count ; i++ )
      { mi = METH_INST + methlist[i];
        if ( GEN_QUANTS[mi->quant].flags & DEFAULT_QUANTITY ) continue;
        if ( mi->flags & IMPLICIT_INSTANCE )
           sprintf(msg," %s ",GEN_QUANTS[mi->quant].name);
        else sprintf(msg," %s ",mi->name); 
        outstring(msg); 
      }
  }

  /* extra attributes */
  for ( i = 0, ex = web.skel[VERTEX].extras ; i < web.skel[VERTEX].extra_count;
              i++,ex++ )
    { char *at = get_v_extra(v_id,i);
      if ( !(ex->flags & DUMP_ATTR) ) continue;
      if ( ex->dim == 0 ) continue;
      sprintf(msg," %s ",ex->name);
      for ( k = 0 ; k < ex->dim ; k++ )
         switch ( ex->type )
         { case REAL_ATTR: 
#ifdef LONGDOUBLE
                sprintf(msg+strlen(msg),"%2.*Lg ",DPREC,((REAL*)at)[k]); 
#else
                sprintf(msg+strlen(msg),"%2.15g ",((REAL*)at)[k]); 
#endif 
                break;
           case INTEGER_ATTR: sprintf(msg+strlen(msg),"%d ",((int *)at)[k]); 
                break;
           case ULONG_ATTR: sprintf(msg+strlen(msg),"%ld ",
                ((unsigned long *)at)[k]); 
                break;
         }
      outstring(msg);
    }

  sprintf(msg,"\n"); outstring(msg); 
  outfd = old_fd;
} /* end vertex_dump() */

/***************************************************************************
*
* function: edge_dump()
*
* purpose: write out one edge definition
*
*/

void edge_dump(e_id,fd)
edge_id e_id;
FILE *fd;  /* destination file */
{ int i,k;
  MAP surfenmap;
  MAP quantmap;
  int attr;
  FILE *old_fd = outfd;
  struct extra *ex;

  outfd = fd;

  if ( !valid_id(e_id) ) return;
  attr = get_eattr(e_id);

  sprintf(msg,"%3d     ",1+ordinal(e_id)); outstring(msg); 
  if ( web.modeltype == LAGRANGE )
  {
      vertex_id *v = get_edge_vertices(e_id);
      int num = web.skel[EDGE].ctrlpts;
      for ( i = 0 ; i < num ; i++ )
        { sprintf(msg,"%3d  ",ordinal(v[i])+1); outstring(msg); }
  }
  else
  {
    if ( web.representation == SIMPLEX )
     { vertex_id *v = get_edge_vertices(e_id);
        for ( i = 0 ; i <= web.dimension-1 ; i++ )
          { sprintf(msg,"%3d  ",ordinal(v[i])+1); outstring(msg); }
     }
    else
    { sprintf(msg,"%3d  %3d    ", 1+ordinal(get_edge_tailv(e_id)),
                1+ordinal(get_edge_headv(e_id))); 
      outstring(msg); 
    }
    if ( web.modeltype == QUADRATIC )
    { sprintf(msg,"  /*midpt*/ %3d ",1+ordinal(get_edge_midv(e_id))); 
      outstring(msg); 
    }
  }

  if ( web.torus_flag )
  { WRAPTYPE wrap = get_edge_wrap(e_id);
    for ( i = 0 ; i < SDIM ; i++, wrap >>= TWRAPBITS )
      switch ( wrap & WRAPMASK  )
      {
          case  NEGWRAP : sprintf(msg," -"); outstring(msg);   break;
          case  0       : sprintf(msg," *"); outstring(msg);   break;
          case  POSWRAP : sprintf(msg," +"); outstring(msg);   break;
          default : 
              sprintf(errmsg,"Bad wrap %lX on edge %d period %d\n",
                  wrap&WRAPMASK,ordinal(e_id)+1,i+1);
              kb_error(1005,errmsg,WARNING);
              sprintf(msg," bad"); outstring(msg); 
            break;
      }
  }
  else if ( web.symmetry_flag )
  { WRAPTYPE wrap = get_edge_wrap(e_id);
    sprintf(msg," wrap %ld ",wrap); outstring(msg); 
  }
  if ( (attr & DENSITY) || ((web.representation==STRING) &&
                    (get_edge_density(e_id) != 1.0)))
#ifdef LONGDOUBLE
    { sprintf(msg,"  density %1.*Lg ",DPREC,get_edge_density(e_id)); outstring(msg); }
#else
    { sprintf(msg,"  density %1.15g ",get_edge_density(e_id)); outstring(msg); }
#endif 
  if ( attr & BOUNDARY )
     { struct boundary *bdry = get_edge_boundary(e_id);
       sprintf(msg,"  boundary %d ",bdry-web.boundaries); outstring(msg); 
     }                    
  if ( attr & CONSTRAINT )
  { conmap_t * conmap = get_e_constraint_map(e_id);
    sprintf(msg,"  constraints "); outstring(msg); 
    for ( i = 1 ; i <= (int)conmap[0] ; i++ )
    { sprintf(msg,"%d ",(conmap[i]&CONMASK)); outstring(msg); }
  }

  if ( attr & NEGBOUNDARY ) outstring(" orientation -1 ");
  if ( attr & FIXED )
      { sprintf(msg,"fixed "); outstring(msg); }
  if ( attr & BARE_NAKED ) { sprintf(msg," bare "); outstring(msg); }
  if ( attr & NO_REFINE ) outstring(" no_refine ");

  if ( get_edge_color(e_id)  != DEFAULT_EDGE_COLOR )
  { sprintf(msg," color %s ",COLORNAME(get_edge_color(e_id))); outstring(msg); }
  if ( (fd == old_fd) && (web.representation != SIMPLEX) )
  { calc_edge(e_id);
    sprintf(msg," /*length %g*/ ",(DOUBLE)get_edge_length(e_id)); outstring(msg); 
  }
  if ( elptr(e_id)->method_count )
  { struct method_instance *mi;
    int meth_offset = web.skel[EDGE].extras[web.meth_attr[EDGE]].offset;
    int *methlist = (int*)((char*)elptr(e_id) + meth_offset);
    for ( i = 0 ; i < elptr(e_id)->method_count ; i++ )
      { int mm = methlist[i];
        mi = METH_INST + abs(mm);
        if ( GEN_QUANTS[mi->quant].flags & DEFAULT_QUANTITY ) continue;
        if ( mi->flags & IMPLICIT_INSTANCE )
           sprintf(msg,"%s%c ",GEN_QUANTS[mi->quant].name,(mm<0?'-':' '));
        else sprintf(msg,"%s%c ",mi->name,(mm<0?'-':' ')); 
        outstring(msg); 
      }
  }
  if ( get_eattr(e_id) & SURF_ENERGY )
  { surfenmap = get_e_surfen_map(e_id) & ~web.surf_global_map;
     if ( surfenmap )
     {  sprintf(msg,"  energy "); outstring(msg);
         for ( i = 0 ; i < web.surfen_count ; i++,surfenmap >>= 1 )
            if ( surfenmap & 1 )
               { sprintf(msg,"%d ",i); outstring(msg); }
     }
  }
  if ( get_eattr(e_id) & SURF_QUANTITY )
  { quantmap = get_e_quant_map(e_id) & ~web.quant_global_map;
     if ( quantmap )
     {  sprintf(msg," quantity "); outstring(msg); 
        for ( i = 0 ; i < web.quantity_count ; i++,quantmap >>= 1 )
          if ( quantmap & 1 )
          { sprintf(msg,"%d ",i); outstring(msg); }
     }
  }
  i = get_original(e_id);
  if ( (i!=NO_ORIGINAL) && (i != 1+ordinal(e_id)) ) 
     { sprintf(msg," original %d",i); outstring(msg); }

  /* extra attributes */
  for ( i = 0, ex = web.skel[EDGE].extras ; i < web.skel[EDGE].extra_count;
              i++,ex++ )
    { char *at = get_e_extra(e_id,i);
      if ( !(ex->flags & DUMP_ATTR) ) continue;
      if ( ex->dim == 0 ) continue;
      sprintf(msg," %s ",ex->name);
      for ( k = 0 ; k < ex->dim ; k++ )
         switch ( ex->type )
         { case REAL_ATTR: 
#ifdef LONGDOUBLE
                sprintf(msg+strlen(msg),"%2.*Lg ",DPREC,((REAL*)at)[k]); 
#else
                sprintf(msg+strlen(msg),"%2.15g ",((REAL*)at)[k]); 
#endif 
                break;
            case INTEGER_ATTR: sprintf(msg+strlen(msg),"%d ",((int *)at)[k]); 
                break;
            case ULONG_ATTR: sprintf(msg+strlen(msg),"%ld ",((unsigned long *)at)[k]); 
                break;
         }
      outstring(msg);
    }

  outstring("\n"); 

  outfd = old_fd;
}  /* end edge_dump() */

/***************************************************************************
*
* function: facet_dump()
*
* purpose: write out one facet definition
*
*/

void facet_dump(f_id,fd)
facet_id f_id;
FILE *fd;  /* destination file */
{ int i,k;
  int per_line = 0; 
  MAP surfenmap;
  MAP quantmap;
  edge_id e_id;
  facetedge_id fe;
  FILE *old_fd = outfd;
  struct extra *ex;
  int attr = get_fattr(f_id);

  outfd = fd;

  if ( !valid_id(f_id) ) return;
  { sprintf(msg,"%3d  ",1+ordinal(f_id)); outstring(msg); }
    if ( web.representation == SIMPLEX )
    { vertex_id *v = get_facet_vertices(f_id);
      int num;
      if (web.modeltype == LAGRANGE)
         num = binom_coeff(web.lagrange_order+web.dimension,web.dimension);
      else num = web.dimension+1;
      for ( i = 0 ; i < num ; i++ )
         { sprintf(msg," %3d",ordinal(v[i])+1); outstring(msg); }
    }
    else
    {
      generate_facet_fe_init();
      while ( generate_facet_fe(f_id,&fe) ) 
      { e_id = get_fe_edge(fe);
        sprintf(msg," %3d",inverted(e_id) ? -1-ordinal(e_id) : 
             1+ordinal(e_id) );
        outstring(msg); 
        per_line++;
        if ( per_line >= 10 )
           { sprintf(msg," \\\n              "); outstring(msg); 
             per_line = 0;
           }
      }
      if ( (web.modeltype == LAGRANGE) && (web.representation != STRING) )
          { int num = binom_coeff(web.lagrange_order+web.dimension,web.dimension);
             vertex_id *v = get_facet_vertices(f_id);
             outstring(" vertices ");
             for ( i = 0 ; i < num ; i++ )
              { sprintf(msg,"%3d ",ordinal(v[i])+1); outstring(msg); }
          }
    }
    if ( attr & BOUNDARY )
      { struct boundary *bdry = get_facet_boundary(f_id);
        sprintf(msg,"  boundary %d ",bdry-web.boundaries); outstring(msg); 
      }                    
    if ( attr & CONSTRAINT )
      { conmap_t * conmap = get_f_constraint_map(f_id);
        outstring(" constraints ");
        for ( i = 1 ; i <= (int)conmap[0] ; i++ )
        { sprintf(msg,"%d ",conmap[i]&CONMASK); outstring(msg); }
      }
    if ( attr & NEGBOUNDARY ) outstring(" orientation -1 ");
    if ( attr & SURF_ENERGY )
    { surfenmap = get_f_surfen_map(f_id) & ~web.surf_global_map;
      if ( surfenmap )
      { 
         outstring(" energy ");
         for ( i = 0 ; i < web.surfen_count ; i++,surfenmap >>= 1 )
            if ( surfenmap & 1 )
              { sprintf(msg,"%d ",i); outstring(msg); }
      }
    }
    if ( attr & SURF_QUANTITY )
    { quantmap = get_f_quant_map(f_id) & ~web.quant_global_map;
      if ( quantmap )
      { 
         outstring(" quantity ");
         for ( i = 0 ; i < web.quantity_count ; i++,quantmap >>= 1 )
            if ( quantmap & 1 )
              { sprintf(msg,"%d ",i); outstring(msg); }
      }
    }
    if ( attr & DENSITY )
#ifdef LONGDOUBLE
    { sprintf(msg,"  density %1.*Lg ",DPREC,get_facet_density(f_id)); outstring(msg); }
#else
    { sprintf(msg,"  density %1.15g ",get_facet_density(f_id)); outstring(msg); }
#endif 
    if ( attr&FIXED )  outstring(" fixed " ); 
    if ( attr & NO_REFINE ) outstring(" no_refine ");
    if ( attr & NODISPLAY ) outstring(" nodisplay"); 
    if ( get_tag(f_id) ) 
        { sprintf(msg," tag %d ",get_tag(f_id)); outstring(msg); }
    if ( phase_flag && (web.representation == STRING) )
     { sprintf(msg," phase %d ",get_f_phase(f_id)); outstring(msg); }
    if ( get_facet_color(f_id) != DEFAULT_FACET_COLOR  )
     { sprintf(msg," color %s ",COLORNAME(get_facet_color(f_id)));
       outstring(msg); }
    if ( get_facet_backcolor(f_id) !=  get_facet_color(f_id) )
     { sprintf(msg," backcolor %s ",COLORNAME(get_facet_backcolor(f_id))); 
       outstring(msg); }
    if ( elptr(f_id)->method_count )
    { struct method_instance *mi;
      int meth_offset = web.skel[FACET].extras[web.meth_attr[FACET]].offset;
      int *methlist = (int*)((char*)elptr(f_id) + meth_offset);
      for ( i = 0 ; i < (int)elptr(f_id)->method_count ; i++ )
        { int mm = methlist[i];
          mi = METH_INST + abs(mm);
          if ( GEN_QUANTS[mi->quant].flags & DEFAULT_QUANTITY ) continue;
          if ( mi->flags & IMPLICIT_INSTANCE )
             sprintf(msg," %s%c ",GEN_QUANTS[mi->quant].name,(mm<0?'-':' '));
          else sprintf(msg," %s%c ",mi->name,(mm<0?'-':' ')); 
          outstring(msg); 
        }
    }
    sprintf(msg," /*area %g*/", (DOUBLE)get_facet_area(f_id)); outstring(msg); 
    i = get_original(f_id);
    if ( (i!=NO_ORIGINAL) && (i != 1+ordinal(f_id)) )
        { sprintf(msg," original %d",i); outstring(msg); }

  /* extra attributes */
  for ( i = 0, ex = web.skel[FACET].extras ; i < web.skel[FACET].extra_count;
              i++,ex++ )
    { char *at = get_f_extra(f_id,i);
      if ( !(ex->flags & DUMP_ATTR) ) continue;
      if ( ex->dim == 0 ) continue;
      sprintf(msg," %s ",ex->name);
      for ( k = 0 ; k < ex->dim ; k++ )
         switch ( ex->type )
         { case REAL_ATTR: 
#ifdef LONGDOUBLE
                sprintf(msg+strlen(msg),"%2.*Lg ",DPREC,((REAL*)at)[k]); 
#else
                sprintf(msg+strlen(msg),"%2.15g ",((REAL*)at)[k]); 
#endif 
                break;
           case INTEGER_ATTR: sprintf(msg+strlen(msg),"%d ",((int *)at)[k]); 
                break;
           case ULONG_ATTR: sprintf(msg+strlen(msg),"%ld ",((unsigned long *)at)[k]); 
                break;
         }
      outstring(msg);
    }

  outstring("\n"); 

  outfd = old_fd;
}  /* end facet_dump() */

/*************************************************************************
*
* function: body_dump()
*
* purpose: Write out definition of one body.
*
*/

void body_dump(b_id,fd)
body_id b_id;
FILE *fd;  /* destination file */
{ 
  REAL den;
  int per_line = 0;
  facet_id f_id;
  int i,k;
  FILE *old_fd = outfd;
  struct extra *ex;
  REAL bvol;
  outfd = fd;

  if ( !valid_id(b_id) ) return;
  { sprintf(msg,"%3d      ",1+ordinal(b_id)); outstring(msg); }
    FOR_ALL_FACETS(f_id)
    { if ( equal_id(get_facet_body(f_id),b_id) )
      { sprintf(msg," %3d",1+ordinal(f_id)); outstring(msg); 
        per_line++;
      }
      if ( equal_id(get_facet_body(facet_inverse(f_id)),b_id) )
      { sprintf(msg," %3d",-1-ordinal(f_id)); outstring(msg); 
        per_line++;
      }
      if ( per_line >= 10 )
      { sprintf(msg," \\\n          "); outstring(msg); 
        per_line = 0;
      }
    }
    bvol = get_body_volume(b_id);
    if ( get_battr(b_id) & FIXEDVOL )
    { sprintf(msg,
#ifdef LONGDOUBLE
        "  volume %1.*Lg  /*actual: %1.*Lg*/ lagrange_multiplier %1.*Lg ",
                     DPREC,get_body_fixvol(b_id), DPREC,
                     bvol,DPREC,get_body_pressure(b_id));
#else
        "  volume %1.15g  /*actual: %1.15g*/ lagrange_multiplier %1.15g ",
                    get_body_fixvol(b_id), 
                    bvol,get_body_pressure(b_id));
#endif 
      outstring(msg); 
    }
    if ( (get_body_volconst(b_id) != 0.0) )
#ifdef LONGDOUBLE
      { sprintf(msg, "  volconst %1.*Lg ",DPREC,get_body_volconst(b_id)); 
#else
      { sprintf(msg, "  volconst %1.15g ",get_body_volconst(b_id)); 
#endif 
        outstring(msg); }

    if ( web.torus_flag && ((bvol > web.torusv) || (bvol < 0.0)) )
#ifdef LONGDOUBLE
    { sprintf(msg, "  actual_volume %1.*Lg ",DPREC,bvol); 
#else
    { sprintf(msg, "  actual_volume %1.15g ",bvol); 
#endif 
      outstring(msg); }

    if ( get_battr(b_id) & PRESSURE ) 
#ifdef LONGDOUBLE
      { sprintf(msg,"  pressure %1.*Lg",DPREC, get_body_pressure(b_id)); outstring(msg); }
#else
      { sprintf(msg,"  pressure %1.15g", get_body_pressure(b_id)); outstring(msg); }
#endif 
    den = get_body_density(b_id);
    if ( den != 0.0 )
#ifdef LONGDOUBLE
         { sprintf(msg,"  density %1.*Lg ",DPREC,den); outstring(msg); }
#else
         { sprintf(msg,"  density %1.15g ",den); outstring(msg); }
#endif 
    if ( phase_flag && (web.representation == SOAPFILM) )
         { sprintf(msg," PHASE %d ",get_b_phase(b_id)); outstring(msg); }
    i = get_original(b_id);
    if ( (i!=NO_ORIGINAL) && (i != 1+ordinal(b_id)) ) 
    { sprintf(msg," original %d",i); 
    outstring(msg); }
    if ( elptr(b_id)->method_count )
      { struct method_instance *mi;
        int meth_offset = web.skel[BODY].extras[web.meth_attr[BODY]].offset;
        int *methlist = (int*)((char*)elptr(b_id) + meth_offset);
        for ( i = 0 ; i < (int)elptr(b_id)->method_count ; i++ )
              { mi = METH_INST + methlist[i];
                 if ( mi->flags & IMPLICIT_INSTANCE )
                    sprintf(msg,"%s ",GEN_QUANTS[mi->quant].name);
                 else sprintf(msg,"%s ",mi->name); 
                 outstring(msg); 
              }
      }

  /* extra attributes */
  for ( i = 0, ex = web.skel[BODY].extras ; i < web.skel[BODY].extra_count;
              i++,ex++ )
    { char *at = get_b_extra(b_id,i);
      if ( !(ex->flags & DUMP_ATTR) ) continue;
      if ( ex->dim == 0 ) continue;
      sprintf(msg," %s ",ex->name);
      for ( k = 0 ; k < ex->dim ; k++ )
         switch ( ex->type )
         { case REAL_ATTR: 
#ifdef LONGDOUBLE
                sprintf(msg+strlen(msg),"%2.*Lg ",DPREC,((REAL*)at)[k]); 
#else
                sprintf(msg+strlen(msg),"%2.15g ",((REAL*)at)[k]); 
#endif 
                break;
           case INTEGER_ATTR: sprintf(msg+strlen(msg),"%d ",((int *)at)[k]); 
                break;
           case ULONG_ATTR: sprintf(msg+strlen(msg),"%ld ",((unsigned long *)at)[k]); 
                break;
         }
      outstring(msg);
    }

  outstring("\n"); 

  outfd = old_fd;
} /* end body_dump */

static int readflag; /* whether READ section has been started */
static FILE *dumpfd;

/**************************************************************************
*
* function: toggle_save()
*
* purpose: write ON toggle state to standard output (which has been 
*          redirected for dump purposes)
*
*/

void toggle_save(togglename)
char *togglename;
{ if ( readflag == 0 ) { outstring("\nread\n"); readflag = 1; }
  sprintf(msg,"%s on\n",togglename);
  outstring(msg);
}

/**************************************************************************
*
* function: toggle_save_off()
*
* purpose: write OFF toggle state to standard output (which has been 
*          redirected for dump purposes)
*
*/

void toggle_save_off(togglename) /* for default on toggles */
char *togglename;
{ if ( readflag == 0 ) { outstring("\nread\n"); readflag = 1; }
  sprintf(msg,"%s off\n",togglename);
  outstring(msg);
}

/***********************************************************************
*
* function: do_dump()
*
* purpose: dump whole datafile.  Open file, call section dumps.
*
*/
void do_dump(name)
char *name;
{
  vertex_id v_id;
  edge_id e_id;
  facet_id f_id;
  body_id b_id;
  char defaultname[100];

  if ( name == NULL )
    { /* construct default name */
      strncpy(defaultname,datafilename,sizeof(defaultname)-5);
      strcat(defaultname,".dmp");
      name = defaultname;
    }

  sprintf(msg,"Dumping to %s.\n",name);
  outstring(msg);
  dumpfd = fopen(name,"w");

  if ( dumpfd == NULL )
    { sprintf(errmsg,"Cannot open file %s.\n",name);
      kb_error(1006,errmsg,RECOVERABLE);
    }


  fprintf(dumpfd,"// %s: Dump of structure.\n\n",name); 

  top_dump(dumpfd);

  /* vertex dump */
  fputs("\nvertices        /*  coordinates  */    \n",dumpfd);
  FOR_ALL_VERTICES(v_id)
     vertex_dump(v_id,dumpfd);
     
  /* edge dump */
  fputs("\nedges  \n",dumpfd);
  FOR_ALL_EDGES(e_id)
     edge_dump(e_id,dumpfd);

  /* facet dump */
  if ( web.representation == SIMPLEX )
     fputs("\nfaces    /* vertex set */      \n",dumpfd);
  else
     fputs("\nfaces    /* edge loop */      \n",dumpfd);
  FOR_ALL_FACETS(f_id)
     facet_dump(f_id,dumpfd);

  /* body dump */
  calc_content(Q_FIXED|Q_INFO);
  fputs("\nbodies  /* facets */\n",dumpfd);
  FOR_ALL_BODIES(b_id)
     body_dump(b_id,dumpfd);

  bottom_dump(dumpfd);

  fclose(dumpfd);
} /* end do_dump() */

/********************************************************************
*
* function: list_proc()
*
* purpose: list one procedure, recursing if necessary for
*             dependencies.
*/
void list_proc(p,level)
int p;     /* number of global variable of procedure */
int level; /* for recursion depth check */
{ struct expnode *ex = &globals[p].value.proc;
  struct treenode *node;
  int pp;

  if ( level > web.global_count ) 
  { sprintf(errmsg,"'%s' involved in cyclic definition.\n",globals[p].name);
     kb_error(1007,errmsg,WARNING);
  }
  else
  {
    /* check dependencies */
    for ( node = ex->start ; node != ex->root ; node ++  )
      if ( node->type == PROCEDURE_ )
      { pp = node->op1.intval;
         if ( globals[pp].proc_timestamp < proc_timestamp )
            list_proc(pp,level+1);
      }
  }

  /* print */
  if ( globals[p].flags & PERMANENT )
    sprintf(msg,"%s ::= ",globals[p].name);
  else sprintf(msg,"%s := ",globals[p].name);
  outstring(msg);
  outstring(print_express(&globals[p].value.proc,'X'));
  outstring("\n");
  globals[p].proc_timestamp = proc_timestamp;
} /* end list_proc() */

/*********************************************************************
*
* function: list_procedures()
*
* purpose: list defined procedures in time order.
*
*/
void list_procedures()
{ int i;

  proc_timestamp++;

  for ( i = 0 ; i < web.global_count ; i++ )
     {
        if ( globals[i].flags & SUBROUTINE ) 
            if ( globals[i].proc_timestamp < proc_timestamp )
              list_proc(i,0);
     }

  /* redefined single letters */
  for ( i = 0 ; i < 128 ; i++ )
     if ( single_redefine[i].start )
     { 
        sprintf(msg,"%c :::= ",i);
        outstring(msg);
        outstring(print_express(&single_redefine[i],'X'));
        outstring("\n");
     }
} /* end list_procedures() */

/****************************************************************************
*
* function: bottom_dump()
*
* purpose: print info at bottom of fe dump file.
*          Called also by bottominfo command.
*/

void bottom_dump(fd)
FILE *fd;
{ int i;
  FILE *old_fd = outfd;
  int e_type;

  outfd = fd;
  readflag = 0;


  outstring("\nread\n"); readflag = 1;
  /* defined commands */
  for ( i = 0 ; i < web.global_count ; i++ )
     { 
        if ( globals[i].flags & STRINGVAL )
          {
             sprintf(msg,"%s := ",globals[i].name);
             convert_string(globals[i].value.string,msg+strlen(msg));
             strcat(msg,"\n");
             outstring(msg);
          }
     }
  /* extra attributes with function definitions */
  for ( e_type = VERTEX ; e_type <= FACETEDGE ; e_type++ )
  { struct extra *ex;
     for ( i = 0, ex = web.skel[e_type].extras ; 
                i < web.skel[e_type].extra_count ; i++, ex++ )
     { if ( !(ex->flags & DUMP_ATTR) ) continue;
       if ( !ex->code.start )  continue;
        sprintf(msg,"\ndefine %s attribute %s %s",typenames[e_type], ex->name,
          attr_type_name[ex->type]);
        if ( ex->dim > 1 )
          sprintf(msg+strlen(msg)-1,"[%d]",ex->dim);
        outstring(msg);
        outstring(" function \n");
        outstring(print_express(&ex->code,'X'));
        outstring("\n");
     }
  }
  list_procedures();

  if ( show_expr[EDGE] ) 
    { 
      outstring(print_express(show_command+EDGE,'X'));
      outstring("\n");
    }
  if ( show_expr[FACET] ) 
    { 
      outstring(print_express(show_command+FACET,'X'));
      outstring("\n");
    }

  /* misc state saving */
  switch(torus_display_mode)
  { case TORUS_RAW_MODE: outstring("raw_cells\n"); break;
    case TORUS_CLIPPED_MODE: outstring("clipped\n"); break;
    case TORUS_CONNECTED_MODE: outstring("connected\n"); break;
  }
  if ( user_thickness_flag )
  { sprintf(msg,"thickness := %g\n",(double)thickness); outstring(msg); }
  if ( hessian_slant_cutoff != 0.0 ) 
  { sprintf(msg,"hessian_slant_cutoff := %g\n",(double)hessian_slant_cutoff); outstring(msg); }
  if ( everything_quantities_flag ) outstring("convert_to_quantities\n");
  if ( conj_grad_flag ) toggle_save("conj_grad");
  if ( quantities_only_flag && !everything_quantities_flag) 
      toggle_save("quantities_only");
  if ( !ribiere_flag ) toggle_save_off("ribiere");
  if ( assume_oriented_flag ) toggle_save("assume_oriented");
  if ( view_4D_flag ) toggle_save("view_4D");
  if ( kusner_flag ) toggle_save("kusner");
  if ( estimate_flag ) toggle_save("estimate");
  if ( unit_normal_flag ) toggle_save("deturck");
  if ( sqgauss_flag ) toggle_save("sqgauss");
  if ( autopop_flag ) toggle_save("autopop");
  if ( autochop_flag ) toggle_save("autochop");
  if ( quiet_flag ) toggle_save("quiet");
  if ( !hessian_quiet_flag ) toggle_save_off("hessian_quiet");
  if ( quiet_go_flag ) toggle_save("quietgo");
  if ( old_area_flag ) toggle_save("old_area");
  if ( approx_curve_flag ) toggle_save("approx_curv");
  if ( runge_kutta_flag ) toggle_save("runge_kutta");
  if ( check_increase_flag ) toggle_save("check_increase");
  if ( web.area_norm_flag ) toggle_save("area_normalization");
  if ( conf_edge_curv_flag ) toggle_save("conf_edge");
  if ( effective_area_flag ) toggle_save("effective_area");
  if ( web.torus_body_flag ) toggle_save("connected");
  if ( web.torus_clip_flag ) toggle_save("clipped");
  if ( thickenflag ) toggle_save("thicken");
  if ( !innerflag ) toggle_save_off("show_inner");
  if ( !outerflag ) toggle_save_off("show_outer");
  if ( colorflag ) toggle_save("colormap");
  if ( hessian_by_diff_flag ) toggle_save("hessian_diff");
  if ( hessian_normal_flag ) toggle_save("hessian_normal");
  if ( hessian_normal_perp_flag ) toggle_save("hessian_normal_perp");
  if ( hessian_double_normal_flag ) toggle_save("hessian_double_normal");
  if ( post_project_flag ) toggle_save("post_project");
  if ( mean_curv_int_flag ) toggle_save("mean_curvature_integral");
  if ( normal_curvature_flag ) toggle_save("normal_curvature");
  if ( div_normal_curvature_flag ) toggle_save("div_normal_curvature");
  if ( !shading_flag ) toggle_save_off("shading");
  if ( !color_flag ) toggle_save_off("facet_colors");
  if ( boundary_curvature_flag ) toggle_save("boundary_curvature");
  if ( normal_motion_flag ) toggle_save("normal_motion");
  if ( check_pinning_flag ) toggle_save("pinning");
  if ( !metric_convert_flag ) toggle_save_off("metric_conversion");
  if ( autorecalc_flag ) toggle_save("autorecalc");
  if ( make_pos_def_flag ) toggle_save("force_pos_def");
  if ( !gv_binary_flag ) toggle_save_off("gv_binary");
  if ( self_similar_flag ) toggle_save("self_similar");
  if ( interp_bdry_param ) toggle_save("interp_bdry_param");
  if ( web.bodycount && !web.gravflag ) toggle_save_off("gravity");
  if ( hessian_linear_metric_flag )
  { toggle_save("linear_metric");
#ifdef LONGDOUBLE
     sprintf(msg,"linear_metric_mix := %2.*Lg\n",DPREC,linear_metric_mix);
#else
     sprintf(msg,"linear_metric_mix := %2.15g\n",(DOUBLE)linear_metric_mix);
#endif 
     outstring(msg);
  }
  if ( transform_expr[0] )
     { 
        sprintf(msg,"transform_expr \"%s\"\n",transform_expr);
        outstring(msg);
     }
  if ( !transforms_flag && transform_count) toggle_save_off("transforms");
  outstring("\n\n");
  if ( web.target_tolerance != DEFAULT_TARGET_TOLERANCE )
  { sprintf(msg,"target_tolerance := %2.15g\n",(DOUBLE)web.target_tolerance);
    outstring(msg);
  }
  if ( brightness != DEFAULT_BRIGHTNESS )
  { sprintf(msg,"brightness := %2.15g\n",(DOUBLE)brightness);
    outstring(msg);
  }
  /* windup */
  outfd = old_fd;
} /* end bottom_dump */

/*********************************************************************
*
* function: facetedge_dump()
*
* purpose:  List facetedge with facet and edge links.  Meant for
*           screen display for debugging, not standard dump file.
*
*/
void facetedge_dump(fe,fd)
facetedge_id fe;
FILE *fd;
{ edge_id e_id;
  facet_id f_id;
  FILE *old_fd = outfd;
  int i,k; struct extra *ex;
  outfd = fd;

  e_id = get_fe_edge(fe);
  f_id = get_fe_facet(fe);
  sprintf(msg,"%6d    %5d  %5d  %8d %8d     %8d %8d",
      1+ordinal(fe), inverted(e_id) ? -1-ordinal(e_id) : 1+ordinal(e_id),
            inverted(f_id) ? -1-ordinal(f_id) : 1+ordinal(f_id),
            1+ordinal(get_prev_edge(fe)), 1+ordinal(get_next_edge(fe)),
            1+ordinal(get_prev_facet(fe)), 1+ordinal(get_next_facet(fe))); 
  outstring(msg); 

  /* extra attributes */
  for ( i = 0, ex = web.skel[FACETEDGE].extras ; 
                i < web.skel[FACETEDGE].extra_count; i++,ex++ )
    { char *at = get_extra(fe,i);
      if ( !(ex->flags & DUMP_ATTR) ) continue;
      if ( ex->dim == 0 ) continue;
      sprintf(msg," %s ",ex->name);
      for ( k = 0 ; k < ex->dim ; k++ )
         switch ( ex->type )
         { case REAL_ATTR: 
#ifdef LONGDOUBLE
                sprintf(msg+strlen(msg),"%2.*Lg ",DPREC,((REAL*)at)[k]); 
#else
                sprintf(msg+strlen(msg),"%2.15g ",((REAL*)at)[k]); 
#endif 
                break;
            case INTEGER_ATTR: sprintf(msg+strlen(msg),"%d ",((int *)at)[k]); 
                break;
            case ULONG_ATTR: sprintf(msg+strlen(msg),"%ld ",((unsigned long *)at)[k]); 
                break;
         }
      outstring(msg);
    }
  outstring("\n");
  outfd = old_fd;
} /* end facetedge_dump() */


/*********************************************************************
*
*  Function:  dump_force()
*
*  Purpose:    List forces at vertices for debugging purposes.
*/

void dump_force()
{
  char name[400];

  FILE *fd;
  vertex_id v_id;
  int i;
  FILE *old_fd = outfd;


  prompt("Enter name of dump file: ",name,sizeof(name));
  if ( name[0] == '\0' ) return;

  fd = fopen(name,"w");

  if ( fd == NULL )
    { sprintf(errmsg,"Cannot open file %s.\n",name);
      kb_error(1008,errmsg,RECOVERABLE);
    }

  outfd = fd;
  sprintf(msg,"%s: Dump of force.\n\n",name); outstring(msg); 

  /* vertex dump */
  outstring("\nvertices              coordinates      \n");
  FOR_ALL_VERTICES(v_id)
  { REAL *x,*f,mag;
    x = get_coord(v_id);
    switch(SDIM)
        { case 2:
#ifdef LONGDOUBLE
          sprintf(msg,"%3d    %17.*Lf %17.*Lf  |x| = %17.*Lf ",
             1+ordinal(v_id),DPREC,x[0],DPREC,x[1],DPREC,sqrt(SDIM_dot(x,x))); 
#else
          sprintf(msg,"%3d    %17.15f %17.15f  |x| = %17.15f ",
             1+ordinal(v_id),x[0],x[1],sqrt(SDIM_dot(x,x))); 
#endif 
             outstring(msg); 
            break;
          case 3:
#ifdef LONGDOUBLE
          sprintf(msg,"%3d  %17.*Lf %17.*Lf %17.*Lf  |x| = %17.*Lf ",
             1+ordinal(v_id),DPREC,x[0],DPREC,x[1],DPREC,x[2],DPREC,sqrt(SDIM_dot(x,x))); 
#else
          sprintf(msg,"%3d  %17.15f %17.15f %17.15f  |x| = %17.15f ",
             1+ordinal(v_id),x[0],x[1],x[2],sqrt(SDIM_dot(x,x))); 
#endif 
             outstring(msg); 
            break;
         }
    if ( get_vattr(v_id) & CONSTRAINT )
    { conmap_t *conmap = get_v_constraint_map(v_id);
      sprintf(msg,"  constraints "); outstring(msg); 
      for ( i = 1 ; i < (int)conmap[0] ; i++ )
      if ( !(get_constraint(conmap[i])->attr & GLOBAL) )
        { sprintf(msg,"%d ",conmap[i]&CONMASK); outstring(msg); }
    }
    if ( get_vattr(v_id)&FIXED ) { sprintf(msg, "  fixed "); outstring(msg); }
    sprintf(msg,"\n"); outstring(msg); 

    f = get_force(v_id);
    mag = sqrt(SDIM_dot(f,f));
    switch(SDIM)
    { case 2:
#ifdef LONGDOUBLE
          sprintf(msg,"f:    %17.*Lf %17.*Lf |f| = %g\n\n",
             DPREC,f[0],DPREC,f[1],(DOUBLE)mag); outstring(msg); 
#else
          sprintf(msg,"f:    %17.15f %17.15f |f| = %g\n\n",
             f[0],f[1],(DOUBLE)mag); outstring(msg); 
#endif 
          break;
      case 3:
#ifdef LONGDOUBLE
          sprintf(msg,"f:    %17.*Lf %17.*Lf %17.*Lf  |f| = %g\n\n",
             DPREC,f[0],DPREC,f[1],DPREC,f[2],(DOUBLE)mag); outstring(msg); 
#else
          sprintf(msg,"f:    %17.15f %17.15f %17.15f  |f| = %g\n\n",
             f[0],f[1],f[2],(DOUBLE)mag); outstring(msg); 
#endif 
      break;
    }
  }

  fclose(fd);

  outfd = old_fd;
} /* end dump_force */

/***********************************************************************************
*
* function: list_attributes()
*
* purpose: List all defined attributes of elements (the ones with entries
*     in attribute table, not flags like FIXED).
*
*/

void list_attributes()
{ int n,k;

  outstring(
     "Extra attributes (and dynamically allocated internal attributes)\n");
  for ( n = 0 ; n < NUMELEMENTS ; n++ )
  {  
      sprintf(msg,"%s attributes:\n",typenames[n]); outstring(msg);
      for ( k = 0 ; k < web.skel[n].extra_count ; k++ )
      { if ( web.skel[n].extras[k].dim == 0 ) continue;
         sprintf(msg,"%32s %s[%d]\n",web.skel[n].extras[k].name,
              attr_type_name[web.skel[n].extras[k].type],web.skel[n].extras[k].dim);
         outstring(msg);
      }
  }
} /* end list_attributes() */

