/*************************************************************
*  This file is part of the Surface Evolver source code.     *
*  Programmer:  Ken Brakke, brakke@susqu.edu                 *
*************************************************************/


/*****************************************************************
*
*  File: evalmore.c
*
*  Purpose: Overflow from eval() in evaltree.c.
*/

#include "include.h" 
#include "ytab.h"
#include "lex.h"
#ifdef MAC_APP
#define S_IFIFO 0x100
#elif !defined(MAC_CW)
#include <sys/stat.h>
#ifndef S_IFIFO
#define S_IFIFO _S_IFIFO
#endif
#endif
/*
int wait ARGS((int*));
int fork ARGS((void));
int execlp();
int stat();
*/
void flip_toggle ARGS((int*,int,char*));
void set_body ARGS((element_id,body_id));
REAL * more_other_stuff ARGS((struct treenode *,REAL *,int*,int*,element_id));

/************************************************************************
*
* Function: other_stuff()
*
* Purpose: execute command nodes, overflow from eval().
*
* Return: stacktop pointer
*/

REAL * other_stuff(node,stacktop,recalc_flag,update_display_flag,q_id,newstack)
struct treenode * node;
REAL *stacktop;  /* current */
int *recalc_flag;
int *update_display_flag;
element_id q_id;
REAL *newstack; /* stack base */
{ int i,j,n,k;
  char *s,*h;
  REAL scale;
  int old; /* old state of toggle */
  element_id id,fe,f_id,b_id;

  switch ( node->type )
  {
    case REORDER_STORAGE_:
       reorder_storage(); break;

    case RENUMBER_ALL_:
       renumber_all(); break;

    case VIEW_MATRIX_:
       i = (int)(*(stacktop--));
       k = (int)(*(stacktop--));
       if ( (k < 1) || (k > SDIM+1) || (i < 1) || (i > SDIM+1) )
       { sprintf(errmsg,
           "Illegal index: view_matrix[%d][%d] (must be 1 to %d)\n",k,i,SDIM);
         kb_error(2512,errmsg,RECOVERABLE);
       }
       *(++stacktop) = view[k-1][i-1];
       break;

    case PAUSE_:
       prompt("Paused; hit RETURN to continue.",msg,msgmax);
       break;

    case HELP_KEYWORD:
       keyword_help(node->op1.string);
       break;
  
    case CREATE_VERTEX_:
     { vertex_id v_id;
       REAL x[MAXCOORD];
       for ( i = SDIM-1 ; i >= 0 ; i-- ) x[i] = *(stacktop--);
       v_id = new_vertex(x,NULLID);
       new_vertex_id = ordinal(v_id) + 1;
       set_original(v_id,new_vertex_id);
       *(++stacktop) = (REAL)new_vertex_id;
       *recalc_flag = 1;
       break;
     }

    case CREATE_EDGE_:
      { vertex_id v_id2; 
        vertex_id v_id1; 
        int v2 = (int)(*stacktop--);
        int v1 = (int)(*stacktop--);
        edge_id e_id;
        v_id2 = get_ordinal_id(VERTEX,v2-1); 
        v_id1 = get_ordinal_id(VERTEX,v1-1); 
        if ( !valid_id(v_id1) )
        { sprintf(errmsg,"Invalid tail vertex %d in NEW_EDGE.\n",v1);
           kb_error(1938,errmsg,RECOVERABLE);
        }
        if ( !valid_id(v_id2) )
        { sprintf(errmsg,"Invalid head vertex %d in NEW_EDGE.\n",v2);
           kb_error(1939,errmsg, RECOVERABLE);
        }
        e_id = new_edge(v_id1,v_id2,NULLID);
        new_edge_id = ordinal(e_id)+1;
        set_original(e_id,new_edge_id);
        *(++stacktop) = (REAL)new_edge_id;
        *recalc_flag = 1;
      }
      break;

    case CREATE_FACET_:
    { facetedge_id old_fe = NULLID;
       edge_id e_id;
       vertex_id tv,hv;
#define ORD_ID(id) ((id)<0 ? inverse_id(get_ordinal_id(EDGE,-1-(id)))  : \
                                   get_ordinal_id(EDGE,(id)-1))
       if ( (web.representation == STRING) || (web.representation == SOAPFILM) )
       { n = node[node->left].op1.intval; /* size of edge list */
         for ( i = 0 ; i < n ; i++ ) 
           if ( !valid_id(ORD_ID((int)(stacktop[-i]))) )
               { sprintf(msg,"Invalid edge %g in NEW_FACET.\n",(double)stacktop[-i]);
                  kb_error(1974,msg,RECOVERABLE);
               }
         /* check continuity */
         i = 1-n;
         e_id = ORD_ID((int)(stacktop[i]));
         hv = get_edge_headv(e_id);
         for ( i++ ; i <= 0 ; i++ )
          { e_id = ORD_ID((int)(stacktop[i]));
            tv = get_edge_tailv(e_id);
            if ( !equal_id(hv,tv) )
            { sprintf(errmsg,
               "Inconsistency: edge %d tail vertex %d disagrees with previous head %d.\n",
                           ordinal(e_id)+1,ordinal(tv)+1,ordinal(hv)+1);
                   kb_error(1975,errmsg,RECOVERABLE);
              }
            hv = get_edge_headv(e_id);
          }
         if ( web.representation == SOAPFILM)
          {
            e_id = ORD_ID((int)(stacktop[0]));
            hv = get_edge_headv(e_id);
            e_id = ORD_ID((int)(stacktop[1-n]));
            tv = get_edge_tailv(e_id);
            if ( !equal_id(hv,tv) ) 
            { sprintf(errmsg,
                      "Inconsistency in first edge tail vertex and last edge head.\n");
               kb_error(1977,errmsg,RECOVERABLE);
            }
          }
         f_id = new_facet();
         new_facet_id = ordinal(f_id) + 1;
         set_original(f_id,new_facet_id);
         for ( i = 0 ; i < n ; i++ ) 
          { facetedge_id edge_fe;
            e_id = ORD_ID((int)(stacktop[i-n+1]));
            fe = new_facetedge(f_id,e_id);
            if ( !valid_id(get_facet_fe(f_id)) )
               set_facet_fe(f_id,fe);
            else
            { set_next_edge(old_fe,fe);
               set_prev_edge(fe,old_fe);
            }
            old_fe = fe;
            /* add to edge facet list, not in geometric order */
            edge_fe = get_edge_fe(e_id);
            if ( valid_id(edge_fe) )
              { /* insert in chain */
                set_next_facet(fe,get_next_facet(edge_fe));
                set_prev_facet(fe,edge_fe);
                set_prev_facet(get_next_facet(fe),fe);
                set_next_facet(edge_fe,fe);
              }
            else
              { set_next_facet(fe,fe);
                set_prev_facet(fe,fe);
                set_edge_fe(e_id,fe);      /* link edge to rest of world */
              }
          }
          if ( equal_id(get_fe_tailv(get_facet_fe(f_id)),get_fe_headv(fe)) )
          {
            set_next_edge(fe,get_facet_fe(f_id));  /* close up ring */
            set_prev_edge(get_facet_fe(f_id),fe);
          }
          if ( web.representation == SOAPFILM ) 
          { fe = get_facet_fe(f_id);
            for ( i = 0 ; i < n ; i++,fe = get_next_edge(fe) )
                fe_reorder(get_fe_edge(fe));
            if ( n > 3 ) face_triangulate(f_id,n);
          }
          stacktop -= n; /* pop id's */
       }
       else if ( web.representation == SIMPLEX )
       {  vertex_id *v;
          n = node[node->left].op1.intval; /* size of edge list */
          for ( i = 0 ; i < n ; i++ )
            if ( !valid_id(get_ordinal_id(VERTEX,(int)stacktop[-i]-1) ) )
               { sprintf(msg,"Invalid vertex %d in NEW_FACET.\n",(int)stacktop[-i]);
                 kb_error(1978,msg,RECOVERABLE);
               }
          f_id = new_facet();
          v = get_facet_vertices(f_id);
          for ( i = 0 ; i < n ; i++ ) v[i] = get_ordinal_id(VERTEX,(int)stacktop[-n+i+1]-1);
          stacktop -= n; /* pop id's */
       }
    }
    new_facet_id = ordinal(f_id) + 1;
    *(++stacktop) = (REAL)new_facet_id;
    *recalc_flag = 1;
    break;

    case CREATE_BODY_:
       { body_id b_id = new_body();
          new_body_id = ordinal(b_id) + 1;
          set_original(b_id,new_body_id);
          *(++stacktop) = (REAL)new_body_id;
       }
       *recalc_flag = 1;
       break;

    case GET_TORUS_PERIODS_:
         j = (int)(*stacktop--);
         i = (int)(*stacktop--);
         if ( (j<1) || (j > SDIM))
         { sprintf(errmsg,"Torus_periods index %d; must be between 1 and %d.\n",j,SDIM);
           kb_error(4327,errmsg,RECOVERABLE);
         }
         if ((i<1) || (i>SDIM) )
         { sprintf(errmsg,"Torus_periods index %d; must be between 1 and %d.\n",i,SDIM);
           kb_error(4327,errmsg,RECOVERABLE);
         }
         *++stacktop = web.torus_period[i-1][j-1]; /* 1-based indexing */
         break;

    case GET_INVERSE_PERIODS_:
         j = (int)(*stacktop--);
         i = (int)(*stacktop--);
         if ( (j<1) || (j > SDIM) )
         { sprintf(errmsg,"Inverse_periods index %d; must be between 1 and %d.\n",j,SDIM);
           kb_error(4328,errmsg,RECOVERABLE);
         }
         if ( (i<1) || (i>SDIM) )
         { sprintf(errmsg,"Inverse_periods index %d; must be between 1 and %d.\n",i,SDIM);
           kb_error(4328,errmsg,RECOVERABLE);
         }
         *++stacktop = web.inverse_periods[i-1][j-1]; /* 1-based indexing */
         break;

    case SET_GAP_CONSTANT_:
       web.spring_constant = *(stacktop--); 
       if ( everything_quantities_flag )
               GEN_QUANTS[gap_quantity_num].modulus = web.spring_constant;
       if ( web.spring_constant == 0.0 ) web.convex_flag = 0;
       else web.convex_flag = 1;
       *recalc_flag = 1;
       break;

    case GET_INTERNAL_:
       *++stacktop = get_internal_variable(node->op1.intval);
          break;
  
    case SET_INTERNAL_:
    { REAL oldvalue;
      REAL val; 
      /* get old value */
      switch(node->op1.intval)
      { case V_TOLERANCE: oldvalue = web.tolerance; break;
        case V_LAST_ERROR: oldvalue = last_error; break;
        case V_BRIGHTNESS: oldvalue = brightness; break;
        case V_BACKGROUND: oldvalue = background_color; break;
        case V_TARGET_TOLERANCE: oldvalue = web.target_tolerance;
              break;
        case V_HESS_EPSILON: oldvalue = hessian_epsilon; break;
        case GRAV_CONST_: oldvalue = web.grav_const; break;
        case V_THICKNESS: oldvalue = thickness; break;
        case V_HESSIAN_SLANT_CUTOFF: oldvalue = hessian_slant_cutoff; break;
        case V_AMBIENT_PRESSURE: oldvalue = web.pressure; break;
        case V_DIFFUSION: oldvalue = web.diffusion_const;
            web.diffusion_flag = 1; break;
        case V_SCALE_SCALE: oldvalue = web.scale_scale; break;
        case V_SCALE: oldvalue = web.scale; break;
        case V_SCALE_LIMIT: oldvalue = web.maxscale; break;
        case V_TIME: oldvalue = total_time; break;
        case V_PICKVNUM: oldvalue = pickvnum; break;
        case V_PICKENUM: oldvalue = pickenum; break;
        case V_PICKFNUM: oldvalue = pickfnum; break;
        case V_JIG_TEMP: oldvalue = web.temperature; break;
        case V_LINEAR_METRIC_MIX: oldvalue = linear_metric_mix; break;
        case V_GAP_CONSTANT: oldvalue = web.spring_constant; break;
        case V_QUADRATIC_METRIC_MIX: 
              oldvalue = quadratic_metric_mix; break;
        case V_RANDOM_SEED: oldvalue = random_seed;break;
        case V_RANDOM: kb_error(2864,"Cannot set random. Set random_seed instead.\n",WARNING); stacktop--; break;
        case V_INTEGRAL_ORDER: oldvalue = web.gauss2D_order; break;
        case V_INTEGRAL_ORDER_1D: oldvalue = web.gauss1D_order; break;
        case V_INTEGRAL_ORDER_2D: oldvalue = web.gauss2D_order; break;
        default: 
          kb_error(1209,"Internal: illegal variable number.\n",RECOVERABLE);
      }
      val = *(stacktop--);
      switch ( node->op2.intval )
      { case ASSIGN_:    break;
        case PLUSASSIGN_: val +=  oldvalue; break;
        case SUBASSIGN_: val = oldvalue - val; break;
        case MULTASSIGN_: val = oldvalue * val; break;
        case DIVASSIGN_: val = oldvalue / val; break;
      }
      switch(node->op1.intval)
      { case V_TOLERANCE: web.tolerance = val; break;
        case V_BRIGHTNESS: 
           if ( (val < 0.0) || (val > 1.0) )
               kb_error(2874,"Brightness must be between 0 and 1.\n",RECOVERABLE);
            else { brightness = val; update_display(); }
               break;
        case V_BACKGROUND:
           background_color = (int)val;
           update_display();
           break;

        case V_TARGET_TOLERANCE: web.target_tolerance = val;
              break;
        case V_HESS_EPSILON: hessian_epsilon = val; break;
        case GRAV_CONST_: web.grav_const = val; break;
        case V_DIFFUSION: web.diffusion_const = val; break;
        case V_AMBIENT_PRESSURE: web.pressure = val; break;
        case V_HESSIAN_SLANT_CUTOFF: hessian_slant_cutoff = val; break;
        case V_THICKNESS: thickness = val; user_thickness_flag = 1;
                  update_display();break;
        case V_SCALE_SCALE: web.scale_scale = val; break;
        case V_SCALE: web.scale = val; break;
        case V_SCALE_LIMIT: web.maxscale = val; break;
        case V_LAST_ERROR: last_error = (int)val; break;
        case V_TIME: total_time = val; break;
        case V_PICKVNUM: pickvnum = (int)val; break;
        case V_PICKENUM: pickenum = (int)val; break;
        case V_PICKFNUM: pickfnum = (int)val; break;
        case V_JIG_TEMP: web.temperature = val; break;
        case V_LINEAR_METRIC_MIX: linear_metric_mix=val; break;
        case V_GAP_CONSTANT: web.spring_constant = val; break;
        case V_QUADRATIC_METRIC_MIX: 
              quadratic_metric_mix=val; break;
        case V_RANDOM_SEED: random_seed = (int)val;
              srand(random_seed); srand48(random_seed); 
              kb_initr(random_seed); break;
        case V_RANDOM: kb_error(2864,"Cannot set random. Set random_seed instead.\n",WARNING); stacktop--; break;
        case V_INTEGRAL_ORDER: 
              web.gauss1D_order = (int)val;
              set_by_user_gauss_1D = web.gauss1D_order;
              web.gauss2D_order = (int)val;
              set_by_user_gauss_2D = web.gauss2D_order;
              gauss_setup();
              if ( web.modeltype == LAGRANGE )
              { gauss_lagrange_setup(
                   (web.dimension==1)?1:web.dimension-1,
                       web.lagrange_order,web.gauss1D_order);
                gauss_lagrange_setup(web.dimension,
                   web.lagrange_order,web.gauss2D_order);
              }
              *recalc_flag = 1;
              break;
        case V_INTEGRAL_ORDER_1D: 
              web.gauss1D_order = (int)val;
              set_by_user_gauss_1D = web.gauss1D_order;
              if ( web.modeltype == LAGRANGE )
                  gauss_lagrange_setup(
                   (web.dimension==1)?1:web.dimension-1,
                       web.lagrange_order,web.gauss1D_order);
              gauss_setup();
              *recalc_flag = 1;
              break;
        case V_INTEGRAL_ORDER_2D: 
              web.gauss2D_order = (int)val;
              set_by_user_gauss_2D = web.gauss2D_order;
              gauss_setup();
              if ( web.modeltype == LAGRANGE )
                gauss_lagrange_setup(web.dimension,
                   web.lagrange_order,web.gauss2D_order);
              *recalc_flag = 1;
              break;
        default: 
          kb_error(1209,"Internal: illegal variable number.\n",RECOVERABLE);

      }
      break;
    } /* end SET_INTERNAL_ */

    case SIZEOF_:
      *++stacktop = (REAL)web.skel[node->op2.intval].extras[node->op1.intval].dim;
       break;

    case TORDUP_:
       i = (int)*(stacktop--);
       if ( ! web.torus_flag )
       { outstring("Torus model not in effect.\n");
         break;
       }
       if ( i < 1 || i > SDIM ) outstring("Improper period number.\n");
       else tordup(i-1);
       break;
     
    case SKINNY_:
       sprintf(msg,"New edges: %d\n",web.refine_count = skinny(*(stacktop--)));
       outstring(msg);
       recalc();
       break;

    case PUSHQPRESSURE_:
     *++stacktop = GEN_QUANTS[globals[node->op1.intval].value.quant].pressure;
          break;
    case PUSHQTARGET_:
     *++stacktop = GEN_QUANTS[globals[node->op1.intval].value.quant].target;
          break;
    case PUSHQMODULUS_:
     *++stacktop = GEN_QUANTS[globals[node->op1.intval].value.quant].modulus;
          break;
    case PUSHQTOLERANCE_:
     *++stacktop = GEN_QUANTS[globals[node->op1.intval].value.quant].tolerance;
          break;
    case PUSHMMODULUS_:
     *++stacktop = METH_INST[globals[node->op1.intval].value.meth_inst].modulus;
          break;
    case PUSHQVALUE_:
      { struct gen_quant *q 
              = GEN_QUANTS + globals[node->op1.intval].value.quant;
        if ( q->timestamp<global_timestamp )
                   calc_quants(q->flags&(Q_INFO|Q_ENERGY|Q_FIXED));
        *++stacktop = q->value;
      }
      break;

    case PUSHMVALUE_:
      { struct method_instance *q 
              = METH_INST + globals[node->op1.intval].value.meth_inst;
        if ((q->timestamp<graph_timestamp) || (q->timestamp<web_timestamp))
                   calc_quants((Q_INFO|Q_ENERGY|Q_FIXED));
        *++stacktop = q->value;
      }
      break;

    case PUSHQVOLCONST_:
      *++stacktop = GEN_QUANTS[globals[node->op1.intval].value.quant].volconst;
      break;
  
    case PUSH_NAMED_QUANTITY:
    case PUSH_METHOD_INSTANCE_:
       *++stacktop = (REAL)node->op1.intval;
       break;
    
    case FIX_QUANTITY_:
    { struct gen_quant *q =
           GEN_QUANTS + node->op1.intval;
       q->flags &= ~(Q_ENERGY|Q_INFO);
       q->flags |= Q_FIXED;
    }
    break;

    case UNFIX_QUANTITY_:
    { struct gen_quant *q =
           GEN_QUANTS + node->op1.intval;
       q->flags &= ~(Q_ENERGY|Q_FIXED);
       q->flags |= Q_INFO;
    }
    break;

   case UNFIX_PARAMETER_:
   { struct global *p = globals + node->op1.intval;
      int old = p->flags & OPTIMIZING_PARAMETER;
      if ( !old )
      { if ( optparamcount >= MAXOPTPARAM-1 )
        { sprintf(errmsg,"Too many optimizing parameters.  Increase MAXOPTPARAM in extern.h and recompile.\n");
           kb_error(2674,errmsg,RECOVERABLE);
        }
        p->flags |= OPTIMIZING_PARAMETER;
        memset((char*)&optparam[optparamcount],0,sizeof(struct optparam));
        optparam[optparamcount++].pnum = node->op1.intval;
        sprintf(msg,"%s now optimizing parameter. (was not)\n",p->name);
      }
      else sprintf(msg,"%s now optimizing parameter. (already was)\n",p->name);
      break;
   }

   case FIX_PARAMETER_:
    { struct global *p = globals + node->op1.intval;
      int old = p->flags & OPTIMIZING_PARAMETER;
      int spot;
      if ( old )
      { for ( spot = 0 ; spot < optparamcount ; spot++ )
           if ( optparam[spot].pnum == node->op1.intval ) break;
        if ( spot == optparamcount )
        { sprintf(errmsg,"`%s` not found in optimizing parameter list.\n",p->name);
           kb_error(2675,errmsg,WARNING);
           break;
        }
        p->flags &= ~OPTIMIZING_PARAMETER;
        optparam[spot] = optparam[--optparamcount];
        sprintf(msg,"%s now nonoptimizing parameter. (was not)\n",p->name);
      }
      else sprintf(msg,"%s now nonoptimizing parameter. (already was)\n",p->name);
      break;
    }

   /* attribute setting in set statements */
   case SET_ATTRIBUTE_: 
   case SET_ATTRIBUTE_L: 
    if ( node->op1.intval ) id = (node+node->op1.intval)->op2.id;
    else id = q_id;
    parallel_update_flag[id_type(id)] = 1; /* set when element info changed */
    switch ( node->op2.intval )
    {
      case SET_FIXED_: set_attr(id,FIXED); break;
      case SET_TRIPLE_PT_: set_attr(id,TRIPLE_PT); --stacktop; break;
      case SET_TETRA_PT_: set_attr(id,TETRA_PT); --stacktop; break;
      case SET_AXIAL_POINT_: set_attr(id,AXIAL_POINT); --stacktop; break;

      case SET_EXTRA_ATTR_:
         { struct extra *ext;
           int inx; /* index */
           n = node->op3.intval[0] & ((1<<ESHIFT)-1);
           if ( node->op1.intval ) id = (node+node->op1.intval)->op2.id;
           else id = q_id;
           ext = web.skel[id_type(id)].extras + n;
           if ( ext->dim > 1 ) 
             { inx = (int)*(stacktop--) - 1; /* 1 based index */
               if ( inx >= ext->dim )
               { sprintf(errmsg,"Attribute %s index too big, %d. Maximum is %d.\n",
                       ext->name,inx+1,ext->dim);
                  kb_error(1210,errmsg,RECOVERABLE);
               }
               if ( inx < 0 )
               { sprintf(errmsg,"Attribute %s index is %d; must be positive .\n",
                       ext->name,inx+1); 
                  kb_error(1211,errmsg,RECOVERABLE);
               }
             }
           else inx = 0;
           switch ( ext->type )
           { case REAL_ATTR:  ((REAL*)get_extra(id,n))[inx] = *(stacktop--); 
                break;
              case INTEGER_ATTR:
                ((int*)get_extra(id,n))[inx] = (int)*(stacktop--);
                break;
              case ULONG_ATTR:
                ((unsigned long*)get_extra(id,n))[inx] = (unsigned long)*(stacktop--);
                break;
           }
        }
        break;

      case SET_DENSITY_:
        { REAL density = *(stacktop--);
           if ( density != 1.0 ) set_attr(id,DENSITY);
           switch ( id_type(id) )
              { case EDGE: set_edge_density(id,density); 
                    pressure_set_flag = 0; break;
                case FACET: set_facet_density(id,density);
                    pressure_set_flag = 0; break;
                case BODY: set_body_density(id,density);
                    web.gravflag = 1;
                    if (gravity_quantity_num >= 0 )
                        GEN_QUANTS[gravity_quantity_num].modulus =
                              web.grav_const;
                    break;
              }
           *recalc_flag = 1;
          }
        break;

      case SET_PHASE_:
        { int phase = (int)*(stacktop--);
           switch ( id_type(id) )
              { case BODY: set_b_phase(id,phase); break;
                case FACET: set_f_phase(id,phase); break;
              }
           *recalc_flag = 1;
        }
        break;

       case SET_VOLCONST_: 
        { REAL v = *(stacktop--);
          REAL oldvcon = get_body_volconst(id);
          set_body_volconst(id,v); 
          set_body_volume(id,get_body_volume(id)+v-oldvcon);
          set_body_oldvolume(id,get_body_oldvolume(id)+v-oldvcon);
          if ( everything_quantities_flag )
            GEN_QUANTS[get_body_volquant(id)].volconst = v;
           *recalc_flag = 1;
          break;
        }

       case SET_TARGET_: 
        { REAL v = *(stacktop--);
           if ( get_attr(id) & PRESSURE )
              kb_error(2013,"Must unset body pressure before fixing target.\n",
                  RECOVERABLE);
          set_attr(id,FIXEDVOL);
          set_body_fixvol(id,v); 
          *recalc_flag = 1;
          break;
        }

       case SET_PRESSURE_: 
        { REAL p = *(stacktop--);
           if ( get_attr(id) & FIXEDVOL )
              kb_error(2013,"Must unset body target before fixing pressure.\n",
                  RECOVERABLE);
          set_attr(id,PRESSURE);
          web.pressflag = 1;
          set_body_pressure(id,p); 
          if ( everything_quantities_flag )
          { struct gen_quant *q = GEN_QUANTS + get_body_volquant(id);
            if ( p == 0.0 )
            { q->modulus = 1.0;
               if ( !(q->flags & Q_INFO) )
               { q->flags &= ~(Q_ENERGY|Q_FIXED);
                 q->flags |= Q_INFO;
                 reset_quant_bits();
               }
            }
            else 
            { q->modulus = -p;
               if ( !(q->flags & Q_ENERGY ) )
               { q->flags &= ~(Q_INFO|Q_FIXED);
                  q->flags |= Q_ENERGY;
                  reset_quant_bits();
               }
            }
          }
          *recalc_flag = 1;
          break;
          }

       case SET_OPACITY_:
          facet_alpha = *(stacktop)--;
          break;

       case SET_CONSTRAINT_:
        { int con = (int)*(stacktop--);
           struct constraint *constr;
          if ( (con<0) || (con>=MAXCON) || !web.constraint_addr[con])
            { sprintf(errmsg,"Illegal constraint number: %d.\n",con);
               kb_error(1212,errmsg,RECOVERABLE);
            }
           set_attr(id,CONSTRAINT);
           constr = get_constraint(con); 
           if ( constr->attr & CON_ENERGY )
              set_attr(id, BDRY_ENERGY);
           if ( constr->attr & CON_CONTENT )
              set_attr(id, BDRY_CONTENT);
           switch ( id_type(id) )
              { case VERTEX: set_v_constraint_map(id,con); break;
                case EDGE: set_e_constraint_map(id,con); break;
                case FACET: set_f_constraint_map(id,con); break;
              }
           *recalc_flag = 1;
          }
        break;

        case SET_ORIGINAL_:
          set_original(id,(int)*(stacktop--));
          break;

        case SET_COLOR_:
        { int color = (int)*(stacktop--);
           switch ( id_type(id) )
              { case EDGE: set_edge_color(id,color); break;
                case FACET: set_facet_color(id,color); break;
              }
           *update_display_flag = 1;
          }
        break;

        case SET_FRONTCOLOR_:
        { int color = (int)*(stacktop--);
           set_facet_frontcolor(id,color);
           *update_display_flag = 1;
          }
          break;

        case SET_BACKCOLOR_:
        { int color = (int)*(stacktop--);
           set_facet_backcolor(id,color);
           *update_display_flag = 1;
          }
          break;

        case SET_WRAP_:
        { int w = (int)*(stacktop--);
           set_edge_wrap(id,w);
           *update_display_flag = 1;
          }
          break;
        case SET_TAG_:
        { int tag = (int)*(stacktop--);
           switch ( id_type(id) )
              { 
                case FACET: set_tag(id,tag); break;
              }
          }
        break;

        case SET_COORD_: case SET_COORD_1: 
          if ( node->right ) /* indexed */
          { int dim = (int)*(stacktop--)-1;
            if ( (dim >= SDIM) || (dim < 0) )
                { sprintf(errmsg,
                   "Trying to set coordinate %d, in space dimension %d.\n",
                        dim,SDIM);
                   kb_error(2333,errmsg,RECOVERABLE);
                }
              get_coord(id)[dim] = *(stacktop--);
              *recalc_flag = 1;
              break;
           }
           /* else fall through */

         case SET_COORD_2:
         case SET_COORD_3: case SET_COORD_4: case SET_COORD_5:
         case SET_COORD_6: case SET_COORD_7: case SET_COORD_8:
           get_coord(id)[node->op2.intval-SET_COORD_1] = *(stacktop--);
           if ( node->op2.intval-SET_COORD_1 <= SDIM ) *recalc_flag = 1;
           break;

         case SET_PARAM_: case SET_PARAM_1: case SET_PARAM_2:
         case SET_PARAM_3: case SET_PARAM_4:
          if ( get_vattr(id) & BOUNDARY )
            { struct boundary *boundary = get_boundary(id);
               REAL *param = get_param(id);
               REAL *xx = get_coord(id);
               int pnum = node->op2.intval-SET_PARAM_1; 
               if ( node->right ) pnum += (int)*(stacktop--);
               if ( pnum > boundary->pcount )
               { sprintf(errmsg,"Parameter number is %d; maximum is %d.\n",
                      pnum,boundary->pcount);
                   kb_error(1213,errmsg,RECOVERABLE);
               }
               param[pnum] = *(stacktop--);
               for ( k = 0 ; k < SDIM ; k++ )
                  xx[k] = eval(boundary->coordf[k],param,id);
               *recalc_flag = 1;
            }
           else /* just parking a value probably */ 
              { REAL *param = get_param(id);
                int pnum = node->op2.intval-SET_PARAM_1;
                if ( node->right ) pnum += (int)*(stacktop--);
                param[pnum] = *(stacktop--);
              }
          break;

          case SET_FRONTBODY_:
          case SET_BACKBODY_:
            b_id = get_ordinal_id(BODY,(int)*(stacktop--)-1);
            if ( !valid_id(b_id) ) 
               kb_error(1995,"Invalid body in SET FRONTBODY or BACKBODY.\n",RECOVERABLE);
            set_body((node->op2.intval==SET_FRONTBODY_?id:inverse_id(id)),b_id);
            *recalc_flag = 1;
            break;
 
         default: 
            sprintf(errmsg,"Unhandled SET attribute %d\n",node->op2.intval);
            kb_error(4343,errmsg,RECOVERABLE);
            break;

      } /* end SET_ATTRIBUTE */
      break;

   /*****************************
   * assignop attribute values  *
   * for single element assigns *
   *****************************/
   case SET_ATTRIBUTE_A: 
   { int assign_type = node[1].op1.intval;
   if ( node->op1.intval ) id = (node+node->op1.intval)->op2.id;
   else id = q_id;
   parallel_update_flag[id_type(id)] = 1; /* set when element info changed */
   switch ( node->op2.intval )
   {  /* Those attributes not fitted for arithmetic assignment 
         have error test for that in yexparse() */

      case SET_ORIENTATION_:
          if ( *(stacktop--) >= 0.0 )
              unset_attr(id,NEGBOUNDARY);
          else  set_attr(id,NEGBOUNDARY);
          break;

      case SET_EXTRA_ATTR_:
        { struct extra *ext;
           int inx; /* index */
           n = node->op3.intval[0] & ((1<<ESHIFT)-1);
           if ( node->op1.intval ) id = (node+node->op1.intval)->op2.id;
           else id = q_id;
           ext = web.skel[id_type(id)].extras + n;
           if ( ext->dim > 1 ) 
            { inx = (int)*(stacktop--) - 1; /* 1 based index */
               if ( inx >= ext->dim )
               { sprintf(errmsg,"Attribute %s index too big, %d. Maximum is %d.\n",
                       ext->name,inx+1,ext->dim);
                  kb_error(1210,errmsg,RECOVERABLE);
               }
               if ( inx < 0 )
               { sprintf(errmsg,"Attribute %s index is %d; must be positive .\n",
                       ext->name,inx+1); 
                  kb_error(1211,errmsg,RECOVERABLE);
               }
            }
           else inx = 0;
           switch ( ext->type )
           { case REAL_ATTR:  
                switch ( assign_type )
                { case ASSIGN_:
                    ((REAL*)get_extra(id,n))[inx] = *(stacktop--); break;
                   case PLUSASSIGN_:
                    ((REAL*)get_extra(id,n))[inx] += *(stacktop--); break;
                   case SUBASSIGN_:
                    ((REAL*)get_extra(id,n))[inx] -= *(stacktop--); break;
                   case MULTASSIGN_:
                    ((REAL*)get_extra(id,n))[inx] *= *(stacktop--); break;
                   case DIVASSIGN_:
                    ((REAL*)get_extra(id,n))[inx] /= *(stacktop--); break;
                }
                break;
              case INTEGER_ATTR:
                switch ( assign_type )
                { case ASSIGN_:
                    ((int*)get_extra(id,n))[inx] = (int)*(stacktop--); break;
                   case PLUSASSIGN_:
                    ((int*)get_extra(id,n))[inx] += (int)*(stacktop--); break;
                   case SUBASSIGN_:
                    ((int*)get_extra(id,n))[inx] -= (int)*(stacktop--); break;
                   case MULTASSIGN_:
                    ((int*)get_extra(id,n))[inx] = 
                       (int)(((int*)get_extra(id,n))[inx] * *(stacktop--));break;
                   case DIVASSIGN_:
                    ((int*)get_extra(id,n))[inx] = 
                       (int)(((int*)get_extra(id,n))[inx] / *(stacktop--));break;
                }
                break;
              case ULONG_ATTR:
                switch ( assign_type )
                { case ASSIGN_:
                    ((unsigned long*)get_extra(id,n))[inx] = 
                        (unsigned long)*(stacktop--); break;
                   case PLUSASSIGN_:
                    ((unsigned long*)get_extra(id,n))[inx] += 
                        (unsigned long)*(stacktop--); break;
                   case SUBASSIGN_:
                    ((unsigned long*)get_extra(id,n))[inx] -= 
                        (unsigned long)*(stacktop--); break;
                   case MULTASSIGN_:
                    ((unsigned long*)get_extra(id,n))[inx] = 
                       (unsigned long)(((unsigned long*)get_extra(id,n))[inx] *
                           *(stacktop--));break;
                   case DIVASSIGN_:
                    ((unsigned long*)get_extra(id,n))[inx] = 
                       (unsigned long)(((unsigned long*)get_extra(id,n))[inx] / 
                           *(stacktop--));break;
                }
                break;
           }
        }
        break;
      case SET_DENSITY_:
        { REAL density = *(stacktop--);
           REAL oldvalue = 0.0;
           if ( density != 1.0 ) set_attr(id,DENSITY);
           switch ( id_type(id) )
              { case EDGE: oldvalue = get_edge_density(id); break;
                case FACET: oldvalue = get_facet_density(id); break;
                case BODY: oldvalue = get_body_density(id); break;
              }
           switch ( assign_type )
              { case ASSIGN_: break;
                case PLUSASSIGN_: density += oldvalue; break;
                case SUBASSIGN_: density = oldvalue - density; break;
                case MULTASSIGN_: density *= oldvalue; break;
                case DIVASSIGN_: density = oldvalue / density; break;
              }
           switch ( id_type(id) )
              { case EDGE: set_edge_density(id,density); 
                    pressure_set_flag = 0; break;
                case FACET: set_facet_density(id,density);
                    pressure_set_flag = 0; break;
                case BODY: set_body_density(id,density);
                    web.gravflag = 1;
                    if (gravity_quantity_num >= 0 )
                        GEN_QUANTS[gravity_quantity_num].modulus =
                              web.grav_const;
                    break;
              }
           *recalc_flag = 1;
          }
        break;

      case SET_PHASE_:
        { int phase = (int)*(stacktop--);
           switch ( id_type(id) )
              { case BODY: set_b_phase(id,phase); break;
                case FACET: set_f_phase(id,phase); break;
              }
           *recalc_flag = 1;
        }
        break;

      case SET_VOLCONST_: 
        { REAL v = *(stacktop--);
           REAL oldvalue = get_body_volconst(id); 
           switch ( assign_type )
              { case ASSIGN_:    break;
                case PLUSASSIGN_: v +=  oldvalue; break;
                case SUBASSIGN_: v = oldvalue - v; break;
                case MULTASSIGN_: v = oldvalue * v; break;
                case DIVASSIGN_: v = oldvalue / v; break;
              }
          set_body_volconst(id,v); 
          set_body_volume(id,get_body_volume(id)+v-oldvalue);
          set_body_oldvolume(id,get_body_oldvolume(id)+v-oldvalue);
          if ( everything_quantities_flag )
            GEN_QUANTS[get_body_volquant(id)].volconst = v;
           *recalc_flag = 1;
          break;
        }
      case SET_TARGET_: 
        { REAL v = *(stacktop--);
           REAL oldvalue = get_body_fixvol(id); 
           switch ( assign_type )
              { case ASSIGN_:    break;
                case PLUSASSIGN_: v +=  oldvalue; break;
                case SUBASSIGN_: v = oldvalue - v; break;
                case MULTASSIGN_: v = oldvalue * v; break;
                case DIVASSIGN_: v = oldvalue / v; break;
              }
           if ( get_attr(id) & PRESSURE )
              kb_error(2013,"Must unset body pressure before fixing target.\n",
                  RECOVERABLE);
          set_attr(id,FIXEDVOL);
          set_body_fixvol(id,v); 
          *recalc_flag = 1;
          break;
        }
      case SET_PRESSURE_: 
        { REAL p = *(stacktop--);
           REAL oldvalue = get_body_pressure(id); 
           switch ( assign_type )
              { case ASSIGN_:    break;
                case PLUSASSIGN_: p +=  oldvalue; break;
                case SUBASSIGN_: p = oldvalue - p; break;
                case MULTASSIGN_: p = oldvalue * p; break;
                case DIVASSIGN_: p = oldvalue / p; break;
              }
           if ( get_attr(id) & FIXEDVOL )
              kb_error(2013,"Must unset body target before fixing pressure.\n",
                  RECOVERABLE);
          set_attr(id,PRESSURE);
          set_body_pressure(id,p);    web.pressflag = 1;
          if ( everything_quantities_flag )
          { struct gen_quant *q = GEN_QUANTS + get_body_volquant(id);
            if ( p == 0.0 )
            { q->modulus = 1.0;
               if ( !(q->flags & Q_INFO) )
               { q->flags &= ~(Q_ENERGY|Q_FIXED);
                  q->flags |= Q_INFO;
                  reset_quant_bits();
               }
            }
            else 
            { q->modulus = -p;
               if ( !(q->flags & Q_ENERGY ) )
               { q->flags &= ~(Q_INFO|Q_FIXED);
                  q->flags |= Q_ENERGY;
                  reset_quant_bits();
               }
            }
          }
           *recalc_flag = 1;
          break;
        }
      case SET_OPACITY_:
          facet_alpha = *(stacktop)--;
          break;

      case SET_CONSTRAINT_:
        { int con = (int)*(stacktop--);
           struct constraint *constr;
           if ( (con<0) || (con>=MAXCON) || !web.constraint_addr[con])
           { sprintf(errmsg,"Illegal constraint number: %d.\n",con);
              kb_error(1212,errmsg,RECOVERABLE);
           } 
           set_attr(id,CONSTRAINT);
           constr = get_constraint(con); 
           if ( constr->attr & CON_ENERGY )
              set_attr(id, BDRY_ENERGY);
           if ( constr->attr & CON_CONTENT )
              set_attr(id, BDRY_CONTENT);
           switch ( id_type(id) )
              { case VERTEX: set_v_constraint_map(id,con); break;
                case EDGE: set_e_constraint_map(id,con); break;
                case FACET: set_f_constraint_map(id,con); break;
              }
           *recalc_flag = 1;
          }
        break;

      case SET_ORIGINAL_:
        set_original(id,(int)*(stacktop--));
        break;

      case SET_COLOR_:
        { int color = (int)*(stacktop--);
           switch ( id_type(id) )
              { case EDGE: set_edge_color(id,color); break;
                case FACET: set_facet_color(id,color); break;
              }
           *update_display_flag = 1;
        }
        break;

      case SET_FRONTCOLOR_:
        { int color = (int)*(stacktop--);
           set_facet_frontcolor(id,color);
           *update_display_flag = 1;
        }
        break;

      case SET_BACKCOLOR_:
        { int color = (int)*(stacktop--);
           set_facet_backcolor(id,color);
           *update_display_flag = 1;
        }
        break;

      case SET_WRAP_:
        { int w = (int)*(stacktop--);
           set_edge_wrap(id,w);
           *update_display_flag = 1;
        }
        break;

      case SET_TAG_:
        { int tag = (int)*(stacktop--);
           switch ( id_type(id) )
              { 
                case FACET: set_tag(id,tag); break;
              }
        }
        break;

       case SET_COORD_: case SET_COORD_1: 
          if ( node->right ) /* indexed */
          { int dim = (int)*(stacktop--)-1;
            REAL oldvalue = get_coord(id)[dim]; 
            REAL v;
            if ( (dim >= SDIM) || (dim < 0) )
                { sprintf(errmsg,
                   "Trying to set coordinate %d, in space dimension %d.\n",
                        dim,SDIM);
                   kb_error(2333,errmsg,RECOVERABLE);
                }
            v = *(stacktop--);
            switch ( assign_type )
              { case ASSIGN_:  break;
                case PLUSASSIGN_: v +=  oldvalue; break;
                case SUBASSIGN_: v = oldvalue - v; break;
                case MULTASSIGN_: v = oldvalue * v; break;
                case DIVASSIGN_: v = oldvalue / v; break;
              }
              get_coord(id)[dim] = v;
              *recalc_flag = 1;
              break;
           }
           /* else fall through */

       case SET_COORD_2:
       case SET_COORD_3: case SET_COORD_4: case SET_COORD_5:
       case SET_COORD_6: case SET_COORD_7: case SET_COORD_8:
          { REAL v = *(stacktop--);
            REAL oldvalue = get_coord(id)[node->op2.intval-SET_COORD_1];
            switch ( assign_type )
              { case ASSIGN_:  break;
                case PLUSASSIGN_: v +=  oldvalue; break;
                case SUBASSIGN_: v = oldvalue - v; break;
                case MULTASSIGN_: v = oldvalue * v; break;
                case DIVASSIGN_: v = oldvalue / v; break;
              }
            get_coord(id)[node->op2.intval-SET_COORD_1] = v;
            if ( node->op2.intval-SET_COORD_1 <= SDIM ) *recalc_flag = 1;
          }
          break;

       case SET_PARAM_: case SET_PARAM_1: case SET_PARAM_2:
       case SET_PARAM_3: case SET_PARAM_4:
          if ( get_vattr(id) & BOUNDARY )
            { struct boundary *boundary = get_boundary(id);
               REAL *param = get_param(id);
               REAL *xx = get_coord(id);
               REAL v;
               int pnum = node->op2.intval-SET_PARAM_1; 
               REAL oldvalue = param[pnum];
               if ( node->right ) pnum += (int)*(stacktop--);
               if ( pnum > boundary->pcount )
               { sprintf(errmsg,"Parameter number is %d; maximum is %d.\n",
                       pnum,boundary->pcount);
                  kb_error(1213,errmsg,RECOVERABLE);
               }
               v = *(stacktop--);
               switch ( assign_type )
               { case ASSIGN_:  break;
                 case PLUSASSIGN_: v +=  oldvalue; break;
                 case SUBASSIGN_: v = oldvalue - v; break;
                 case MULTASSIGN_: v = oldvalue * v; break;
                 case DIVASSIGN_: v = oldvalue / v; break;
               }
               param[pnum] = v; 
               for ( k = 0 ; k < SDIM ; k++ )
                  xx[k] = eval(boundary->coordf[k],param,id);
               *recalc_flag = 1;
            }
           else /* just parking a value probably */ 
              { REAL *param = get_param(id);
                int pnum = node->op2.intval-SET_PARAM_1;
                REAL oldvalue = param[pnum];
                REAL v;
                if ( node->right ) pnum += (int)*(stacktop--);
               v = *(stacktop--);
               switch ( assign_type )
               { case ASSIGN_:  break;
                 case PLUSASSIGN_: v +=  oldvalue; break;
                 case SUBASSIGN_: v = oldvalue - v; break;
                 case MULTASSIGN_: v = oldvalue * v; break;
                 case DIVASSIGN_: v = oldvalue / v; break;
               }

               param[pnum] = v;
             }
             break;

          case SET_FRONTBODY_:
          case SET_BACKBODY_:
            b_id = get_ordinal_id(BODY,(int)*(stacktop--)-1);
            if ( !valid_id(b_id) ) 
               kb_error(1995,"Invalid body in SET FRONTBODY or BACKBODY.\n",RECOVERABLE);
            set_body((node->op2.intval==SET_FRONTBODY_?id:inverse_id(id)),b_id);
            *recalc_flag = 1;
            break;

         default: 
            sprintf(errmsg,"Unhandled SET attribute %d\n",node->op2.intval);
            kb_error(4344,errmsg,RECOVERABLE);
            break;

        } /* end assigned SET_ATTRIBUTE_A */
      }
      break;

      /******************************/
      /* initializing element loops */
      /******************************/

#define STACKSAVE /* for break, continue */  \
      if ( node->stackpos ) \
    node[node->stackpos].stackpos = stacktop - newstack;

      case INIT_VERTEX_: /* all vertices */
           *(element_id *)++stacktop = q_id; /* save old element */
           *(vertex_id*)++stacktop=web.skel[VERTEX].last; /* sentinel */
           *(element_id *)++stacktop = web.skel[VERTEX].used;
           STACKSAVE
           break;

      case INIT_EDGE_VERTEX_: /* edge endpoints */
           if ( node->op2.intval ) id = (node+node->op2.intval)->op2.id;
           else id = q_id;
           *(element_id *)++stacktop = id; /* save old element */
           *(int *)++stacktop = 0; /* element number */
           *(element_id **)++stacktop = get_edge_vertices(id);
           STACKSAVE
           break;

      case INIT_FACET_VERTEX_: /* facet vertices */
           if ( node->op2.intval ) id = (node+node->op2.intval)->op2.id;
           else id = q_id;
           *(element_id *)++stacktop = id; /* save old element */
           if ( (web.representation == SIMPLEX) ||
                   (web.modeltype==LAGRANGE) )
           { *(int *)++stacktop = 0 ; /* counter */
              *(int *)++stacktop = id ; /* facet */
           }
           else 
           { *(element_id *)++stacktop = get_facet_fe(id); /* start */
              *(element_id *)++stacktop = get_facet_fe(id); /* current */
           }
           STACKSAVE
           break;

      case INIT_BODY_VERTEX_: /* vertices on body */
           kb_error(1214,"Can't do body vertices yet.\n",RECOVERABLE);

           break;

      case INIT_EDGE_: /* all edges */
           *(element_id *)++stacktop = q_id; /* save old element */
           *(edge_id *)++stacktop = web.skel[EDGE].last; /* sentinel */
           *(element_id *)++stacktop = web.skel[EDGE].used;
           STACKSAVE
           break;

      case INIT_VERTEX_EDGE_:
           if ( vedge_timestamp < top_timestamp ) 
              { make_vedge_lists(); vedge_timestamp = top_timestamp; }
           if ( node->op2.intval ) id = (node+node->op2.intval)->op2.id;
           else id = q_id;
           *(element_id *)++stacktop = id; /* save old element */
           if ( get_vattr(id) & Q_MIDFACET )
           { *(element_id *)++stacktop = NULLID;
             *(element_id *)++stacktop = NULLID;
           }
           else
           { *(element_id *)++stacktop = get_vertex_edge(id);
             *(element_id *)++stacktop = get_vertex_edge(id);
           }
           STACKSAVE
           break;
           
      case INIT_FACET_EDGE_:
           if ( node->op2.intval ) id = (node+node->op2.intval)->op2.id;
           else id = q_id;
           *(element_id *)++stacktop = id; /* save old element */
           if ( web.representation == STRING )
              *(element_id *)++stacktop = get_facet_fe(id); /* start */
           else
              *(int *)++stacktop = 0; /* counter */
           *(element_id *)++stacktop = get_facet_fe(id);
           STACKSAVE
           break;
           
      case INIT_BODY_EDGE_:
           kb_error(1215,"Can't do body edges yet.\n",RECOVERABLE);
           break;

      case INIT_FACET_: /* all facets */
           *(element_id *)++stacktop = q_id; /* save old element */
           *(facet_id *)++stacktop = web.skel[FACET].last; /* sentinel */
           *(element_id *)++stacktop = web.skel[FACET].used;
           STACKSAVE
           break;

      case INIT_VERTEX_FACET_:
           if ( vfacet_timestamp < top_timestamp )  make_vfacet_lists(); 
           if ( node->op2.intval ) id = (node+node->op2.intval)->op2.id;
           else id = q_id;
           *(element_id *)++stacktop = id; /* save old element */
           if ( get_vattr(id) & Q_MIDPOINT )
           {
              *(element_id *)++stacktop = fe = get_vertex_fe(id); 
              f_id = get_fe_facet(fe);
           }
           else if ( get_vattr(id) & Q_MIDFACET )
           {
              *(element_id *)++stacktop = f_id = get_vertex_edge(id); 
           }
           else /* regular vertex */
           { f_id = get_first_vertex_facet(id);
              *(element_id *)++stacktop = f_id;    /* sentinel */
           }
           *(element_id *)++stacktop = f_id; 
           STACKSAVE
           break;

      case INIT_EDGE_FACET_:
           if ( node->op2.intval ) id = (node+node->op2.intval)->op2.id;
           else id = q_id;
           *(element_id *)++stacktop = id; /* save old element */
           *(element_id *)++stacktop = get_edge_fe(id); /* start */
           *(element_id *)++stacktop = get_edge_fe(id);
           STACKSAVE
           break;

      case INIT_BODY_FACET_:
           if ( bfacet_timestamp < top_timestamp ) 
              { make_bfacet_lists(); bfacet_timestamp = top_timestamp; }
           if ( node->op2.intval ) id = (node+node->op2.intval)->op2.id;
           else id = q_id;
           *(element_id *)++stacktop = q_id; /* save old element */
           *(element_id *)++stacktop = get_fe_facet(get_body_fe(id));
           *(element_id *)++stacktop = get_fe_facet(get_body_fe(id));
           STACKSAVE
           break;

      case INIT_BODY_: /* all bodies */
           *(element_id *)++stacktop = q_id; /* save old element */
           *(body_id *)++stacktop = web.skel[BODY].last; /* sentinel */
           *(element_id *)++stacktop = web.skel[BODY].used;
           STACKSAVE
           break;

      case INIT_FACET_BODY_: 
           if ( node->op2.intval ) id = (node+node->op2.intval)->op2.id;
           else id = q_id;
           *(element_id *)++stacktop = id; /* save old element */
           *(int *)++stacktop = 0; /* element number */
           *(element_id *)++stacktop = NULLID;
           STACKSAVE
           break;

      case INIT_VERTEX_BODY_:
           kb_error(1216,"Can't do vertex bodies yet.\n",RECOVERABLE);
           break;

      case INIT_EDGE_BODY_:
           kb_error(1217,"Can't do edge bodies yet.\n",RECOVERABLE);
           break;

      case INIT_FACETEDGE_: /* all edges */
           *(element_id *)++stacktop = q_id; /* save old element */
           *(edge_id *)++stacktop = web.skel[FACETEDGE].last; /* sentinel */
           *(element_id *)++stacktop = web.skel[FACETEDGE].used;
           STACKSAVE
           break;

      case SINGLE_ELEMENT_INIT_: 
           *(element_id *)++stacktop = q_id; /* save old element */
           *(++stacktop) = 0.0 ; /* iteration count */
           stacktop++;  /* dummy local to pop */
           STACKSAVE;
           break;


    /*******************/
    /* simple commands */
    /*******************/

    case NOP_:
    case NULLBLOCK_:
    case NULLCMD_:
        break;

    case DEFINE_:
        expand_attribute(node->op2.intval,node->op1.intval,node->op3.intval[0]);
        break;

    case HISTORY_:
        if ( history_space == NULL ) { outstring("No history.\n"); break;}
        for ( k = history_count-1 ; k >= 0 ; k-- )
          { h = history_space+history_list[k];
            sprintf(msg,"%d) ",history_number-k); outstring(msg);
            outstring(h);
          }
          break;

    case REDIRECT_:
    case REDIRECTOVER_:
        { struct stat st;
           int retval;
           if ( node->left ) s = *(char**)(stacktop--);
           else s = node->op1.string;
           *(FILE **)(++stacktop) = outfd;
           retval = stat(s,&st);
           outfd = NULL;
           if ( ((retval == 0) && (st.st_mode & S_IFIFO)) 
            || (node->type==REDIRECTOVER_) )
              outfd = fopen(s,"w");
           else outfd = fopen(s,"a");
           if ( outfd == NULL )
              { 
#ifdef MAC_APP
                sprintf(errmsg,"Cannot open file %s. \n",s);
#else
                perror(s);
                sprintf(errmsg,"Cannot open redirection file %s.\n",s);
#endif
                kb_error(1218,errmsg,RECOVERABLE);

              }
          }
           break;

    case REDIRECT_END_:
           fclose(outfd);
           /* outfd = stdout; */
           outfd = *(FILE**)(stacktop--);
           break;

    case PIPE_:
           if ( node->left ) s = *(char**)(stacktop--);
           else s = node->op1.string;
           *(FILE **)(++stacktop) = outfd;
           if ( s[0] == '+' )  /* kludge to permit appending to file */
           outfd = fopen(s+1,"a+");
#ifdef NOPIPE
           else
           outfd = fopen(s,"w");
#else
           outfd = popen(s,"w");
#endif
           if ( outfd == NULL )
              { 
                outfd = *(FILE**)(stacktop--);
#ifdef MAC_APP
                sprintf(errmsg,"Cannot open pipe %s. \n",s);
#else
                perror(s);
                sprintf(errmsg,"Cannot open pipe %s.\n",s);
#endif
                kb_error(1219,errmsg,RECOVERABLE);

              }
           broken_pipe_flag = 0; /* reset */
           break;
 
       case PIPE_END_:
#ifdef NOPIPE
           fclose(outfd);
#else
           pclose(outfd);
#endif
           /* outfd = stdout; */
           outfd = *(FILE**)(stacktop--);
           break;


       case SINGLE_REDEFD_:
          if ( single_redefine[node->op1.intval].start )
            eval(&single_redefine[node->op1.intval],NULL,NULLID);
          else letter_command(node->op1.intval);
          break;

       case SINGLE_LETTER_:
          letter_command(node->op1.intval);
          break;

       case DEBUG_:
          outstring(yydebug?"YACC debugging was ON.":
              "YACC debugging was OFF.\n");
          yydebug = (node->op1.intval==ON_) ? 1 : 0;
          outstring(yydebug?"Now ON.\n":"Now OFF.\n");
          break;

       case ITDEBUG_:
          outstring(itdebug?"Iteration debugging was ON.":
              "Iteration debugging was OFF.\n");
          itdebug = (node->op1.intval==ON_) ? 1 : 0;
          outstring(itdebug?"Now ON.\n":"Now OFF.\n");
          break;

       case MEMDEBUG_:
          old = memdebug;
          memdebug = (node->op1.intval==ON_) ? 1 : 0;
#ifdef M_DEBUG
    mallopt(M_DEBUG,memdebug);
#endif
          outstring(memdebug?"Memory debugging ON.":
                                   "Memory debugging OFF.");
          outstring(old?" (was on)\n":" (was off)\n");
          break;

       case RECALC_: recalc(); break;

       case CLOSE_SHOW_: close_graphics(); break;

       case TOPINFO_: top_dump(outfd); break;
       case BOTTOMINFO_: bottom_dump(outfd); break;

       case LIST_PROCS_:
          outstring("//Named procedures: \n");
          list_procedures();
          break;
    
       case LIST_ATTRIBUTES_:
          outstring("//Element attributes: \n");
          list_attributes();
          break;
    
       case DIRICHLET_:
       case DIRICHLET_SEEK_:
          scale = dirichlet(node->type==DIRICHLET_?0:1);
          recalc();
#ifdef LONGDOUBLE
          sprintf(msg,"%3d.  energy: %#*.*Lg  scale: %#g\n",1,
              DWIDTH,DPREC, web.total_energy,(DOUBLE)scale);
#else
          sprintf(msg,"%3d. %s: %#17.15g energy: %#17.15g  scale: %#g\n",1,
              areaname,web.total_area,web.total_energy,scale);
#endif 
          outstring(msg);
          break;

       case SOBOLEV_:
       case SOBOLEV_SEEK_:
          scale = sobolev(node->type==SOBOLEV_?0:1);
          recalc();
#ifdef LONGDOUBLE
          sprintf(msg,"%3d.  energy: %#*.*Lg  scale: %#g\n",1,
              DWIDTH,DPREC,web.total_energy,(DOUBLE)scale);
#else
          sprintf(msg,"%3d. %s: %#17.15g energy: %#17.15g  scale: %#g\n",1,
              areaname,web.total_area,web.total_energy,scale);
#endif 
          outstring(msg);
          break;

       case HESSIAN_SEEK_:
       { REAL maxscale; 
          if ( node->left ) maxscale = *(stacktop--);
          else maxscale = 1.;
          scale = hessian_seek(maxscale);
          recalc();
#ifdef LONGDOUBLE
          sprintf(msg,"%3d.  energy: %#*.*Lg  scale: %#g\n",1,
              DWIDTH,DPREC,web.total_energy,(DOUBLE)scale);
#else
          sprintf(msg,"%3d. %s: %#17.15g energy: %#17.15g  scale: %#g\n",1,
              areaname,web.total_area,web.total_energy,scale);
#endif 
          outstring(msg);
        }
        break;

       case HESSIAN_SADDLE_:
       { REAL maxscale; 
          if ( node->left ) maxscale = *(stacktop--);
          else maxscale = 100.;
          hessian_saddle(maxscale);
       }
       break;

       case HESSIAN_MENU_: 
           hessian_menu();
           recalc();
#ifdef LONGDOUBLE
          sprintf(msg,"%3d.  energy: %#*.*Lg \n",1, 
               DWIDTH,DPREC,web.total_energy);
#else
          sprintf(msg,"%3d. %s: %#17.15g energy: %#17.15g \n",1,
              areaname,web.total_area,web.total_energy);
#endif 
           outstring(msg);
           break;

       case HESSIAN_ : 
           hessian_auto();
           recalc();
#ifdef LONGDOUBLE
          sprintf(msg,"%3d.  energy: %#*.*Lg  \n",1, DWIDTH,
               DPREC,web.total_energy);
#else
          sprintf(msg,"%3d. %s: %#17.15g energy: %#17.15g\n",1,
              areaname,web.total_area,web.total_energy);
#endif 
           outstring(msg);
           break;

       case SHELL_:  /* execute subshell */
#if defined(__WIN32__) || defined(_WIN32)
       spawnlp(P_WAIT,"command",NULL);
#else
#if defined(MAC_APP) || defined(MAC_CW)
       kb_error(1221,"No subshell on Mac.\n",RECOVERABLE);

#else
       if ( fork() == 0 )
        { execlp("sh",NULL);
           perror("sh");
        }
       else wait(NULL);
#endif
#endif
          break;

    case PROCEDURE_:
          eval(&globals[node->op1.intval].value.proc,NULL,NULLID);
          break;

    case CMDLIST_: case COMMAND_BLOCK_:
          /* no action, just holds tree together */
          break;
    

       case SHOW_TRANS_:
           view_transform(*(char**)(stacktop--));
           break;

       case CHECK_:
          run_checks();
          break;

       case CONVERT_TO_QUANTS_:
          if ( everything_quantities_flag )
            outstring("Everything already quantities.\n");
          else
          { convert_to_quantities();
            recalc();
          }
          break;

       case SHOW_VOL_:
          show_volumes();
          break;

     case LOGFILE_:
          start_logfile(*(char**)(stacktop--));
          outstring("Logfile ON.\n");
          break;

     case GEOMVIEW_:
          geomview_command(*(char**)(stacktop--));
          break;

     case IS_DEFINED_:
         /* replace string pointer with parse-time evaluation */
         *stacktop = (REAL)node->op1.intval;
         break;

     case SYSTEM_:
       system(*(char**)(stacktop--));
       break;

     case CHDIR_:
#ifdef MSC
       k = _chdir(*(char**)(stacktop--));
       if ( k < 0 )
       { kb_error(3222,"Unable to change directory. \n",WARNING);
         if ( !strchr(*(char**)(stacktop+1),'\\') )
           outstring("Try using \\\\ or / instead of \\.\n");
       }
#else
       k = chdir(*(char**)(stacktop--));
       if ( k < 0 )
         kb_error(3222,"Unable to change directory.\n",WARNING);
#endif

       break;

     case READ_:
       exec_file(NULL,*(char**)(stacktop--));
       break;

     case VIEW_TRANSFORMS_:
       read_transforms((int)(*stacktop--));
       break;

     case TRANSFORM_DEPTH_:
       generate_transforms((int)(*stacktop--));
       update_display();
       break;

     case TRANSFORM_EXPR_:
       transform_gen_expr(*(char**)(stacktop--));
       transforms_flag = 1;
       update_display();
       break;

     case GRAVITY_:
       old = web.gravflag;
       web.gravflag = (node->op1.intval==ON_) ? 1 : 0;
       outstring(web.gravflag ? "Gravity ON." : "Gravity OFF.");
       outstring(old?" (was on)\n":" (was off)\n");
       recalc();
       break;

     case LINEAR_:
          if ( web.modeltype == QUADRATIC )
          { quad_to_linear(); recalc(); break; }
          else if ( web.modeltype == LAGRANGE )
            lagrange_to_linear();
          break;

     case QUADRATIC_:
          if ( web.modeltype == LINEAR )
          { linear_to_quad(); recalc(); break; }
          else if ( web.modeltype == LAGRANGE )
            lagrange_to_quad();
          break;

     case LAGRANGE_:
          if ( web.modeltype == LINEAR )
            linear_to_lagrange((int)(*stacktop--)); 
          else if ( web.modeltype == QUADRATIC )
            quad_to_lagrange((int)(*stacktop--));
          else
            lagrange_to_lagrange((int)(*stacktop--));
          break;

     case TOGGLEVALUE:
       *++stacktop = (REAL)get_toggle_value(node->op1.intval);
       break;

     case FACET_COLORS_:
       old = color_flag;
       color_flag = (node->op1.intval==ON_) ? 1 : 0; 
       outstring(color_flag ? "Facet colors ON.":
                                        "Facet colors OFF.");
          outstring(old?" (was on)\n":" (was off)\n");
       break;

     case SHADING_:
       old = shading_flag;
       shading_flag = (node->op1.intval==ON_) ? 1 : 0; 
       outstring(shading_flag ? "Facet shading ON.":
                                        "Facet shading OFF.");
          outstring(old?" (was on)\n":" (was off)\n");
       break;

       case PREPRINTF_:
       case PRESPRINTF_:
           break;
       case GO_:
          iterate();
          break;

       default:
          stacktop = more_other_stuff(node,stacktop,recalc_flag,
              update_display_flag,q_id);
          break;
  }
  return stacktop;
} /* end other_stuff() */

/******************************************************************
*
* Function: flip_toggle()
*
* Purpose: Utility function to flip a toggle. Put here so Mac 68K
*     can make a short jump to it within this file.
*/

void flip_toggle(flag,newstate,phrase)
int *flag;  /* toggle variable */
int newstate;
char *phrase; /* to print */
{ int old;
  old = *flag;
  *flag = (newstate==ON_) ? 1 : 0;
  outstring(phrase);
  outstring(*flag ? " ON." : " OFF.");
  if ( old < 0 ) outstring(" (was unset)\n");
  else if ( old == 0 ) outstring(" (was off)\n");
  else outstring(" (was on)\n");
}

/************************************************************************
*
* Function: more_other_stuff()
*
* Purpose: execute command nodes, overflow from eval() and other_stuff().
*
* Return: stacktop pointer
*
*/

REAL * more_other_stuff(node,stacktop,recalc_flag,update_display_flag,q_id)
struct treenode * node;
REAL *stacktop;  /* current */
int *recalc_flag;
int *update_display_flag;
element_id q_id;
{ 
  char response[100];
  int old; /* old state of toggle */
  int n,k,i;
  char *s;

  switch ( node->type )
  {
    case WRAP_VERTEX_:
         { vertex_id v_id = get_ordinal_id(VERTEX,(int)(stacktop[-1])-1);
           int wrap = (element_id)(stacktop[0]);
           REAL newx[MAXCOORD];
           REAL *x = get_coord(v_id);
           edge_id e_id,start_e;

           stacktop -= 2;
           sym_wrap(x,newx,wrap);
           for ( i = 0 ; i < SDIM ; i++ ) x[i] = newx[i];
           if ( vedge_timestamp < top_timestamp ) make_vedge_lists();
           start_e = e_id = get_vertex_edge(v_id);
           do
           { set_edge_wrap(e_id,sym_compose(wrap,get_edge_wrap(e_id)));
             e_id = get_next_tail_edge(e_id);
           } while ( !equal_id(start_e,e_id) );
           *update_display_flag = 1;
         }
         break;

    case NORMAL_MOTION_:
         old = normal_motion_flag;
         normal_motion_flag = (node->op1.intval==ON_) ? 1 : 0; 
         outstring(normal_motion_flag ? "Normal motion ON.":
                                          "Normal motion OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         if ( normal_motion_flag )
             begin_normal_motion();
         else end_normal_motion();
         break;

    case VIEW_4D_: 
         old = view_4D_flag;
         view_4D_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(view_4D_flag ? "4D graphics  ON." :
                                            "4D graphics  OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         update_display();
         break;
     
    case PINNING_: 
         old = check_pinning_flag;
         check_pinning_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(check_pinning_flag ? "Constraint pinning ON." :
                                            "Constraint pinning OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;
     
    case METRIC_CONVERSION_: 
         old = metric_convert_flag;
         metric_convert_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(metric_convert_flag ? "Metric conversion ON." :
                                            "Metric conversion OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;
     

    case SELF_SIMILAR_: 
         old = self_similar_flag;
         self_similar_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(self_similar_flag ? "Self similarity mode ON." :
                                            "Self similarity mode OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case GV_BINARY_: 
         old = gv_binary_flag;
         gv_binary_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(gv_binary_flag ? "Geomview binary mode ON." :
                                            "Geomview binary mode OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;
     
    case KUSNER_: 
         old = kusner_flag;
         kusner_flag = (node->op1.intval==ON_) ? 1 : 0;
         conf_edge_curv_flag = 0;
         outstring(kusner_flag ? "Edge square curvature ON." :
                                            "Edge square curvature OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         recalc();
         break;
     
    case BOUNDARY_CURVATURE_: 
         old = boundary_curvature_flag;
         boundary_curvature_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(boundary_curvature_flag ? "Boundary curvature ON." :
                                            "Boundary curvature OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         recalc();
         break;
     
    case CONF_EDGE_SQCURV_: 
         old = conf_edge_curv_flag;
         conf_edge_curv_flag = (node->op1.intval==ON_) ? 1 : 0;
         kusner_flag = 0;
         outstring(conf_edge_curv_flag ? 
                "Conformal edge square curvature ON." :
                "Conformal edge square curvature OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         recalc();
         break;

    case POST_PROJECT_: 
         old = post_project_flag;
         post_project_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(post_project_flag ? "Post-projection ON." :
                "Post-projection OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case MEAN_CURV_INT_: 
         old = mean_curv_int_flag;
         mean_curv_int_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(mean_curv_int_flag ? "Mean curvature integral ON." :
                "Mean curvature integral OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case AUTORECALC_: 
         old = autorecalc_flag;
         autorecalc_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(autorecalc_flag ? "Autorecalc ON." : "Autorecalc OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case FORCE_POS_DEF_: 
         old = make_pos_def_flag;
         make_pos_def_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(make_pos_def_flag ? "Force positive definite ON." : 
                "Force positive definite OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case ESTIMATE_: 
         old = estimate_flag;
         estimate_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(estimate_flag ? "Estimation ON." : "Estimation OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case TRANSFORMS_: 
         old = transforms_flag;
         transforms_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(transforms_flag ? "Transform showing ON." : 
                                    "Transform showing OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         update_display();
         break;

    case DETURCK_:
         old = unit_normal_flag;
         unit_normal_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(sqgauss_flag ? "Unit normal motion ON." :
                                            "Unit normal motion OFF.");
         if ( unit_normal_flag )
              { sprintf(msg,"Enter unit normal weight factor(%g): ",
                         (DOUBLE)deturck_factor);
                 prompt(msg,response,sizeof(response));
                 const_expr(response,&deturck_factor);
              }
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case SQGAUSS_: 
         old = sqgauss_flag;
         sqgauss_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(sqgauss_flag ? "Squared Gaussian curvature ON." :
                                            "Squared Gaussian curvature OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         recalc();
         break;

    case STABILITY_TEST_: 
         stability_test();
         break;

    case AUTOPOP_: 
         old = autopop_flag;
         autopop_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(autopop_flag ? "Autopopping ON.\n" : "Autopopping OFF.\n");
         outstring(old?" (was on)\n":" (was off)\n");
         if ( autopop_flag )
            { n = 0; 
              if ( web.representation == STRING )
                 sprintf(msg,"Number of vertices popped: %d\n",
                    web.pop_count = n=verpop_str());
              else
                 sprintf(msg,"Number of vertices popped: %d\n",
                            web.pop_count = n = edgepop_film());
              outstring(msg);
              if ( n > 0 ) recalc();
            }
         break;

    case AUTOCHOP_:
         old = autochop_flag;
         autochop_flag = (node->op1.intval==ON_) ? 1 : 0;
         if ( autochop_flag )
              { 
                 sprintf(msg,"Autochopping ON. Chop length %g ",(DOUBLE)autochop_size);
                 outstring(msg);
                 outstring(old?" (was on)\n":" (was off)\n");
                 outstring("Set autochop length with  AUTOCHOP := value\n");
              }
         else 
         { outstring("Autochopping OFF.");
           outstring(old?" (was on)\n":" (was off)\n");
         }
         break;

    case UTEST_: 
         simplex_delaunay_test();
         break;

    case OLD_AREA_: 
         old = old_area_flag;
         old_area_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(old_area_flag ? "old_area ON." : "old_area OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case APPROX_CURV_: 
         old = approx_curve_flag;
         approx_curve_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(approx_curve_flag ? "approx_curvature ON." :
                                            "approx_curvature OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case H_INVERSE_METRIC_: 
         old = web.h_inverse_metric_flag;
         web.h_inverse_metric_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(web.h_inverse_metric_flag ? "h_inverse_metric ON." :
                                            "h_inverse_metric OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case ASSUME_ORIENTED_: 
         old = assume_oriented_flag;
         assume_oriented_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(assume_oriented_flag ? "assume_oriented ON." :
                                            "assume_oriented OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case POSTSCRIPT_:
         if ( ps_colorflag < 0 )  ps_colorflag = 0 ;
         if ( gridflag < 0 ) gridflag = 0;
         if ( crossingflag < 0 ) crossingflag = 0;
         if ( labelflag < 0 ) labelflag = 0;
         if ( torus_display_mode == TORUS_DEFAULT_MODE ) 
         { torus_display_mode = TORUS_RAW_MODE;
            web.torus_body_flag = 0; 
            web.torus_clip_flag = 0; 
         }
         strncpy(ps_file_name,*(char**)(stacktop--),sizeof(ps_file_name));
         do_gfile('3',ps_file_name);
         break;

    case GEOMPIPE_:  /* to command */
         old = geompipe_flag;
         do_gfile('C',*(char**)(stacktop--));
         break;

    case GEOMPIPE_TOGGLE_: 
         old = geompipe_flag;
         do_gfile((node->op1.intval==ON_) ? 'A' : 'B',NULL);
         outstring(geompipe_flag ? "geompipe ON." :
                                            "geompipe OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case GEOMVIEW_TOGGLE_: 
         old = geomview_flag;
         if ( !geomview_flag && (node->op1.intval==ON_) )
             do_gfile( '8',NULL);
         else if ( geomview_flag && !(node->op1.intval==ON_) )
             do_gfile( '9',NULL);
         outstring(geomview_flag ? "geomview ON." :
                                            "geomview OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case LOGFILE_TOGGLE_:
         old = logfile_flag;
         if ( !logfile_flag && (node->op1.intval==ON_) )
             start_logfile(NULL);
         else if ( logfile_flag && (node->op1.intval!=ON_) )
             stop_logfile();
         outstring(logfile_flag ? "Logfile ON." :
                                            "Logfile OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case JIGGLE_TOGGLE_: 
         old = web.jiggle_flag;
         web.jiggle_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(web.jiggle_flag ? "jiggling ON." :
                                            "jiggling OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case RIBIERE_CG_: 
         old = ribiere_flag;
         ribiere_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(ribiere_flag ? "Polak-Ribiere conjugate gradient ON." :
                                            "Polak-Ribiere conjugate gradient OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         if ( ribiere_flag )
         { int r_attr;
            vertex_id v_id;
            if ( find_attribute(VERTEX,RIBIERE_ATTR_NAME) == -1 )
              add_attribute(VERTEX,RIBIERE_ATTR_NAME,REAL_ATTR,SDIM,0,NULL);
            r_attr = find_attribute(VERTEX,RIBIERE_ATTR_NAME);
            FOR_ALL_VERTICES(v_id)
            { REAL *g = (REAL*)get_extra(v_id,r_attr);
              for ( i = 0 ; i < SDIM ; i++ ) g[i] = 0.0;
            }
         }
         break;

    case CONJ_GRAD_:
         old = conj_grad_flag;
         conj_grad_flag = (node->op1.intval==ON_) ? 1 : 0;
         if ( conj_grad_flag )
          { conj_grad_flag = 1;
             reset_conj_grad();
             outstring("Conjugate gradient now ON.");
             outstring(old?" (was on)\n":" (was off)\n");
             if ( web.motion_flag )
                     kb_error(1222,"Fixed scale is ON! Not a good idea with conjugate gradient.\n",WARNING);
          }
         else
            { if ( cg_hvector ) myfree((char *)cg_hvector);
              cg_hvector = NULL;
              cg_oldsum = 0.0;
              conj_grad_flag = 0;
              outstring("Conjugate gradient now OFF.");
              outstring(old?" (was on)\n":" (was off)\n");
            }
         break;

     case MEAN_CURV_:
         web.norm_check_flag = 0; /* default OFF */
         old = web.area_norm_flag;
         web.area_norm_flag = (node->op1.intval==ON_) ? 1 : 0;
         if ( web.area_norm_flag ) 
            { outstring("Area normalization ON.");
            }
         else
            { outstring("Area normalization OFF.");
            }
         outstring(old?" (was on)\n":" (was off)\n");
         calc_energy();  /* to make sure vertex areas set */
         break;


    case SHOW_ALL_QUANTITIES_: 
          flip_toggle(&show_all_quantities,node->op1.intval,"show_all_quantities");
          break;
    case PSCOLORFLAG_: 
          flip_toggle(&ps_colorflag,node->op1.intval,"pscolorflag");
          break;
    case GRIDFLAG_:
          flip_toggle(&gridflag,node->op1.intval,"gridflag");
          break;
    case CROSSINGFLAG_: 
          flip_toggle(&crossingflag,node->op1.intval,"crossingflag");
          break;
    case LABELFLAG_:
          flip_toggle(&labelflag,node->op1.intval,"labelflag");
          break;
    case TORUS_FILLED_:
          if ( !web.torus_flag )
          { kb_error(3431,"torus_filled is invalid toggle because not in torus model.\n",
              WARNING);
            break;
          }
          flip_toggle(&web.full_flag,node->op1.intval,"torus_filled");
          break;

    case VERBOSE_:
         old = verbose_flag;
         verbose_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(verbose_flag ? "Verbose ON." : "Verbose OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case QUIET_:
         old = quiet_flag;
         quiet_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(quiet_flag ? "Quiet ON." : "Quiet OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case BACKCULL_:
         old = backcull_flag;
         backcull_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(backcull_flag ? "backcull ON." : "backcull OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         update_display();
         break;

    case VOLGRADS_EVERY_:
         old = volgrads_every_flag;
         volgrads_every_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(volgrads_every_flag ? "Volgrads_every ON." : "Volgrads_every OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case ZENER_DRAG_:
         old = zener_drag_flag;
         zener_drag_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(zener_drag_flag ? "Zener_drag ON." : "Zener_drag OFF.");
         outstring(old?" (was on)":" (was off)");
         { int zd = lookup_global(ZENER_COEFF_NAME);
            if ( zd >= 0 ) 
              sprintf (msg,"  (%s: %g)\n",ZENER_COEFF_NAME,(double)globals[zd].value.real);
            else sprintf(msg,"  (%s not set)\n",ZENER_COEFF_NAME);
            outstring(msg);
         }
         break;

    case QUIETGO_:
         old = quiet_go_flag;
         quiet_go_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(quiet_go_flag ? "QuietGo ON." : "QuietGo OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case HESSIAN_NORMAL_:
         old = hessian_normal_flag;
         hessian_normal_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(hessian_normal_flag ? "hessian_normal  ON." :
             "hessian_normal OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case HESSIAN_SPECIAL_NORMAL_:
         old = hessian_special_normal_flag;
         hessian_special_normal_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(hessian_special_normal_flag ? "hessian_special_normal ON." :
             "hessian_special_normal OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case HESSIAN_NORMAL_PERP_:
         old = hessian_normal_perp_flag;
         hessian_normal_perp_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(hessian_normal_perp_flag ? "hessian_normal_perp  ON." :
             "hessian_normal_perp OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;


    case HESSIAN_DOUBLE_NORMAL_:
         old = hessian_double_normal_flag;
         hessian_double_normal_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(hessian_double_normal_flag ? "hessian_double_normal  ON." :
             "hessian_double_normal OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case HESSIAN_NORMAL_ONE_:
         if ( node->op1.intval==ON_ && 
         ((web.representation == SIMPLEX) || (SDIM - web.dimension > 1)))
            kb_error(1923,"HESSIAN_NORMAL_ONE only for STRING or SOAPFILM hypersurfaces.\n",RECOVERABLE);

         old = hessian_normal_one_flag;
         hessian_normal_one_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(hessian_normal_one_flag ? "hessian_normal_one  ON." :
             "hessian_normal_one OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case HESSIAN_QUIET_:
         old = hessian_quiet_flag;
         hessian_quiet_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(hessian_quiet_flag ? "hessian_quiet  ON." :
             "hessian_quiet OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case HESSIAN_DIFF_:
         old = hessian_by_diff_flag;
         if ( !web.pressure_flag && count_fixed_vol() )
            kb_error(1223,"Hessian_diff not implemented for constrained quants.\n",
              RECOVERABLE);
         hessian_by_diff_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(hessian_by_diff_flag ? "Hessian by differences ON." :
                                            "Hessian by differences OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case INTERP_BDRY_PARAM_:
         old = interp_bdry_param;
         interp_bdry_param = (node->op1.intval==ON_) ? 1 : 0;
         outstring(interp_bdry_param ? "interpolation of boundary parameters  ON." :
             "interpolation of boundary parameters OFF (using extrapolation).");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case RITZ_:
      { int krydim;  /* Krylov subspace dimension */
         krydim = (int)*(stacktop--);
         ritz_command(*(stacktop--),krydim);
      }
      break;

    
    case LANCZOS_:
      { int krydim;  /* Krylov subspace dimension */
         if ( node->right ) krydim = (int)*(stacktop--);
         else krydim = 100;
         lanczos_command(*(stacktop--),krydim);
      }
      break;

    case MOVE_:
         calc_volgrads(DO_OPTS);
         move_vertices(ACTUAL_MOVE,*(stacktop--));
         vgrad_end();
#ifdef LONGDOUBLE
         sprintf(msg,"1.  energy: %*.*Lg  stepsize: %g\n",DWIDTH,DPREC,
            web.total_energy,(DOUBLE)stacktop[1]);
#else
         sprintf(msg,"1. %s: %17.15g energy: %17.15g  stepsize: %g\n",
            areaname,web.total_area,web.total_energy,stacktop[1]);
#endif 
         update_display();
         break;

    case EIGENPROBE_:
      { int iters;
         if ( node->right ) iters = (int)*(stacktop--);
         else iters = 0;
         eigenprobe_command(*(stacktop--),iters);
      }
      break;


    case EXPRLIST_:
             break;  /* leave expression on stack */
 
    case GET_TRANSFORM_EXPR_: /* put ptr in both halves */
       { int pp = (sizeof(REAL))/sizeof(char*);
         int nn;
         ++stacktop;
         for ( nn = 0 ; nn < pp  ; nn++ ) 
                ((char **)stacktop)[nn] = transform_expr;
       }
       break;

    case DATAFILENAME_: /* put ptr in both halves */
       { int pp = (sizeof(REAL))/sizeof(char*);
         int nn;
         ++stacktop;
         for ( nn = 0 ; nn < pp  ; nn++ ) 
               ((char **)stacktop)[nn] = datafilename;
       }
       break;

    case QUOTATION_: /* put ptr in both halves */
       { int pp = (sizeof(REAL))/sizeof(char*);
         int nn;
         *++stacktop = 0.0;
         for ( nn = 0 ; nn < pp  ; nn++ ) 
                 ((char **)stacktop)[nn] = node->op1.string;
       }
       break;

    case PRINTFHEAD_:
          if ( node->op1.string ) s = node->op1.string;
          else
              s = *(char**)(stacktop--);
          outstring(s);
          break;

     case SPRINTFHEAD_:
          if ( node->op1.string )
             *(char **)(++stacktop) = node->op1.string;
          /* else already on stack */
          break;

     case PRINTF_:
     case SPRINTF_:
     { char format[1000];
       n = node[node->right].op1.intval;
       if ( node[node->left].op1.string ) 
           s = node[node->left].op1.string;
       else
           s = *(char**)(stacktop-n);
       /* remove old string munges */
       { char *p,*c;
          /* strip %0.0s from format string */
          while ( (c = strstr(s,"%0.0s")) != NULL )
          {  for ( p = c+5 ; *p ; c++,p++ ) *c = *p;
              *c = 0;
          }
          while ( (c = strstr(s,"%.0s")) != NULL )
          {  for ( p = c+4 ; *p ; c++,p++ ) *c = *p;
              *c = 0;
          }
       }
       /* tinker with format string */
       { /* search for %... */
          char *sp = s;
          char *f = format;
          int pp;
          while ( *sp )
          { if ( *sp != '%' ) {*(f++) = *(sp++); continue; }
             while ( *sp && !isalpha(*sp) ) *(f++) = *(sp++);
             switch ( *sp )
             { case 's':
                 *(f++) = *(sp++);
                 /* assuming REAL sizes on stack rounded up to multiple
                     of pointer sizes */
                 pp = (sizeof(REAL) + sizeof(char*) - 1)/sizeof(char*);
                 for ( k = 1 ; k < pp ; k++ )  
                   { strcpy(f,"%0.0s"); f += strlen(f); }
                 break;
               case 'd': kb_error(7621,
     "Integer format %d is improper; use %f or %g since Evolver values are floating point.\n",WARNING); break;
#ifdef LONGDOUBLE
               case 'f': case 'g': case 'e':
                   *(f++) = 'L'; 
                   *(f++) = *(sp++);
                   break;
#endif
               default: *(f++) = *(sp++);
             }
          }
          *f = 0;
          s = format;
       }
#ifdef LONGDOUBLE
/* stack cleaning for odd-length REALs */
sprintf(msg,"",0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);
#endif
       sprintf(msg, s,
          n >= 1 ? stacktop[1-n] : 0.0,
          n >= 2 ? stacktop[2-n] : 0.0,
          n >= 3 ? stacktop[3-n] : 0.0,
          n >= 4 ? stacktop[4-n] : 0.0,
          n >= 5 ? stacktop[5-n] : 0.0,
          n >= 6 ? stacktop[6-n] : 0.0,
          n >= 7 ? stacktop[7-n] : 0.0,
          n >= 8 ? stacktop[8-n] : 0.0,
          n >= 9 ? stacktop[9-n] : 0.0,
          n >= 10 ? stacktop[10-n] : 0.0);
       stacktop -= n; /* pop args */
       if ( node[node->left].op1.string == NULL ) stacktop--;
       if ( node->type == PRINTF_ )
       { int old_flag = quiet_flag;
          quiet_flag = 0;
          outstring(msg);
          quiet_flag = old_flag;
       }
       else
          { char *str = mycalloc(strlen(msg)+2,1);
             strcpy(str,msg);
             *(char **)(++stacktop) = str;
             }
     }
     break;

     case PRINT_: /* verb */
#ifdef LONGDOUBLE
          sprintf(msg,"%*.*Lg\n",DWIDTH,DPREC,*(stacktop--));
#else
          sprintf(msg,"%20.15g\n",*(stacktop--));
#endif 
          old = quiet_flag; quiet_flag = 0;
          outstring(msg);
          quiet_flag = old;
          break;

     case STRPRINT_: /* verb */
          sprintf(msg,"%s\n",*(char**)(stacktop--));
          outstring(msg);
          break;

     case PRINT_LETTER_:
         if ( single_redefine[node->op1.intval].start )
            outstring(print_express(&single_redefine[node->op1.intval],'X'));
         else { sprintf(msg,"%c",node->op1.intval); outstring(msg);}
         outstring("\n\n");
         break;

     case PRINT_PROCEDURE_:
         outstring(print_express(&globals[node->op1.intval].value.proc,
                              'X'));
         outstring("\n\n");
         break;

     case EPRINT_: /* print and pass on value inside expression */
#ifdef LONGDOUBLE
          sprintf(msg,"%*.*Lg\n",DWIDTH,DPREC,*stacktop);
#else
          sprintf(msg,"%20.15g\n",*stacktop);
#endif 
          outstring(msg);
          break;

     case SHOWQ_:
#ifdef NeXTapp
         NeXT_display();
#else
         display();
#endif
         break;

    case SET_THICKEN_ :
      thickness = *(stacktop--);
      break;

    case JIGGLE_:
      web.temperature = *(stacktop--);
      jiggle();
      recalc();
      break;
     
    case NOTCH_:
      if ( web.representation == SIMPLEX )
         kb_error(1224, "Notching not implemented for simplex representation.\n",
            RECOVERABLE);
      web.max_angle = *(stacktop--);
      if ( web.max_angle <= 0.0 )
         kb_error(1225,"Notching angle not positive.\n",RECOVERABLE);
      if ( web.representation == STRING )
        { sprintf(msg,"refine edge ee where max(ee.vertex,dihedral) > %f",
              (DOUBLE)web.max_angle);
          command(msg);
          web.notch_count = web.refine_count;
        }
      else 
        { sprintf(msg,"Number of edges notched: %d\n",
                          web.notch_count = ridge_notcher(web.max_angle));
          outstring(msg);
          recalc();
        }
      break;

    case SET_AUTOCHOP_:
      autochop_size = *(stacktop--);
      break;

    case SET_AMBIENT_PRESSURE_:
      web.pressure = *(stacktop--);
      if ( web.pressure > 0.00000001 )
         {
            if ( !web.full_flag && !valid_id(web.outside_body) )
                              add_outside();
            web.projection_flag = 0;
            web.pressure_flag = 1;
         }
      else
         {
            web.projection_flag = 1;
            web.pressure_flag = 0;
         }
      recalc();
      break;

    case SET_DIFFUSION_:
         web.diffusion_const = *(stacktop--);
         web.diffusion_flag = 1;
         break;

    case AREAWEED_:
         sprintf(msg,"Skinny triangles weeded: %d\n",
              web.delete_count = areaweed(*(stacktop--)));
         outstring(msg);
         energy_init = 0;
         recalc();
         break;

    case METIS_:
    case KMETIS_:
         metis_partition_dual((int)(*(stacktop--)),node->type);
         break;

    case OMETIS_:
         if ( node->left )
            metis_vertex_order((int)(*(stacktop--)));
         else metis_vertex_order(100);
         break;

         case EDGEWEED_:
         sprintf(msg,"Deleted edges: %d\n",
              web.delete_count = edgeweed(*(stacktop--)));
         outstring(msg);
         recalc();
         break;

         case TOGGLE_QUANTITY_:
         { struct quantity *q;
            if ( (node->op1.intval < 0) || (node->op1.intval >= QUANTMAX))
              kb_error(1226,"Illegal quantity number.\n",RECOVERABLE);

            q  = get_quant(node->op1.intval);
            if ( !(q->attr & IN_USE ) )
              kb_error(1227,"Illegal quantity number.\n",RECOVERABLE);

            if ( node->op2.intval == ON_ )
              q->attr |= QFIXED;
            else q->attr &= ~QFIXED;
         }
         reset_conj_grad();
         break;

    case SET_QUANTITY_:
         { struct quantity *q;
            if ( (node->op1.intval < 0) || (node->op1.intval >= QUANTMAX))
              kb_error(1228,"Illegal quantity number.\n",RECOVERABLE);

            q  = get_quant(node->op1.intval);
            if ( !(q->attr & IN_USE ) )
              kb_error(1229,"Illegal quantity number.\n",RECOVERABLE);

            q->target = *(stacktop--);
         }
         reset_conj_grad();
         break;

    case OPTIMIZE_:
         web.motion_flag = (node->op1.intval==ON_) ? 0 : 1;
         if ( web.motion_flag )
            sprintf(msg,"Scale fixed at %g.\n",(DOUBLE)web.scale);
         else sprintf(msg,"Scale optimizing with bound %g.\n",(DOUBLE)web.maxscale);
         outstring(msg);
         break;

    case SET_OPTIMIZE_:
         web.maxscale = *(stacktop--);
         web.motion_flag = 0;
         sprintf(msg,"Scale optimizing with bound %g.\n",(DOUBLE)web.maxscale);
         outstring(msg);
         break;

    case SET_SCALE_:
         web.scale = *(stacktop--);
         web.motion_flag = 1;
         sprintf(msg,"Scale fixed at %g.\n",(DOUBLE)web.scale);
         outstring(msg);
         break;

    case SET_GRAVITY_:
         switch ( node->op1.intval )
         { case ASSIGN_: web.grav_const = *(stacktop--); break;
            case PLUSASSIGN_: web.grav_const += *(stacktop--); break;
            case SUBASSIGN_: web.grav_const -= *(stacktop--); break;
            case MULTASSIGN_: web.grav_const *= *(stacktop--); break;
            case DIVASSIGN_: web.grav_const /= *(stacktop--); break;
         }
         old = web.gravflag;
         if ( web.grav_const != 0.0 )
         { web.gravflag = 1;
            sprintf(msg,"Gravity is now ON with gravitational constant %g.",
              (DOUBLE)web.grav_const);
         }
         else
         { web.gravflag = 0;
            sprintf(msg,"Gravity is now OFF.");
         }
         if (gravity_quantity_num >= 0 )
                          GEN_QUANTS[gravity_quantity_num].modulus =
                              web.gravflag ? web.grav_const : 0.0;
         outstring(msg);
         outstring(old?" (was on)\n":" (was off)\n");
         recalc();
         break;

    case SET_MODEL_:
         switch ( (int)*(stacktop--) )
         { 
            case LINEAR: 
                if ( web.modeltype == QUADRATIC )
                 { outstring("Changing to LINEAR model. (was QUADRATIC)\n");
                    quad_to_linear(); recalc(); }
                else if ( web.modeltype == LAGRANGE )
                 { outstring("Changing to LINEAR model. (was LAGRANGE)\n");
                    lagrange_to_linear(); recalc(); }
                 else outstring("Model already LINEAR.\n");
                 break;
            case QUADRATIC: 
                if ( web.modeltype == LINEAR )
                 { outstring("Changing to QUADRATIC model. (was LINEAR)\n");
                    linear_to_quad(); recalc(); }
                else if ( web.modeltype == LAGRANGE )
                 { outstring("Changing to QUADRATIC model. (was LAGRANGE)\n");
                    lagrange_to_quad(); recalc(); }
                else outstring("Model already QUADRATIC.\n");
                 break;
            default: 
                 if ( stacktop[1] > 2. )
                 { if ( web.modeltype == LINEAR )
                   { outstring("Changing to LAGRANGE model. (was LINEAR)\n");
                         linear_to_lagrange((int)stacktop[1]); recalc(); }
                   else if ( web.modeltype == QUADRATIC )
                    { outstring("Changing to LAGRANGE model. (was QUADRATIC)\n");
                         quad_to_lagrange((int)stacktop[1]); recalc(); }
                   else lagrange_to_lagrange((int)stacktop[1]);
                 }
                 else
              kb_error(1230,"Bad model choice. Legal: 1 (linear), 2 (quadratic), > 2 (Lagrange)\n",
                  RECOVERABLE);
         }
         recalc();
         break;

    case INVOKE_P_MENU_:
         display_file((int)*(stacktop--));
         break;

    case EDGEDIVIDE_:
         sprintf(msg,"New edges: %d\n",
             web.refine_count = articulate(*(stacktop--)));
         outstring(msg);
         recalc();
            break;

     case SET_SGLOBAL_:
       { struct global *g = globals + node->op1.intval;
         char *s;
         if ( g->value.string ) myfree(g->value.string);
         s = *(char**)(stacktop--);
         if ( s )
         { g->value.string = mycalloc(strlen(s)+1,sizeof(char));
            strcpy(g->value.string,s);
         }
         else 
         { g->value.string = mycalloc(1,sizeof(char));  /* empty */
         } 

         g->flags |= STRINGVAL;
       } break;

     case SET_DELTA_:
       { struct global *g = globals + node->op1.intval;
         g->delta = *(stacktop--);
         break;
       }

     case DEFINE_IDENT_: break;

     case PLUSASSIGN_:
     case SUBASSIGN_:
     case MULTASSIGN_:
     case DIVASSIGN_:
     case SET_GLOBAL_:
       { struct global *g = globals + node->op1.intval;
         REAL *x;

         if ( g->flags & STRINGVAL )
         { myfree(g->value.string); g->flags &= ~STRINGVAL; }

         if ( g->flags & ORDINARY_PARAM )
            x = &g->value.real;
         else if ( g->flags & QUANTITY_TARGET )
            x = &GEN_QUANTS[g->value.quant].target;
         else if ( g->flags & QUANTITY_MODULUS )
           { x = &GEN_QUANTS[g->value.quant].modulus;  
             web_timestamp = ++global_timestamp;
           }
         else if ( g->flags & METHOD_MODULUS )
           { x = &METH_INST[g->value.meth_inst].modulus;
             web_timestamp = ++global_timestamp;
           }
         else 
         { sprintf(errmsg, "Variable %s is read-only.\n",g->name);
            kb_error(1254,errmsg,WARNING);
            break;
         }
         switch(node->type)
         { 
            case SET_GLOBAL_: *x = *(stacktop--); break;
            case PLUSASSIGN_: *x += *(stacktop--); break;
            case SUBASSIGN_: *x -= *(stacktop--); break;
            case MULTASSIGN_: *x *= *(stacktop--); break;
            case DIVASSIGN_: *x /= *(stacktop--); break;
         }
         if ( g->flags & SURFACE_PARAMETER & autorecalc_flag )
            recalc();
       }
       break;

    case ZOOM_:
            if ( node->left ) /* have vertex number */
              { vertex_id v_id;
                 int vnum;
                 int found = 0;
                 web.zoom_radius = *(stacktop--);
                 vnum = (int)*(stacktop--);
                 FOR_ALL_VERTICES(v_id)
                  { if ( vnum == (ordinal(v_id)+1) )
                        { web.zoom_v = v_id; found = 1; break; }
                  }
                 if ( !found )
                  kb_error(1231,"Zoom vertex not found.\n",RECOVERABLE);

              }
            else if ( !valid_element(web.zoom_v) )
                 kb_error(1232,"Zoom vertex not found.\n",RECOVERABLE);

            zoom_vertex(web.zoom_v,web.zoom_radius);
            /* resize(); */
            recalc();
            break;

     case VIEW_MATRIX_LVALUE_: break;  /* just to hold indices */

     case SET_VIEW_MATRIX_:
      { int i = (int)(stacktop[-2]);
        int k = (int)(stacktop[-1]);
        REAL value = *stacktop;
        stacktop -= 3;
         if ( (k < 1) || (k > SDIM+1) || (i < 1) || (i > SDIM+1) )
         { sprintf(errmsg,
             "Illegal index: view_matrix[%d][%d] (must be 1 to %d)\n",i,k,SDIM);
            kb_error(2512,errmsg,RECOVERABLE);
         }
         i--;k--;  /* convert to 0 based indexing */
         switch ( node->op2.intval )
         { case ASSIGN_: view[i][k] = value; break;
            case PLUSASSIGN_: view[i][k] += value; break;
            case SUBASSIGN_: view[i][k] -= value; break;
            case MULTASSIGN_: view[i][k] *= value; break;
            case DIVASSIGN_: view[i][k] /= value; break;
         } 
         *update_display_flag = 1;
      }
      break;


     case SET_QTARGET_:
      { struct gen_quant *q =
             GEN_QUANTS + globals[node->op1.intval].value.quant;
         switch ( node->op2.intval )
         { case ASSIGN_: q->target = *(stacktop--); break;
            case PLUSASSIGN_: q->target += *(stacktop--); break;
            case SUBASSIGN_: q->target -= *(stacktop--); break;
            case MULTASSIGN_: q->target *= *(stacktop--); break;
            case DIVASSIGN_: q->target /= *(stacktop--); break;
         } 
         if ( valid_id(q->b_id) )
              set_body_fixvol(q->b_id,q->target);
      }
      break;

     case SET_QMODULUS_:
      { REAL v = GEN_QUANTS[globals[node->op1.intval].value.quant].modulus;
         switch ( node->op2.intval )
         { case ASSIGN_: v = *(stacktop--); break;
            case PLUSASSIGN_: v += *(stacktop--); break;
            case SUBASSIGN_: v -= *(stacktop--); break;
            case MULTASSIGN_: v *= *(stacktop--); break;
            case DIVASSIGN_: v /= *(stacktop--); break;
         } 
        GEN_QUANTS[globals[node->op1.intval].value.quant].modulus = v;
     }
     break;

     case SET_QTOLERANCE_:
      if ( *stacktop <= 0.0 )
         kb_error(2864,"Tolerance must be positive.\n",RECOVERABLE);
      { REAL v = GEN_QUANTS[globals[node->op1.intval].value.quant].tolerance;
         switch ( node->op2.intval )
         { case ASSIGN_: v = *(stacktop--); break;
            case PLUSASSIGN_: v += *(stacktop--); break;
            case SUBASSIGN_: v -= *(stacktop--); break;
            case MULTASSIGN_: v *= *(stacktop--); break;
            case DIVASSIGN_: v /= *(stacktop--); break;
         } 
        GEN_QUANTS[globals[node->op1.intval].value.quant].tolerance = v;
     }
     break;

     case SET_MMODULUS_:
      { REAL v = METH_INST[globals[node->op1.intval].value.meth_inst].modulus;
         switch ( node->op2.intval )
         { case ASSIGN_: v = *(stacktop--); break;
            case PLUSASSIGN_: v += *(stacktop--); break;
            case SUBASSIGN_: v -= *(stacktop--); break;
            case MULTASSIGN_: v *= *(stacktop--); break;
            case DIVASSIGN_: v /= *(stacktop--); break;
         } 
        METH_INST[globals[node->op1.intval].value.meth_inst].modulus = v;
     }
     break;

     case SET_QVOLCONST_:
      { struct gen_quant *q =
             GEN_QUANTS + globals[node->op1.intval].value.quant;
         switch ( node->op2.intval )
         { case ASSIGN_: q->volconst = *(stacktop--); break;
            case PLUSASSIGN_: q->volconst += *(stacktop--); break;
            case SUBASSIGN_: q->volconst -= *(stacktop--); break;
            case MULTASSIGN_: q->volconst *= *(stacktop--); break;
            case DIVASSIGN_: q->volconst /= *(stacktop--); break;
         } 
         if ( valid_id(q->b_id) )
              set_body_volconst(q->b_id,q->volconst);
      }
    break;

    case RUNGE_KUTTA_:
         old = runge_kutta_flag;
         runge_kutta_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(runge_kutta_flag ? "Runge-Kutta ON." :
                                            "Runge-Kutta OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case CHECK_INCREASE_: 
         old = check_increase_flag;
         check_increase_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(check_increase_flag?"Increase check ON.":
                          "Increase check OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         break;

    case HOMOTHETY_:
         old = web.homothety;
         web.homothety = (node->op1.intval==ON_) ? 1 : 0;
         if ( web.homothety && (web.skel[BODY].count == 0) )
          { web.homothety = 0;
             kb_error(1233,"Cannot do homothety without bodies.\n",RECOVERABLE);
          }
         sprintf(msg,"Homothety adjustment is %s.",
                          web.homothety ? "ON" : "OFF");
         outstring(msg);
         outstring(old?" (was on)\n":" (was off)\n");
         if ( web.homothety )
            { sprintf(msg,"Enter target size (%g): ",(DOUBLE)homothety_target);
              prompt(msg,response,sizeof(response));
              const_expr(response,&homothety_target);
            }
         break;

    case COUNTS_: /* report count of elements and status */
         memory_report();
         break;

    case EXTRAPOLATE_: extrapolate(); break;

    case DIFFUSION_: /* Set diffusion */
         web.diffusion_flag = (node->op1.intval==ON_) ? 1 : 0;
            break;

    case AUTODISPLAY_: 
      old = go_display_flag;
      go_display_flag = (node->op1.intval == ON_) ? 1 : 0;
      outstring(go_display_flag ? "Autodisplay ON." :
                                            "Autodisplay OFF.");
      outstring(old?" (was on)\n":" (was off)\n");
      if ( go_display_flag ) update_display();
      break;

    case SHOW_INNER_: 
      old = innerflag;
      innerflag = (node->op1.intval == ON_) ? 1 : 0;
      outstring(innerflag ? "show inner ON." : "show inner OFF.");
      outstring(old?" (was on)\n":" (was off)\n");
      update_display();
      break;

    case SHOW_OUTER_: 
      old = outerflag;
      outerflag = (node->op1.intval == ON_) ? 1 : 0;
      outstring(outerflag ? "show outer ON." : "show outer OFF.");
      outstring(old?" (was on)\n":" (was off)\n");
      update_display();
      break;

    case INTERP_NORMALS_: 
      normflag = (node->op1.intval == ON_) ? 1 : 0;
      break;

    case COLORMAP_: 
      colorflag = (node->op1.intval == ON_) ? 1 : 0;
      break;

    case AMBIENT_PRESSURE_: 
      web.pressure_flag = (node->op1.intval == ON_) ? 1 : 0;
      if ( web.pressure_flag)
       { if ( !web.full_flag && !valid_id(web.outside_body) )
             add_outside();
         web.projection_flag = 0;
         sprintf(msg,"Ambient pressure ON; ambient pressure = %2.15g\n",
                    (double)web.pressure);
       }
      else
       {
         web.projection_flag = 1;
         sprintf(msg,"Ambient pressure OFF.\n");
       }
      calc_energy();
      outstring(msg);
      break;

    case EFFECTIVE_AREA_:
         if ( SDIM > 3 )
         { kb_error(3210,"effective_area only for dimension 3.\n",RECOVERABLE);}
         old = effective_area_flag;
         effective_area_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(effective_area_flag ? "effective_area ON." :
                                            "effective_area OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         if ( effective_area_flag ) web.area_norm_flag = 1;
         if ( square_curvature_flag ) calc_energy();
         break;

    case NORMAL_CURVATURE_:
         if ( SDIM != 3 )
         { kb_error(3217,"Normal_curvature only for dimension 3.\n",RECOVERABLE);}
         old = normal_curvature_flag;
         normal_curvature_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(normal_curvature_flag ? "normal_curvature ON." :
                                            "normal_curvature OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         if ( square_curvature_flag ) calc_energy();
         break;

    case DIV_NORMAL_CURVATURE_:
         if ( SDIM != 3 )
         { kb_error(3212,"Div_normal_curvature only for dimension 3.\n",RECOVERABLE);}
         old = div_normal_curvature_flag;
         div_normal_curvature_flag = (node->op1.intval==ON_) ? 1 : 0;
         outstring(div_normal_curvature_flag ? "div_normal_curvature ON." :
                                            "div_normal_curvature OFF.");
         outstring(old?" (was on)\n":" (was off)\n");
         if ( square_curvature_flag ) calc_energy();
         break;

    case LONG_JIGGLE_: 
         long_jiggle();
         break;

    case RAW_VERAVG_:
         vertex_average(NOVOLKEEP);
         recalc();
         outstring("Vertex averaging done.\n");
         break;

    case RAWEST_VERAVG_:
         vertex_average(RAWEST);
         recalc();
         outstring("Vertex averaging done.\n");
         break;

    case REBODY_:
      { body_id b_id;
         sprintf(msg,"New bodies: %d\n",rebody());
         outstring(msg);
         calc_content(Q_FIXED|Q_INFO);
         FOR_ALL_BODIES(b_id)
            if ( get_battr(b_id) & FIXEDVOL )  
              set_body_fixvol(b_id,get_body_volume(b_id));
         break;
      }

    case ALICE_:      
         alice();
         break;

    case BURCHARD_:      
         burchard(node->op1.intval);
         break;

    case LOAD_:
         /* startup(*(char**)(stacktop--)); */
         strncpy(loadfilename,*(char**)(stacktop--),sizeof(loadfilename));
         longjmp(loadjumpbuf,1);
         break;

    case DUMP_:
         if ( node->left ) do_dump(*(char**)(stacktop--));
         else do_dump(NULL);
         break;

    case SET_COLORMAP_:
         strncpy(cmapname,*(char**)(stacktop--),sizeof(cmapname));
         break;

    case RAW_CELLS_: 
         web.torus_body_flag = 0; web.torus_clip_flag = 0;
         torus_display_mode = TORUS_RAW_MODE;
         update_display();
         break;

    case CONNECTED_CELLS_: 
         if ( web.skel[BODY].count == 0 )
            kb_error(1234,"There are no bodies to display connectedly.\n",WARNING);

         else { web.torus_body_flag = 1; web.torus_clip_flag = 0; 
                  torus_display_mode = TORUS_CONNECTED_MODE;
                  update_display(); }
         break;

    case CLIPPED_CELLS_:
         if ( !web.torus_flag ) 
            kb_error(2397,"CLIPPED mode has no effect since torus model not used.\n",
                WARNING);
         torus_display_mode = TORUS_CLIPPED_MODE;
         web.torus_body_flag = 0; web.torus_clip_flag = 1;
         update_display();
         break;

    case THICKEN_: 
      thickenflag = (node->op1.intval == ON_) ? 1 : 0;
         update_display();
         break;

    case METRIC_CONVERT_:
         metric_convert_flag = (node->op1.intval == ON_) ? 1 : 0;
         if ( metric_convert_flag )
             outstring("Converting form to vector using metric.\n");
         else outstring("Not using metric to convert form to vector.\n");
         break;

    case QUANTITIES_ONLY_:
         old = quantities_only_flag;
         quantities_only_flag = (node->op1.intval == ON_) ? 1 : 0;
         outstring("Using quantities only ");
         outstring(quantities_only_flag ? "ON.": "OFF.");
         outstring(old ? "(was ON)\n" : "(was OFF)\n");
         break;

    case SQUARED_GRADIENT_:
         old = min_square_grad_flag;
         min_square_grad_flag = (node->op1.intval == ON_) ? 1 : 0;
         outstring("Squared gradient minimization with Hessian ");
         outstring(min_square_grad_flag ? "ON.": "OFF.");
         outstring(old ? "(was ON)\n" : "(was OFF)\n");
         break;

    case LINEAR_METRIC_:
         old = hessian_linear_metric_flag;
         hessian_linear_metric_flag = (node->op1.intval == ON_) ? 1 : 0;
         outstring("Linear interpolation metric with Hessian ");
         outstring(hessian_linear_metric_flag ? "ON.": "OFF.");
         outstring(old ? "(was ON)\n" : "(was OFF)\n");
         break;

    case YSMP_: 
      old = ysmp_flag;
      ysmp_flag = 
          (node->op1.intval == ON_) ? YSMP_FACTORING : MINDEG_FACTORING;
      if ( ysmp_flag == MINDEG_FACTORING )
      {
         sp_mul_func = bk_mul;
         sp_AIJ_setup_func= bk_AIJ_setup;
         sp_constraint_setup_func = bk_constraint_setup;
         sp_hess_project_setup_func= BK_hess_project_setup;
         sp_factor_func = xmd_factor;
         sp_solve_func = xmd_solve;
         sp_solve_multi_func = xmd_solve_multi;
         sp_ordering_func = NULL;
         outstring("Using alternate minimal degree. ");
         switch ( old ) 
         {  case YSMP_FACTORING: outstring("(was YSMP)\n"); break;
             case MINDEG_FACTORING: outstring("(was alt min deg)\n"); break;
             case METIS_FACTORING: outstring("(was metis)\n"); break;
         }
      }
      else
      { 
         sp_mul_func = bk_mul;
         sp_AIJ_setup_func= bk_AIJ_setup;
         sp_constraint_setup_func = bk_constraint_setup;
         sp_hess_project_setup_func= BK_hess_project_setup;
         sp_factor_func = ysmp_factor;
         sp_solve_func = ysmp_solve;
         sp_solve_multi_func = ysmp_solve_multi;
         sp_ordering_func = NULL;
         outstring("Using YSMP. ");
         switch ( old ) 
         {  case YSMP_FACTORING: outstring("(was YSMP)\n"); break;
             case MINDEG_FACTORING: outstring("(was alt min deg)\n"); break;
             case METIS_FACTORING: outstring("(was metis)\n"); break;
         }
      }
      break;

    case METIS_FACTOR_:
         old = ysmp_flag;
         if ( node->op1.intval == ON_ )
         {
#ifndef METIS 
         kb_error(2333,"This Evolver not compiled with METIS.\n",RECOVERABLE);
#else
         ysmp_flag = METIS_FACTORING;
         sp_mul_func = bk_mul;
         sp_AIJ_setup_func= bk_AIJ_setup;
         sp_constraint_setup_func = bk_constraint_setup;
         sp_hess_project_setup_func= BK_hess_project_setup;
         sp_factor_func = tree_factor;
         sp_solve_func = tree_solve;
         sp_solve_multi_func = tree_solve_multi;
         sp_ordering_func = metis_order;
         outstring("Using Metis tree factoring. ");
         switch ( old ) 
         {  case YSMP_FACTORING: outstring("(was YSMP)\n"); break;
             case MINDEG_FACTORING: outstring("(was alt min deg)\n"); break;
             case METIS_FACTORING: outstring("(was metis)\n"); break;
         }
#endif
         }
         else
         { 
         ysmp_flag = YSMP_FACTORING;
         sp_mul_func = bk_mul;
         sp_AIJ_setup_func= bk_AIJ_setup;
         sp_constraint_setup_func = bk_constraint_setup;
         sp_hess_project_setup_func= BK_hess_project_setup;
         sp_factor_func = ysmp_factor;
         sp_solve_func = ysmp_solve;
         sp_solve_multi_func = ysmp_solve_multi;
         sp_ordering_func = NULL;
         outstring("Using YSMP. ");
         switch ( old ) 
         {  case YSMP_FACTORING: outstring("(was YSMP)\n"); break;
             case MINDEG_FACTORING: outstring("(was alt min deg)\n"); break;
             case METIS_FACTORING: outstring("(was metis)\n"); break;
         }
         }

         break;

     case BUNCH_KAUFFMAN_:
          flip_toggle(&BK_flag,node->op1.intval,
                 "Bunch-Kaufman version of minimal degree");
             break;

      case CONTINUE_:
      case SET_PROC_END_:
      case SHOW_END_:
      case ELSE_:
      case COND_ELSE_:
         /* just a continue node */
         break;

         default:
            sprintf(errmsg,"Bad expression eval node type: %s.",
                tokname(node->type));
            kb_error(1235,errmsg,RECOVERABLE);

            break;

  }
  return stacktop;
} /* end more_other_stuff() */


/********************************************************************
*
* function: get_toggle_value()
*
* purpose:  get boolean value of toggle for expression 
*/

int get_toggle_value(tog)
int tog;  /* type of toggle */
{
  switch (tog)
    {  case GEOMVIEW_: return geomview_flag; 
        case AMBIENT_PRESSURE_: return web.pressure_flag;
        case ZENER_DRAG_: return zener_drag_flag;
        case BACKCULL_: return backcull_flag;
        case TORUS_FILLED_: return web.full_flag;
        case VOLGRADS_EVERY_: return volgrads_every_flag;
        case LOGFILE_: return logfile_flag;
        case LINEAR_: return (web.modeltype == LINEAR);
        case QUADRATIC_: return (web.modeltype == QUADRATIC);
        case LAGRANGE_: return (web.modeltype == LAGRANGE);
        case KUSNER_: return kusner_flag; 
        case ESTIMATE_: return estimate_flag;
        case DETURCK_: return unit_normal_flag;
        case HOMOTHETY_: return web.homothety;
        case SQGAUSS_: return sqgauss_flag; 
        case AUTOPOP_: return autopop_flag;
        case AUTOCHOP_: return autochop_flag;
        case QUIET_: return quiet_flag; 
        case HESSIAN_QUIET_: return hessian_quiet_flag; 
        case QUIETGO_: return quiet_go_flag; 
        case OLD_AREA_: return old_area_flag;
        case APPROX_CURV_: return approx_curve_flag;
        case RUNGE_KUTTA_: return runge_kutta_flag; 
        case CHECK_INCREASE_: return check_increase_flag; 
        case DEBUG_: return yydebug; 
        case MEAN_CURV_: return web.area_norm_flag; 
        case DIFFUSION_: return web.diffusion_flag; 
        case GRAVITY_: return web.gravflag; 
        case CONJ_GRAD_: return conj_grad_flag; 
        case TRANSFORMS_: return transforms_flag; 
        case CONF_EDGE_SQCURV_: return conf_edge_curv_flag; 
        case EFFECTIVE_AREA_: return effective_area_flag; 
        case RAW_CELLS_: return !web.torus_body_flag && !web.torus_clip_flag; 
        case CONNECTED_CELLS_: return web.torus_body_flag; 
        case CLIPPED_CELLS_: return web.torus_clip_flag; 
        case THICKEN_: return thickenflag; 
        case YSMP_: return ysmp_flag; 
        case LINEAR_METRIC_: return hessian_linear_metric_flag; 
        case METRIC_CONVERT_: return metric_convert_flag; 
        case BUNCH_KAUFFMAN_: return BK_flag; 
        case SHOW_INNER_: return innerflag; 
        case SHOW_OUTER_: return outerflag; 
        case COLORMAP_: return colorflag; 
        case HESSIAN_DIFF_: return hessian_by_diff_flag; 
        case POST_PROJECT_: return post_project_flag; 
        case MEAN_CURV_INT_: return mean_curv_int_flag; 
        case OPTIMIZE_: return !web.motion_flag; 
        case NORMAL_CURVATURE_: return normal_curvature_flag; 
        case DIV_NORMAL_CURVATURE_: return div_normal_curvature_flag; 
        case SHADING_: return shading_flag; 
        case FACET_COLORS_: return color_flag; 
        case BOUNDARY_CURVATURE_: return boundary_curvature_flag; 
        case NORMAL_MOTION_: return normal_motion_flag; 
        case PINNING_: return check_pinning_flag; 
        case VIEW_4D_: return view_4D_flag; 
        case MEMDEBUG_: return memdebug; 
        case ITDEBUG_: return itdebug; 
        case METRIC_CONVERSION_: return metric_convert_flag; 
        case AUTORECALC_: return autorecalc_flag; 
        case FORCE_POS_DEF_: return make_pos_def_flag; 
        case GV_BINARY_: return gv_binary_flag; 
        case SELF_SIMILAR_: return self_similar_flag; 
        case AUTODISPLAY_: return go_display_flag; 
        case RIBIERE_CG_: return ribiere_flag; 
        case ASSUME_ORIENTED_: return assume_oriented_flag; 
        case HESSIAN_NORMAL_: return hessian_normal_flag; 
        case HESSIAN_SPECIAL_NORMAL_: return hessian_special_normal_flag; 
        case HESSIAN_NORMAL_PERP_: return hessian_normal_perp_flag; 
        case HESSIAN_NORMAL_ONE_: return hessian_normal_one_flag; 
        case HESSIAN_DOUBLE_NORMAL_: return hessian_double_normal_flag; 
        case INTERP_BDRY_PARAM_: return interp_bdry_param; 
        case H_INVERSE_METRIC_: return web.h_inverse_metric_flag; 
        case PSCOLORFLAG_: return ps_colorflag;
        case GRIDFLAG_: return gridflag;
        case METIS_FACTOR_: return ysmp_flag == METIS_FACTORING;
        case CROSSINGFLAG_: return crossingflag;
        case LABELFLAG_: return labelflag;
        case SHOW_ALL_QUANTITIES_: return show_all_quantities;
        case QUANTITIES_ONLY_: return quantities_only_flag;
        default:
            sprintf(errmsg,"Internal error:  Toggle value omitted for toggle %d.\n",tog);
             kb_error(1236,errmsg,WARNING);
    }
    return 0;
} /* end get_toggle_value */

/*************************************************************************
*
* function: get_internal_variable()
*
* purpose: get value of various variables visible to user.
*/

REAL get_internal_variable(vartok)
int vartok; /* token number of variable */
{ 
  switch(vartok)
  { case V_RANDOM: return kb_drand();
    case V_VERTEXCOUNT:
      return (REAL)web.skel[VERTEX].count; 
    case V_EDGECOUNT:
      return (REAL)web.skel[EDGE].count; 
    case V_FACETCOUNT:
      return (REAL)web.skel[FACET].count; 
    case V_BODYCOUNT:
      return (REAL)web.skel[BODY].count; 
    case V_FACETEDGECOUNT:
      return (REAL)web.skel[FACETEDGE].count; 
    case V_ENERGY:
      return web.total_energy; 
    case V_GAP_CONSTANT:
      return web.spring_constant; 
    case V_TIME:
      return total_time; 
    case V_JIG_TEMP:
      return web.temperature; 
    case V_AREA:
    case V_LENGTH:
      return web.total_area; 
    case V_ITER_COUNTER:
      return (REAL)gocount; 
    case V_AMBIENT_PRESSURE:
      return web.pressure; 
    case V_DIFFUSION:
      return web.diffusion_const; 
    case V_SCALE:
      return web.scale; 
    case V_SCALE_SCALE:
      return web.scale_scale; 
    case V_SPACE_DIMENSION:
      return (REAL)SDIM; 
    case V_SURFACE_DIMENSION:
      return (REAL)web.dimension; 
    case V_TORUS:
      return (REAL)web.torus_flag; 
    case V_SYMMETRY_GROUP:
      return (REAL)web.symmetry_flag; 
    case V_SIMPLEX:
      return (REAL)(web.representation == SIMPLEX); 
    case V_INTEGRAL_ORDER:
      return ((REAL)web.gauss1D_order+2)/2; 
    case V_INTEGRAL_ORDER_1D:
      return (REAL)web.gauss1D_order; 
    case V_INTEGRAL_ORDER_2D:
      return (REAL)web.gauss2D_order; 
    case V_TOLERANCE:
      return web.tolerance; 
    case V_TARGET_TOLERANCE:
      return web.target_tolerance; 
    case V_THICKNESS:
      return thickness; 
    case V_HESSIAN_SLANT_CUTOFF:
      return hessian_slant_cutoff; 
    case V_HESS_EPSILON:
      return hessian_epsilon; 
    case GRAV_CONST_:
      return web.grav_const; 
    case V_EQUI_COUNT:
      return (REAL)web.equi_count; 
    case V_DELETE_COUNT:
      return (REAL)web.delete_count; 
    case V_REFINE_COUNT:
      return (REAL)web.refine_count; 
    case V_NOTCH_COUNT:
      return (REAL)web.notch_count; 
    case V_DISSOLVE_COUNT:
      return (REAL)web.dissolve_count; 
    case V_POP_COUNT:
      return (REAL)web.pop_count; 
    case V_WHERE_COUNT:
      return (REAL)web.where_count; 
    case V_EIGENPOS:
      return (REAL)eigen_pos; 
    case V_EIGENNEG:
      return (REAL)eigen_neg; 
    case V_EIGENZERO:
      return (REAL)eigen_zero; 
    case V_PICKVNUM:
      return (REAL)pickvnum; 
    case V_PICKENUM:
      return (REAL)pickenum; 
    case V_PICKFNUM:
      return (REAL)pickfnum; 
    case V_LINEAR_METRIC_MIX:
      return linear_metric_mix; 
    case V_QUADRATIC_METRIC_MIX:
      return quadratic_metric_mix; 
    case V_RANDOM_SEED:
      return (REAL)random_seed; 
    case V_LAST_EIGENVALUE:
      return last_eigenvalue; 
    case V_LAST_HESSIAN_SCALE:
      return last_hessian_scale; 
    case V_LAGRANGE_ORDER:
      return (REAL)web.lagrange_order; 
    case V_SCALE_LIMIT:
      return web.maxscale; 
    case V_BRIGHTNESS:
      return brightness; 
    case V_LAST_ERROR: return last_error;
    case V_BACKGROUND: return background_color;
    case V_TRANSFORM_COUNT:
      return transform_count + 1.;
    case V_CLOCK:
#if defined (__SYS_TIMES_H__) && defined(HZ)
      { struct tms t;
        times(&t);
        return t.tms_utime/(REAL)HZ;
      }
#else
#if defined (_SYS_TIMES_H) && defined(CLK_TCK)
      { struct tms t;    /* SGI, anyway */
        times(&t);
        return t.tms_utime/(REAL)CLK_TCK;
      }
#else
#ifdef CLOCKS_PER_SEC
      return clock()/(REAL)CLOCKS_PER_SEC;
#else
      return 0.0;
#endif
#endif
#endif
      break;

    case V_MEMARENA:
#if defined(M_MXFAST) && defined(IRIS)
      { struct mallinfo m = mallinfo();
        return (REAL)m.arena;
      }
#else
#ifdef MSC
      { struct _heapinfo hinfo;
        long mem_use=0,mem_free=0;
        hinfo._pentry = NULL;
        while ( _heapwalk(&hinfo) == _HEAPOK )
           if (hinfo._useflag) { mem_use+= hinfo._size; }
           else {  mem_free += hinfo._size; }
        return (REAL)(mem_use+mem_free);
      }          

#else
#ifdef TC
      { struct heapinfo hinfo;
        long mem_use=0,mem_free=0;
        hinfo.ptr = NULL;
        while ( heapwalk(&hinfo) == _HEAPOK )
           if (hinfo.in_use) { mem_use+= hinfo.size; }
           else {  mem_free += hinfo.size; }
        return (REAL)(mem_use+mem_free);
      }          

#else
      return 0.0;
#endif
#endif
#endif
      break;

    case V_MEMUSED:
#if defined(M_MXFAST) && defined(IRIS)
      { struct mallinfo m = mallinfo();
        return (REAL)(m.uordblks + m.usmblks);
      }
#else
#ifdef MSC
      { struct _heapinfo hinfo;
        long mem_use=0,mem_free=0;
        hinfo._pentry = NULL;
        while ( _heapwalk(&hinfo) == _HEAPOK )
           if (hinfo._useflag) { mem_use+= hinfo._size; }
           else {  mem_free += hinfo._size; }
        return (REAL)(mem_use);
      }          

#else
      return 0.0;
#endif
#endif
      break;

    default: 
      sprintf(errmsg,"Internal: illegal variable number %d.\n",vartok);
      kb_error(1208, errmsg,RECOVERABLE);

  }
  return 0.0;
}

/**********************************************************************
*
*  function: tree_copy()
*
*  purpose: copy tree in linear postorder form
*              allocates space for tree in destination
*
*/

void tree_copy(dest,src)
struct expnode *dest;
struct treenode *src;
{
  struct treenode *enode;
  int count,n;
  if ( dest == NULL ) return;
  if ( dest->start ) myfree((char*)dest->start);
  if ( src == NULL ) { dest->start = dest->root = NULL; return; }
  for ( enode=src; enode->left!=0; enode+=enode->left );
  count = src - enode + 1;
  dest->start = (struct treenode *)mycalloc(count+2,sizeof(struct treenode));
  dest->flag |= USERCOPY;
  memcpy((char*)(dest->start+1),(char*)enode,count*sizeof(struct treenode));
  dest->root = dest->start + count;
  /* make copies of strings */
  for ( enode=dest->start+1, n = 0 ; n < count ; enode++,n++ )
     if ( enode->flags & HAS_STRING )
     { char *s = mycalloc(strlen(enode->op1.string)+1,sizeof(char));
        strcpy(s,enode->op1.string);
        enode->op1.string = s;
     }
  /* copy root to first place */
  dest->start[0] = *src;
  if ( dest->start[0].left ) dest->start[0].left += count;
  if ( dest->start[0].right ) dest->start[0].right += count;

  /* FINISH node */
  dest->start[count+1].type = FINISHED;
}


/*******************************************************************
*
* function: reduce_strings()
*
* purpose:  converts escape sequences to ASCII in place  
*
*/
void reduce_string(s) 
char *s;
{ char *c = s;
  while ( *s ) /* reduce escape sequences */
  if ( *s == '\\' )
  { 
    switch ( s[1] )
    { case 'n' : *(c++) = '\n'; s += 2; break;
      case 't' : *(c++) = '\t'; s += 2; break;
      case 'b' : *(c++) = '\b'; s += 2; break;
      case 'r' : *(c++) = '\r'; s += 2; break;
      case '\\' : *(c++) = '\\'; s += 2; break;
      case 'q' : *(c++) = '"'; s += 2; break;
      default:  *(c++) =  s[1]; s += 2; break;
    }
  }
  else if (*s == '\n' ) s++;  /* omit newlines */
  else *(c++) = *(s++);
  *c = '\0';
}

/***********************************************************************
*
* function: set_body()
*
* purpose: set facet body, or edge body in string model.
*          Can set to NULLID.
*/
void set_body(id,b_id)
element_id id; /* edge or facet, inverted for backbody */
body_id b_id;
{
  if ( id_type(id) == FACET )
  { set_facet_body(id,b_id);
    return;
  }

  if ( (id_type(id) != EDGE) || (web.representation != STRING) )
    kb_error(4345,"Trying to set body of wrong type element.\n",RECOVERABLE);

  if ( !valid_id(b_id) )
  { facetedge_id fe,fe_id,next_fe,prev_fe;
    facet_id f_id;

    fe_id = get_edge_fe(id);
    f_id = get_fe_facet(fe_id);
    if ( !valid_id(f_id) ) return;
    if ( inverted(f_id) ) 
    { fe_id = get_next_facet(fe_id);
      f_id = get_fe_facet(fe_id);
    } 
    if ( !valid_id(f_id) ) return;
    if ( inverted(f_id) ) return; 

    /* check for being at beginning or end of edge arc */
    fe = fe_id;
    if ( valid_id(get_next_edge(fe)) && valid_id(get_prev_edge(fe)) )
    { /* check for full loop */
      do { fe = get_next_edge(fe); }
      while ( valid_id(fe) && !equal_id(fe,fe_id) );
      if ( !valid_id(fe) )
      { if ( verbose_flag )
        { sprintf(msg,"Not dissolving edge %d since would make facet two arcs.\n",
            ordinal(id)+1);
          outstring(msg);
        }
        return;
      }
    }
    /* Okay, ready to party. */
    next_fe = get_next_edge(fe_id);
    prev_fe = get_prev_edge(fe_id);
    if ( valid_id(next_fe) ) 
    { set_prev_edge(next_fe,NULLID);
      set_facet_fe(f_id,next_fe);
    }
    if ( valid_id(prev_fe) )
      set_next_edge(prev_fe,NULLID);
    else if ( !valid_id(next_fe) ) 
      set_facet_fe(f_id,NULLID);
    if ( valid_id(get_next_facet(fe_id)) )
      set_edge_fe(id,get_next_facet(fe_id));
    set_next_facet(get_prev_facet(fe_id),get_next_facet(fe_id));
    set_prev_facet(get_next_facet(fe_id),get_prev_facet(fe_id));
    free_element(fe_id);
  }
  else /* valid id */
  { /* string model */
    facetedge_id oldfe = get_body_fe(b_id);
    facet_id f_id;
    facetedge_id efe,eefe,fe,start_fe;
    facetedge_id newfe;
    int inserted_flag=0;  /* whether a linkup made */

    if ( !valid_id(oldfe) )
        kb_error(3255,
            "Need to set the body's facet before assigning edges.\n",
                          RECOVERABLE);
    f_id = get_fe_facet(oldfe);
    newfe = new_facetedge(f_id,id);
    efe = get_edge_fe(id);
    if ( valid_id(get_fe_facet(efe)) )
    { eefe = get_next_facet(efe);
      set_next_facet(efe,newfe);
      set_next_facet(newfe,eefe);
      set_prev_facet(eefe,newfe);
      set_prev_facet(newfe,efe);
    }
    else
    { if (valid_id(efe)) free_element(efe);
      set_next_facet(newfe,newfe);
      set_prev_facet(newfe,newfe);
      set_edge_fe(id,newfe);
    }
    /* see if can place at head or end of chain */
    fe = start_fe = get_facet_fe(f_id);
    if ( !valid_id(fe) )
    { /* naked facet */
      set_facet_fe(f_id,newfe); inserted_flag = 1;
    }
    else
    { if ( !valid_id(get_prev_edge(fe)) )
      { if ( equal_id(get_fe_tailv(fe),get_edge_headv(id)) )
        { set_next_edge(newfe,fe); set_prev_edge(fe,newfe); 
          set_facet_fe(f_id,newfe); inserted_flag = 1;
        }
      }
      while ( valid_id(get_next_edge(fe)) &&
                        !equal_id(get_next_edge(fe),start_fe) )
           fe = get_next_edge(fe);
      if ( !valid_id(get_next_edge(fe)) )
      { if ( equal_id(get_fe_headv(fe),get_edge_tailv(id)) )
        { set_next_edge(fe,newfe); set_prev_edge(newfe,fe); 
          inserted_flag = 1;
        }
      }
    }
    if ( !inserted_flag ) 
    {   sprintf(errmsg,"Could not link edge %d onto facet %d chain.\n",
           ordinal(id)+1,ordinal(f_id)+1);
        kb_error(3855,errmsg,WARNING);
    }

  }
} /* end set_body() */
