/*************************************************************
*  This file is part of the Surface Evolver source code.     *
*  Programmer:  Ken Brakke, brakke@susqu.edu                 *
*************************************************************/


/* for tracking memory usage */
#ifdef MEMSTRINGS
#define mycalloc(num,size)  kb_calloc(num,size,__FILE__,__LINE__)
#define kb_realloc(ptr,new,old) KB_realloc(ptr,new,old,__FILE__,__LINE__)
#define dmatrix(a,b,c,d) kb_dmatrix(a,b,c,d,__FILE__,__LINE__)
#define dmatrix3(a,b,c) kb_dmatrix3(a,b,c,__FILE__,__LINE__)
#define dmatrix4(a,b,c,d) kb_dmatrix4(a,b,c,d,__FILE__,__LINE__)
#else
#define mycalloc(num,size)  kb_calloc(num,size)
#define kb_realloc(ptr,new,old) KB_realloc(ptr,new,old)
#define dmatrix(a,b,c,d) kb_dmatrix(a,b,c,d)
#define dmatrix3(a,b,c) kb_dmatrix3(a,b,c)
#define dmatrix4(a,b,c,d) kb_dmatrix4(a,b,c,d)
#endif

/* Precision */
#ifdef LONGDOUBLE
#define REAL  long double 
#define DOT    dot 
#define DWIDTH ((sizeof(REAL)==16) ? 35 : 22)
#define DPREC ((sizeof(REAL)==16) ? 32 : 19)
#else
#ifdef FLOAT
#define REAL float
#define DOT  dotf
#define v3d  v3f
#define v2d  v2f  
#else 
#define REAL  double 
#define DOT    dot 
#endif
#endif

/* for things we really want to be plain double */
#define DOUBLE double

#include <time.h>

/* following also works for Cray */
#ifdef GENERIC
/* Adjust header file names if yours are different */
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <setjmp.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>
#include <sys/param.h>
#define is_finite(x) (((x)>(-1e300))&&((x)<1e300)) 
/* MAXALLOC is maximum size allocable by calloc() */
#define    MAXALLOC 0x7FFFFFFFL
/* Some don't declare calloc, getenv, and bsearch in header files */
char *calloc();
char *getenv();
char *bsearch();
/* PATHCHAR is name-separating character in paths */
#define PATHCHAR '/'
/* ENVPATHCHAR is the path separating character in environment strings */
#define ENVPATHCHAR ":"
/* NOPROTO should be defined for systems that don't do ANSI prototypes */
#define NOPROTO
#endif

#ifdef _HPUX_SOURCE
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <setjmp.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>
#include <sys/param.h>
#define is_finite(x) (((x)>(-1e300))&&((x)<1e300)) 
#define    MAXALLOC 0x7FFFFFFFL
char *getenv();
#define PATHCHAR '/'
#define ENVPATHCHAR ":"
#define FCAST (int(*)(const void*,const void *))
#endif

#ifdef LINUX
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <setjmp.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>
#include <sys/param.h>
#define is_finite(x) (((x)>(-1e300))&&((x)<1e300)) 
/* MAXALLOC is maximum size allocable by calloc() */
#define    MAXALLOC 0x7FFFFFFFL
/* Some don't declare calloc, getenv, and bsearch in header files */
char *calloc();
char *getenv();
char *bsearch();
/* PATHCHAR is name-separating character in paths */
#define PATHCHAR '/'
/* ENVPATHCHAR is the path separating character in environment strings */
#define ENVPATHCHAR ":"
#define FCAST (int(*)(const void*,const void *))
#endif

#ifdef __convex_SPP
/* extra parallel stuff beyond _HPUX_SOURCE */
#include <spp_prog_model.h>
#define m_get_myid my_thread
extern int                  m_fork(void (*)(), int,int);
#define m_park_procs()
#define m_rele_procs()
#define m_sync()
#define m_kill_procs()
#define m_set_procs()
#define m_get_numprocs num_procs
#define m_next(void)    not for Convex yet
#define M_LOCK(addr)      lock_gate(&gate1)        
#define M_UNLOCK(addr)    unlock_gate(&gate1)
extern gate_t gate1;
#ifndef PVM
#define SHARED_MEMORY
#endif
#endif

#ifdef DECALPHA
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <setjmp.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>
#include <sys/param.h>
#define is_finite(x) finite(x)
#define    MAXALLOC 0x7FFFFFFFL
char *getenv();
#define PATHCHAR '/'
#define ENVPATHCHAR ":"
#endif


#ifdef MAC_APP
/* this was done with THINK C */
/* Adjust header file names if yours are different */
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <setjmp.h>
#include <signal.h>
#include <unix.h>
#define is_finite(x) (((x)>(-1e300))&&((x)<1e300)) 
/* MAXALLOC is maximum size allocable by calloc() */
#define    MAXALLOC 0x7FFFFFFFL
/* Macs don't have environments */
#define getenv(a)  ":fe;:doc"
/* PATHCHAR is name-separating character in paths */
#define PATHCHAR ':'
/* ENVPATHCHAR is the path separating character in environment strings */
#define ENVPATHCHAR ";"
#define drand48()    ((REAL)rand()/RAND_MAX)
#define srand48(seed)    srand(seed)
#define NOPIPE
/* cast for comparison functions for qsort() and bsearch() */
#define FCAST (__cmp_func)
/* to avoid Think C name conflict in a header file */
#define extend my_extend
#define NO_YACC_DEBUG
/* for workaround of Symantec 8.0.5 68K compiler bug */
#define ANSI_DEF
#endif


#ifdef MAC_CW
/* this was done with CODE WARRIOR */
/* Adjust header file names if yours are different */
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <setjmp.h>
#include <signal.h>
#include <unix.h>
#include <alloca.h>
#define is_finite(x) (((x)>(-1e300))&&((x)<1e300)) 
/* MAXALLOC is maximum size allocable by calloc() */
#define    MAXALLOC 0x7FFFFFFFL
/* Macs don't have environments */
#define getenv(a)  ":fe;:doc;::fe;::doc;:::fe;:::doc"
/* PATHCHAR is name-separating character in paths */
#define PATHCHAR ':'
/* ENVPATHCHAR is the path separating character in environment strings */
#define ENVPATHCHAR ";"
#define drand48()    ((REAL)rand()/RAND_MAX)
#define srand48(seed)    srand(seed)
#define NOPIPE
/* cast for comparison functions for qsort() and bsearch() */
#define FCAST (void*)
/* to avoid Think C name conflict in a header file */
#define extend my_extend
/* for workaround of Symantec 8.0.5 68K compiler bug */
#define ANSI_DEF
#endif


#ifdef IRIS

/* this weirdness was needed on an IRIX 5.2 machine */
/* due to inconsistent use of _SIZE_T and _SIZE_T_ in headers */
#if !defined(_SIZE_T) && !defined(_SIZE_T_)
typedef unsigned size_t;
#endif
#define _SIZE_T
#define _SIZE_T_

/* Kludge for IRIX 6.5 */
#include <ctype.h>
#undef isalpha
#undef isdigit
#undef isprint
#undef islower
#undef isupper

#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <bstring.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <memory.h>
#include <setjmp.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>
#include <sys/param.h>
#include <stropts.h>
#include <poll.h>
#include <fp_class.h>
#define  is_finite(x) finite(x)
#define    MAXALLOC 0x7FFFFFFFL
#define PATHCHAR '/'
#define ENVPATHCHAR ":"
#define FCAST (int(*)(const void*,const void *))
#ifdef SGI_MULTI
#define PARALLEL_MACHINE
#define SHARED_MEMORY
#include <task.h>
extern int _fork(void (*)(), ...);
extern int _park_procs(void);
extern int _rele_procs(void);
extern void _sync(void);
extern int _kill_procs(void);
extern int _get_myid(void);
extern int _set_procs(int);
extern int _get_numprocs(void);
extern unsigned m_next(void);
extern void m_lock(void);              
extern void m_unlock(void); 

#endif
#endif

#ifdef SGI_MULTI
#include <ulocks.h>
#include <task.h>
#define M_LOCK(addr) (locklist[0]?ussetlock(locklist[((long)(addr))&(_MAXLOCKS-1)]):0)
#define M_UNLOCK(addr) (locklist[0]?usunsetlock(locklist[((long)(addr))&(_MAXLOCKS-1)]):0)
#endif


#ifdef KSR
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <setjmp.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>
#include <sys/param.h>
#define  is_finite(x) finite(x)
#define MAXALLOC 0x7FFFFFFFL
#define PATHCHAR '/'
#define ENVPATHCHAR ":"
#define FCAST (int(*)(const void*,const void *))
#define PARALLEL_MACHINE
#define SHARED_MEMORY
#else
#define __align128
#endif


#ifdef KSR
#include <ksr/numarg.h>
#include <pthread.h>

/* KSR says avoid _gspwt() */
#define M_LOCK(addr) {long xi; while(!_gspnwt(addr)) { \
                  while(++xi & 0x7f) continue; \
                  if ( xi > 60*1000000*20/4) abort(); } }

#define M_UNLOCK(addr) _rsp(addr) 
extern __shared pthread_mutex_t ksrlock;
extern int team_id = -1;

extern int ksr_counter;
extern int m_next();
#define NUMPROCS 64
#define m_get_numprocs() NUMPROCS
#define m_rele_procs()     0
#define m_park_procs()     0

extern void m_fork();  
#endif

#ifdef GCC
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <setjmp.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>
#include <sys/param.h>
#define is_finite(x) (((x)>(-1e300))&&((x)<1e300)) 
#define MAXALLOC 0x7FFFFFFFL
#ifdef MSDOS
#define PATHCHAR '\\'
#define ENVPATHCHAR ";"
#else
#define PATHCHAR '/'
#define ENVPATHCHAR ":"
#endif
#define FCAST (int(*)(const void*,const void *))
#define drand48() ((REAL)random()/0x7FFFFFFF)
#define srand48(seed)    srandom(seed)
#endif


#ifdef SUN
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <malloc.h>
#ifdef strdup
#undef strdup
#endif
#include <string.h>
#include <stdio.h>
#include <memory.h>
#include <setjmp.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>
#include <sys/param.h>
#include <values.h>
#define is_finite(x)  finite(x)
#define MAXALLOC 0x7FFFFFFFL
char *getenv();
/* char *calloc(); */
#define PATHCHAR '/'
#define ENVPATHCHAR ":"
/* #define  memmove(dest,src,n) kb_memmove(dest,src,n) */
#define NOPROTO
#endif


#ifdef XENIX386
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <setjmp.h>
#include <signal.h>
#include <values.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/param.h>
#include <prototypes.h>
void srand48(int);
#define is_finite(x) (((x)>(-1e300))&&((x)<1e300)) 
int _filbuf(FILE *);    /* for prototype */
#define    MAXALLOC 0x7FFFFFFFL
#define  memmove(dest,src,n) kb_memmove(dest,src,n)
char *getenv(char *);
#define PATHCHAR '/'
#define ENVPATHCHAR ":"
#endif


#ifdef NeXT
#include <ctype.h>
#include <fcntl.h>
#include <errno.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <setjmp.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>
#include <sys/param.h>
#define is_finite(x)  finite(x)
#define    MAXALLOC 0x7FFFFFFFL
#define PATHCHAR '/'
#define ENVPATHCHAR ":"
#define FCAST (int(*)(const void*,const void *))
#define srand48(seed)    srand(seed)
#endif

#ifdef MSC
#define _USERENTRY
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <float.h>
#include <fcntl.h>
#include <io.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <setjmp.h>
#include <signal.h>
#include <limits.h>
#include <direct.h>
#include <process.h>
#include <sys/types.h>
#define is_finite(x) (((x)>(-1e300))&&((x)<1e300)) 
#define MAXINT INT_MAX
#define    MAXALLOC 0x7FFFFFFFL
#define PATHCHAR '\\'
#define ENVPATHCHAR ";"
#define farcalloc(num,size)  (void far *)halloc(num,size)
#define farfree(ptr)            hfree((void huge *)ptr)
#define FCAST (int(*)(const void*,const void *))
#define NOPIPE
#define _export
#define drand48()    ((REAL)rand()/RAND_MAX)
#define srand48(seed)    srand(seed)
#define FPRESET _fpreset()
/* since graphics in parallel thread */
#define PARALLEL_MACHINE
#endif

#ifdef __WIN32__ 
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <mem.h>
#include <alloc.h>
#include <malloc.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <fcntl.h>
#include <conio.h>
#include <setjmp.h>
#include <signal.h>
#include <values.h>
#include <process.h>
#define is_finite(x) (((x)>(-1e300))&&((x)<1e300)) 
#define    MAXALLOC 0x7FFFFFFFL
#define PATHCHAR '\\'
#define ENVPATHCHAR ";"
#define NOPIPE
#pragma hdrstop
/* so qsort() and bsearch() below agree with prototypes */
#define FCAST (int(*)(const void*,const void *))
#define drand48()    ((REAL)rand()/RAND_MAX)
#define srand48(seed)    srand(seed)
#endif

#ifdef LCC 
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <fcntl.h>
#include <conio.h>
#include <setjmp.h>
#include <signal.h>
#include <process.h>
#define is_finite(x) (((x)>(-1e300))&&((x)<1e300)) 
#define    MAXALLOC 0x7FFFFFFFL
#define PATHCHAR '\\'
#define ENVPATHCHAR ";"
#define NOPIPE
#pragma hdrstop
/* so qsort() and bsearch() below agree with prototypes */
#define FCAST (int(*)(const void*,const void *))
#define drand48()    ((REAL)rand()/RAND_MAX)
#define srand48(seed)    srand(seed)
#endif

#ifdef PVM
#define PARALLEL_MACHINE
#include "pvm_kb.h"
#endif

#ifdef MPI
#define PARALLEL_MACHINE
#include "mpi_kb.h"
#endif

#ifndef SGI_MULTI
#define m_get_myid() 0
#endif

#ifdef TC
#define atold _atold
#endif

#ifndef TC
#ifndef huge
#define huge
#endif
#endif

#ifndef FCAST
#define FCAST
#endif

typedef int DY_OFFSET ;


#include "model.h"

#include "storage.h"

/* Some don't have these manifest constants in math.h (namely, Microsoft) */
#ifndef M_LN2
#define M_E                 2.71828182845904523536
#define M_PI                3.14159265358979323846
#define M_LN2              0.693147180559945309417
#endif

#ifdef LONGDOUBLE
#undef  M_E
#define M_E                 2.7182818284590452353602874713527L
#undef  M_PI
#ifdef TC
#define M_PI (atan(1.0)*4)
#else
#define M_PI                3.1415926535897932384626433832795L
#endif
#undef  M_LN2
#define M_LN2              0.693147180559945309417L
#endif

#ifndef DBL_EPSILON
#define DBL_EPSILON      2.2204460492503131e-16
#endif

/* can undefine or redefine these if your system has decent string functions */
#define stricmp(s1,s2)  kb_stricmp((s1),(s2))
#define strnicmp(s1,s2,n)  kb_strnicmp((s1),(s2),(n))
#ifndef strstr
#define strstr(s1,s2) kb_strstr((s1),(s2))
#endif
#ifndef strupr
#define strupr(a) kb_strupr(a)
#endif

/* Since tolower and toupper don't always  check case before converting */
#undef tolower
#undef toupper
#define tolower(c)    (isupper(c) ? ((c)-'A'+'a') : c)
#define toupper(c)    (islower(c) ? ((c)-'a'+'A') : c)

#ifndef MAXDOUBLE
#define MAXDOUBLE 1.0e38
#endif

#ifdef NOPROTO
#define ARGS(x) ()
#else
#define ARGS(x) x
#endif

#ifdef ENABLE_DLL
#include <dlfcn.h>
#endif

/* Evolver header files */
#include "skeleton.h"
#include "extern.h"
#include "express.h"
#include "quantity.h"
#include "web.h"
#ifdef NOPROTO
#include "noproto.h"
#else
#include "proto.h"
#endif

/* in case of non-parallel machines */
#ifndef M_LOCK
#define M_LOCK(addr)
#define M_UNLOCK(addr) 
#endif

#ifndef MAXINT
#define MAXINT (~(1<<(8*sizeof(int)-1)))
#endif

#ifndef FPRESET
#define FPRESET
#endif

#if defined(LONGDOUBLE) && !defined(NOLONGMATHFUNC)
/* have to do these after math.h */
#define sin sinl
#define cos cosl
#define tan tanl
#define asin asinl
#define acos acosl
#define atan atanl
#define sinh sinhl
#define cosh coshl
#define tanh tanhl
#define asinh asinhl
#define acosh acoshl
#define atanh atanhl
#define exp expl
#define log logl
#define pow powl
#define sqrt sqrtl
#define ceil ceill
#define fabs fabsl
extern REAL fabsl(REAL); /* wasn't in IRIX6.1 math.h */
#define floor floorl
#define fmod fmodl
#define modf modfl
#define atof atold
#endif


